#ifndef ALPHA_PARSING_INCLUDED
#define ALPHA_PARSING_INCLUDED


typedef union   _fpReg
{
     uint64  Asuint64;
     int64   Asint64;
     double  Asdouble;
     uint32  Asuint32;
     int32   Asint32;
     float   Asfloat;
}       fpReg;

// This code is lifted directly from AlphaProcessor.cxx in Bliss.  Eventually,
// it should be factored out of the file to make it reusable.

/************************************ AXP Execution Engine *************************************/


// General support functions used by the op-code simulation...

#	define	GenMaskForBit(zzz)			(((uint64)1)<<(zzz))
#	define	Make64(hhhh,llll)			((((uint64)(hhhh))<<32)|((uint64)(llll)))
#	define	Maxuint64					Make64(0xffffffff, 0xffffffff)
#	define	MaskLower32(zzz)			((zzz) & Make64(0x00000000, 0xffffffff))
#	define	MaskUpper32(zzz)			((zzz) & Make64(0xffffffff, 0x00000000))
#	define	SignExtend32To64(zzz)		(((zzz) & GenMaskForBit(31)) ? ((zzz)|Make64(0xffffffff, 0x00000000)) : (zzz))
#	define	SignExtend21To64(zzz)		(((zzz) & GenMaskForBit(20)) ? ((zzz)|Make64(0xffffffff, 0xffe00000)) : (zzz))
#	define	SignExtend16To64(zzz)		(((zzz) & GenMaskForBit(15)) ? ((zzz)|Make64(0xffffffff, 0xffff0000)) : (zzz))
#	define	SignExtend8To64(zzz)		(((zzz) & GenMaskForBit(7)) ? ((zzz)|Make64(0xffffffff, 0xffffff00)) : (zzz))

#	define	Maskb0(zzz)					((zzz)&Make64(0, 0x000000ff))
#	define	Maskb1(zzz)					((zzz)&Make64(0, 0x0000ff00))
#	define	Maskb2(zzz)					((zzz)&Make64(0, 0x00ff0000))
#	define	Maskb3(zzz)					((zzz)&Make64(0, 0xff000000))
#	define	Maskb4(zzz)					((zzz)&Make64(0x000000ff, 0))
#	define	Maskb5(zzz)					((zzz)&Make64(0x0000ff00, 0))
#	define	Maskb6(zzz)					((zzz)&Make64(0x00ff0000, 0))
#	define	Maskb7(zzz)					((zzz)&Make64(0xff000000, 0))
#	define	b0(zzz)						(((zzz)>>0) & Make64(0, 0xff))
#	define	b1(zzz)						(((zzz)>>8) & Make64(0, 0xff))
#	define	b2(zzz)						(((zzz)>>16) & Make64(0, 0xff))
#	define	b3(zzz)						(((zzz)>>24) & Make64(0, 0xff))
#	define	b4(zzz)						(((zzz)>>32) & Make64(0, 0xff))
#	define	b5(zzz)						(((zzz)>>40) & Make64(0, 0xff))
#	define	b6(zzz)						(((zzz)>>48) & Make64(0, 0xff))
#	define	b7(zzz)						(((zzz)>>56) & Make64(0, 0xff))
#	define	w0(zzz)						((zzz) & Make64(0, 0xffff))
#	define	w1(zzz)						(((zzz)>>16) & Make64(0, 0xffff))
#	define	w2(zzz)						(((zzz)>>32) & Make64(0, 0xffff))
#	define	w3(zzz)						(((zzz)>>48) & Make64(0, 0xffff))
#	define	l0(zzz)						((zzz) & Make64(0, 0xffffffff))
#	define	l1(zzz)						(((zzz)>>32) & Make64(0, 0xffffffff))
#	define	q0(zzz)						(zzz)

#	define	KeepMaskN(yyy,nnn,sss)		(((yyy)&(nnn)) ? (cuint64(0xff)<<(sss)) : 0)
#	define	KeepMask(yyy)				( KeepMaskN(yyy, 0x01, 0)|KeepMaskN(yyy, 0x02, 8)|KeepMaskN(yyy, 0x04, 16)|\
										  KeepMaskN(yyy, 0x08, 24)|KeepMaskN(yyy, 0x10, 32)|KeepMaskN(yyy, 0x20, 40)|\
										  KeepMaskN(yyy, 0x40, 48)|KeepMaskN(yyy, 0x80, 56) )
#	define	ByteZap(zzz,yyy)			((zzz) & KeepMask((~((yyy)))))
/* #	define	ByteZap(zzz,yyy)			((zzz) & KeepMask((~(cuint64(yyy))))) */
#	define	ByteKeep(zzz,yyy)			((zzz) & KeepMask((yyy)))

// Specific support for S (float) and T (double) type formats
#	define	FloatingPointSignBitMask	GenMaskForBit(63)

#	define	MAP_S(yyy)					(	((((yyy) & 0xff) == 0xff) ? 0xfff : 	\
											((((yyy) & 0xff) == 0x00) ? 0x000 :	\
											((((yyy) & 0x80) == 0x80) ? (0x400|((yyy)&0x7f))	:	\
											(0x380|((yyy)&0x7f)) )))	)

#	define	LoadFromSFormat_(xxx)		(	((((xxx) >> 31)&0x01)<<63)	|	\
											((MAP_S(((xxx)>>23)&0xff)) << 52)			|	\
											(((xxx) & 0x7fffff) << 29)		|	\
											(0)	)

#	define	LoadFromSFormat(xxx)		LoadFromSFormat_(cuint64(xxx))
#	define	StoreToSFormat(xxx)			(	((((xxx) >> 62)&0x03)	<< 30)		|	((xxx >> 29)&0x3fffffff)	)



// support functions for op-code parsing

#	define	OpCodeField(zzz)					(((zzz) >> 26) & 0x3f)
#	define	RegisterField_Ra(zzz)				(((zzz) >> 21) & 0x1f)
#	define	RegisterField_Rb(zzz)				(((zzz) >> 16) & 0x1f)
#	define	RegisterField_Rc(zzz)				(((zzz) >> 0) & 0x1f)
#	define	MemoryDisplacementField(zzz)		SignExtend16To64((((zzz) >> 0) & 0xffff))
#	define	FunctionField(zzz)					(((zzz) >> 0) & 0xffff)
#	define	OpFunctionField(zzz)				(((zzz) >> 5) & 0x7f)
#	define	BranchDisplacementField(zzz)		cint64(4*SignExtend21To64((((zzz) >> 0) & 0x1fffff)))
#	define	LiteralOrNotField(zzz)				(((zzz) >> 12) & 0x01)
#	define	LiteralField(zzz)					cuint64(((zzz) >> 13) & 0xff)
#	define	PALCodeField(zzz)					(((zzz) >> 0) & 0x3ffffff)


#endif
