#ifndef INSTRUCTION_INCLUDED
#define INSTRUCTION_INCLUDED


#include"Util/Printable.h"
#include"Util/File.h"
#include"WaveScalarValue.h"
#include<vector>
#include<map>
#include<string>
#include<utility>
#include<assert.h>

#include"WSOpCode.h"

#include<math.h>
using std::pair;
using std::vector;
using std::map;
using std::string;

class Instruction;

typedef pair<Instruction*, uint> Designator;

class Instruction : public Printable {
 public:
     typedef vector<uint> InvocationSet;

     enum InputType {inA = 0, inB = 1 , inC = 2, inD = 3};

     enum OutputType {outA = 0, outB = 1, outC = 2, outD = 3};

#define INST_CLASS_BIT(a) (1 << a)
     enum InstructionClass {
	  generic = 0,
	  load,
	  identityLoad,
	  store,
	  branch
     };

     static Instruction *Create(WSOpCode inst, const char *name =NULL);
     static Instruction *Create(WSOpCode inst, int64 imm, const char *name =NULL);
     static void InitializeISA();
     static map<string, WSOpCode> nameToOpcode;
     static map<WSOpCode, string> opcodeToName;

     static vector<Instruction *> indirectSendTargets;
     static bool printFires;
     static bool alwaysProbe;

     typedef vector<WaveScalarValue>     InputQueue;
     typedef vector<Designator >         Output;
     typedef pair<bool, WaveScalarValue> OutputValue;

     WSOpCode type;

     L
     string name;
     vector<InputQueue > inputs;
     vector<Output> outputs;
     
     vector<InvocationSet> invocationSets;

     int64 immediate;
     WaveScalarValue sunk;
     int key;
     bool silent;

     uint32 instClass;

     Instruction(uint inputs, uint outputs, int64 imm, const char *name="Inst");  

     Designator operator[] (uint i) {
	  return Designator(this, i);
     }

     Designator I(uint i) {
	  return Designator(this, i);
     }

     Designator O(uint i) {
	  return Designator(this, i);
     }

     void SetInvocationSet0() {
     }
     void SetInvocationSet0(uint i1) {
	  invocationSets[0].push_back(i1);
     }
     void SetInvocationSet0(uint i1, uint i2) {
	  invocationSets[0].push_back(i1);
	  invocationSets[0].push_back(i2);
     }
     void SetInvocationSet0(uint i1, uint i2, uint i3) {
	  invocationSets[0].push_back(i1);
	  invocationSets[0].push_back(i2);
	  invocationSets[0].push_back(i3);
     }
     void SetInvocationSet0(uint i1, uint i2, uint i3, uint i4) {
	  invocationSets[0].push_back(i1);
	  invocationSets[0].push_back(i2);
	  invocationSets[0].push_back(i3);
	  invocationSets[0].push_back(i4);
     }


     void SetInvocationSet1() {
     }
     void SetInvocationSet1(uint i1) {
	  invocationSets[1].push_back(i1);
     }
     void SetInvocationSet1(uint i1, uint i2) {
	  invocationSets[1].push_back(i1);
	  invocationSets[1].push_back(i2);
     }
     void SetInvocationSet1(uint i1, uint i2, uint i3) {
	  invocationSets[1].push_back(i1);
	  invocationSets[1].push_back(i2);
	  invocationSets[1].push_back(i3);
     }
     void SetInvocationSet1(uint i1, uint i2, uint i3, uint i4) {
	  invocationSets[1].push_back(i1);
	  invocationSets[1].push_back(i2);
	  invocationSets[1].push_back(i3);
	  invocationSets[1].push_back(i4);
     }

     void SetInvocationSet2() {
     }
     void SetInvocationSet2(uint i1) {
	  invocationSets[2].push_back(i1);
     }
     void SetInvocationSet2(uint i1, uint i2) {
	  invocationSets[2].push_back(i1);
	  invocationSets[2].push_back(i2);
     }
     void SetInvocationSet2(uint i1, uint i2, uint i3) {
	  invocationSets[2].push_back(i1);
	  invocationSets[2].push_back(i2);
	  invocationSets[2].push_back(i3);
     }
     void SetInvocationSet2(uint i1, uint i2, uint i3, uint i4) {
	  invocationSets[2].push_back(i1);
	  invocationSets[2].push_back(i2);
	  invocationSets[2].push_back(i3);
	  invocationSets[2].push_back(i4);
     }

     void SendValueOnOutput(uint output, const WaveScalarValue &v) {
	  uint k;
	  for (k = 0; k < outputs[output].size(); k++) {
	       outputs[output][k].first->Input(outputs[output][k].second, v);
	  }
     }

     const char *ToString(string &buf) const {
	  buf = name;
	  return buf.c_str();
     }

     virtual void Input(uint input, const WaveScalarValue &v) {
	  uint i;
	  InputQueue::iterator j;
	  bool foundAll = true;
	  inputs[input].push_back(v);
	  vector<InputQueue::iterator> matches;

	  for (i = 0; i < inputs.size(); i++) {
	       bool foundMatch = false;
	       for(j = inputs[i].begin(); j != inputs[i].end(); j++) {
		    if (j->Match(v)) {
			 foundMatch = true;
			 matches.push_back(j);
			 break;
		    }
	       }
	       if (!foundMatch) {
		    foundAll = false;
		    break;
	       }
	  }
	  
	  if (foundAll) {
	       vector<OutputValue> outputValues(outputs.size());
	       vector<WaveScalarValue> inputValues(inputs.size());
	       for(i = 0; i < inputs.size(); i++) {
		    inputValues[i] = *matches[i];
		    inputs[i].erase(matches[i]);
	       }
	       if (printFires && 
		   !silent) {
		    File::Stdout.Print("%s fires\n", name.c_str());
	       }
	       Fire0(inputValues, outputValues);
	       assert(outputValues.size() == outputs.size());
	       for(i = 0; i < outputs.size(); i++) {
		    if (outputValues[i].first) {
			 SendValueOnOutput(i, outputValues[i].second);
		    }
	       }
	  }
     }

     virtual void Fire(const WaveScalarValue &v) {
	  assert(0);
     }

     virtual void Fire0(const vector<WaveScalarValue> inputs, vector<OutputValue> &outputs) = 0;
     virtual void Fire1(const vector<WaveScalarValue> inputs, vector<OutputValue> &outputs) = 0;
     virtual void Fire2(const vector<WaveScalarValue> inputs, vector<OutputValue> &outputs) = 0;

     virtual ~Instruction() {};
     
     void DumpQueues(File &f) {
	  uint i;
	  uint j;
	  f.Print("%s:\n", name.c_str());
	  for(i = 0; i < inputs.size(); i++) {
	       f.Print("In %d\n=========\n", i);
	       for(j = 0; j < inputs[i].size(); j++) {
		    string b;
		    f.Print("%3d : %s\n",j, inputs[i][j].ToString(b));
	       }
	       f.Print("\n");
	  }
     }
     
     static void RegisterIndirectSendTarget(Instruction *inst) {
	  inst->key = indirectSendTargets.size();
	  indirectSendTargets.push_back(inst);
     }
	  
};


void operator>>(Designator src, Designator dst);

void FromProbe(Designator from, Designator  to);
void From(Designator from, Designator  to, bool recursive = false);


#endif
