#include "Instruction.h"
#include"WaveScalarValue.h"
#include"AlphaParsing.h"
#include"WSOpCode.h"


class AlphaInstruction : public Instruction {
 public:
     AlphaInstruction(uint inputs, uint outputs, int64 imm, const char *name="AlphaInst") : Instruction(inputs, outputs, imm, name) {}
};

class AlphaStoreInstruction : public AlphaInstruction {
 protected:
     void Write8(uint64 addr, uint64 data) {
	  File::Stdout.Print("Write8 0x%llx to 0x%llx\n", data, addr);
     }
     void Write16(uint64 addr, uint64 data) {
	  File::Stdout.Print("Write16 0x%llx to 0x%llx\n", data, addr);
     }
     void Write32(uint64 addr, uint64 data) {
	  File::Stdout.Print("Write32 0x%llx to 0x%llx\n", data, addr);
     }
     void Write64(uint64 addr, uint64 data) {
	  File::Stdout.Print("Write64 0x%llx to 0x%llx\n", data, addr);
     }
 public:
     AlphaStoreInstruction(uint inputs, uint outputs, int64 imm, const char *name="AlphaStore") : AlphaInstruction(inputs, outputs, imm, name) 
     {
	  instClass |= INST_CLASS_BIT(Instruction::store);
     }
};

class AlphaLoadInstruction : public AlphaInstruction {

 protected:
     uint64 Read8(uint64 addr) {
	  File::Stdout.Print("Read8 at 0x%llx\n", addr);
	  return 0;
     }
     uint64 Read16(uint64 addr) {
	  File::Stdout.Print("Read16 at 0x%llx\n", addr);
	  return 0;
     }
     uint64 Read32(uint64 addr) {
	  File::Stdout.Print("Read32 at 0x%llx\n", addr);
	  return 0;
     }
     uint64 Read32as64(uint64 addr) {
	  File::Stdout.Print("Read32as64 at 0x%llx\n", addr);
	  return 0;
     }
     uint64 Read64(uint64 addr) {
	  File::Stdout.Print("Read64 at 0x%llx\n", addr);
	  return 0;
     }
 public:
     AlphaLoadInstruction(uint inputs, uint outputs, int64 imm, const char *name="AlphaLoad") : AlphaInstruction(inputs, outputs, imm, name) 
     {
	  instClass |= INST_CLASS_BIT(Instruction::load);
     }

};

class AlphaLoadIdentInstruction : public AlphaLoadInstruction {
public:
     AlphaLoadIdentInstruction(uint inputs, uint outputs, int64 imm, const char *name="AlphaLoad") : AlphaLoadInstruction(inputs, outputs, imm, name) 
     {
	  instClass |= INST_CLASS_BIT(Instruction::load);
	  instClass |= INST_CLASS_BIT(Instruction::identityLoad);
     }
};

class AlphaBranchInstruction : public AlphaInstruction {
 protected:
     bool SetNextProgramCounter(uint64 disp) { return true;}
     uint64 GetCurrentProgramCounter() {return 0;};
 public:
     AlphaBranchInstruction(uint inputs, uint outputs, int64 imm, const char *name="AlphaBranch") : AlphaInstruction(inputs, outputs, imm, name)
     {
	  instClass |= INST_CLASS_BIT(Instruction::branch);
     }

};

class AlphaPALInstruction : public AlphaInstruction {
 public:
     AlphaPALInstruction(uint inputs, uint outputs, int64 imm, const char *name="AlphaBranch") : AlphaInstruction(inputs, outputs, imm, name) {}
};

class _src: public Instruction {
 public:
     enum {OUT = 0};
     _src(int64 imm, const char *name="src") :Instruction(0, 1, 0, name){}
     void Fire(const WaveScalarValue &v) {
	  string b;
	  File::Stdout.Print("%s Sends %s\n", name.c_str(), v.ToString(b)); \
	  SendValueOnOutput(0, v);
     }
     void Fire0(const vector<WaveScalarValue> inputs, vector<OutputValue> &outputs) {
	  assert(0);
     }
     void Fire1(const vector<WaveScalarValue> inputs, vector<OutputValue> &outputs) {
	  assert(0);
     }
     void Fire2(const vector<WaveScalarValue> inputs, vector<OutputValue> &outputs) {
	  assert(0);
     }
};

#define TOKEN_INPUT 0
#define WILDCARD_INPUT 1
class _tc: public Instruction {
public:
     enum {TOKEN=0, WILD_CARD=1};
     enum {OUT=0};

     _tc(int64 imm, const char *name="tc") : Instruction(2, 1, 0, name) {
	  SetInvocationSet0 (TOKEN, WILD_CARD);	
	  SetInvocationSet1 ();
	  SetInvocationSet2 ();
     }
     
     void Input(uint input, const WaveScalarValue &v) {

	  InputQueue::iterator r;
	  InputQueue::iterator t;

	  inputs[input].push_back(v);
	  vector<InputQueue::iterator> matches;
	  bool foundMatch = false;

	  for (r = inputs[WILDCARD_INPUT].begin(); 
	       r != inputs[WILDCARD_INPUT].end(); 
	       r++) {
	       for (t = inputs[TOKEN_INPUT].begin(); 
		    t != inputs[TOKEN_INPUT].end(); 
		    t++) {
		    if (r->value == (int64)t->thread) {
			 foundMatch = true;
			 break;
		    } else {
			 File::Stderr.Print("r->value = %lld t->thread = %llu\n", 
					    r->value, 
					    t->thread);
		    }
	       }
	       if(foundMatch) {
		    break;
	       }
	  }
	  
	  if (foundMatch) {
	       WaveScalarValue token = *t;
	       WaveScalarValue request = *r;

	       inputs[WILDCARD_INPUT].erase(r);
	       inputs[TOKEN_INPUT].erase(t);
	       
	       if (printFires && 
		   !silent) {
		    File::Stdout.Print("%s fires\n", name.c_str());
	       }
	       SendValueOnOutput(0, WaveScalarValue(request.thread, request.wave, token.wave));
	  }
     }
     void Fire0(const vector<WaveScalarValue> inputs, vector<OutputValue> &outputs) {assert(0);}
     void Fire1(const vector<WaveScalarValue> inputs, vector<OutputValue> &outputs) {assert(0);}
     void Fire2(const vector<WaveScalarValue> inputs, vector<OutputValue> &outputs) {assert(0);}
};



// This is an inst without the INPUTS and OUTPUTS fields, CPP garbles them if
// they go through to many levels of macro expansion.
#define INST_BODY(OPCODE, BASE_TYPE, IN_COUNT, OUT_COUNT, IMM, IS_MEM, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
_##OPCODE(const char *name = #OPCODE) :BASE_TYPE(IN_COUNT,OUT_COUNT,0,name){ \
     SetInvocationSet0 FIRE_0_SET;					\
     SetInvocationSet1 FIRE_1_SET;					\
     SetInvocationSet2 FIRE_2_SET;					\
     type = OP_ENUM(OPCODE);						\
     }									\
	  								\
_##OPCODE(int64 imm, const char *name = #OPCODE) :BASE_TYPE(IN_COUNT,OUT_COUNT,imm,name){ \
     SetInvocationSet0 FIRE_0_SET;				\
     SetInvocationSet1 FIRE_1_SET;				\
     SetInvocationSet2 FIRE_2_SET;				\
     type = OP_ENUM(OPCODE);					\
}								\
								\
void Fire0(const vector<WaveScalarValue> inputs, vector<OutputValue> &outputs) { \
     FIRE_0_CODE;						\
}								\
void Fire1(const vector<WaveScalarValue> inputs, vector<OutputValue> &outputs) { \
     FIRE_1_CODE;						\
}								\
void Fire2(const vector<WaveScalarValue> inputs, vector<OutputValue> &outputs) { \
     FIRE_2_CODE;						\
}


///////////////////////////////////////////////////////////////////////////////////
/////////////////  INTEGER
///////////////////////////////////////////////////////////////////////////////////                    
#define ALPHA_IARITH_INST(OP, IN_COUNT, OUT_COUNT, IMM, IS_MEM, INPUTS, AUTO_INPUTS, OUTPUTS, AUTO_OUTPUTS, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     class _##OP : public AlphaInstruction {				\
     public:								\
	  INPUTS;							\
	  OUTPUTS;							\
	  AUTO_INPUTS;							\
	  AUTO_OUTPUTS;							\
	  INST_BODY(OP, AlphaInstruction, IN_COUNT, OUT_COUNT, IMM, IS_MEM, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
	       };							
     
#define ALPHA_IARITH_INST_IMM(OP, IN_COUNT, OUT_COUNT, IMM, IS_MEM, INPUTS, AUTO_INPUTS, OUTPUTS, AUTO_OUTPUTS, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     class _##OP : public AlphaInstruction {				\
     public:								\
	  INPUTS;							\
	  OUTPUTS;							\
	  AUTO_INPUTS;							\
	  AUTO_OUTPUTS;							\
	  INST_BODY(OP, AlphaInstruction, IN_COUNT, OUT_COUNT, IMM, IS_MEM, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     };							
     
#define ALPHA_ICMOVE_INST(OP, IN_COUNT, OUT_COUNT, IMM, IS_MEM, INPUTS, AUTO_INPUTS, OUTPUTS, AUTO_OUTPUTS, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     class _##OP : public Instruction {					\
     public:								\
	  INPUTS;							\
	  OUTPUTS;							\
	  AUTO_INPUTS;							\
	  AUTO_OUTPUTS;							\
	  INST_BODY(OP, AlphaInstruction, IN_COUNT, OUT_COUNT, IMM, IS_MEM, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     };							

#define ALPHA_ICMOVE_INST_IMM(OP, IN_COUNT, OUT_COUNT, IMM, IS_MEM, INPUTS, AUTO_INPUTS, OUTPUTS, AUTO_OUTPUTS, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     class _##OP : public AlphaInstruction {				\
     public:								\
	  INPUTS;							\
	  OUTPUTS;							\
	  AUTO_INPUTS;							\
	  AUTO_OUTPUTS;							\
	  INST_BODY(OP, AlphaInstruction, IN_COUNT, OUT_COUNT, IMM, IS_MEM, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     };							
     
#define ALPHA_ILOAD_INST(OP, IN_COUNT, OUT_COUNT, IMM, IS_MEM, INPUTS, AUTO_INPUTS, OUTPUTS, AUTO_OUTPUTS, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     class _##OP : public AlphaLoadInstruction {				\
     public:								\
	  INPUTS;							\
	  OUTPUTS;							\
	  AUTO_INPUTS;							\
	  AUTO_OUTPUTS;							\
	  INST_BODY(OP, AlphaLoadInstruction,IN_COUNT, OUT_COUNT, IMM, IS_MEM, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     };							
     
#define ALPHA_ISTORE_INST(OP, IN_COUNT, OUT_COUNT, IMM, IS_MEM, INPUTS, AUTO_INPUTS, OUTPUTS, AUTO_OUTPUTS, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     class _##OP : public AlphaStoreInstruction {				\
     public:								\
	  INPUTS;							\
	  OUTPUTS;							\
	  AUTO_INPUTS;							\
	  AUTO_OUTPUTS;							\
	  INST_BODY(OP, AlphaStoreInstruction,IN_COUNT, OUT_COUNT, IMM, IS_MEM, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     };							
     
#define ALPHA_IBRANCH_INST(OP, IN_COUNT, OUT_COUNT, IMM, IS_MEM, INPUTS, AUTO_INPUTS, OUTPUTS, AUTO_OUTPUTS, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     class _##OP : public AlphaBranchInstruction {				\
     public:								\
	  INPUTS;							\
	  OUTPUTS;							\
	  AUTO_INPUTS;							\
	  AUTO_OUTPUTS;							\
	  INST_BODY(OP, AlphaBranchInstruction,IN_COUNT, OUT_COUNT, IMM, IS_MEM, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     };							
          

///////////////////////////////////////////////////////////////////////////////////
/////////////////  STUPID CRAP HACK
///////////////////////////////////////////////////////////////////////////////////                    

#define ALPHA_ILOAD_IDENT_INST(OP, IN_COUNT, OUT_COUNT, IMM, IS_MEM, INPUTS, AUTO_INPUTS, OUTPUTS, AUTO_OUTPUTS, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     class _##OP : public AlphaIdentLoadInstruction {				\
     public:								\
	  INPUTS;							\
	  OUTPUTS;							\
	  AUTO_INPUTS;							\
	  AUTO_OUTPUTS;							\
	  INST_BODY(OP, AlphaLoadInstruction,IN_COUNT, OUT_COUNT, IMM, IS_MEM, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     };							

///////////////////////////////////////////////////////////////////////////////////
/////////////////  FLOATING POINT 
///////////////////////////////////////////////////////////////////////////////////                    

#define ALPHA_FARITH_INST(OP, IN_COUNT, OUT_COUNT, IMM, IS_MEM, INPUTS, AUTO_INPUTS, OUTPUTS, AUTO_OUTPUTS, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     class _##OP : public AlphaInstruction {				\
     public:								\
	  INPUTS;							\
	  OUTPUTS;							\
	  AUTO_INPUTS;							\
	  AUTO_OUTPUTS;							\
	  INST_BODY(OP, AlphaInstruction,IN_COUNT, OUT_COUNT, IMM, IS_MEM, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     };							
     
#define ALPHA_FARITH_INST_IMM(OP, IN_COUNT, OUT_COUNT, IMM, IS_MEM, INPUTS, AUTO_INPUTS, OUTPUTS, AUTO_OUTPUTS, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     class _##OP : public AlphaInstruction {				\
     public:								\
	  INPUTS;							\
	  OUTPUTS;							\
	  AUTO_INPUTS;							\
	  AUTO_OUTPUTS;							\
	  INST_BODY(OP, AlphaInstruction,IN_COUNT, OUT_COUNT, IMM, IS_MEM, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     };							
     
#define ALPHA_FCMOVE_INST(OP, IN_COUNT, OUT_COUNT, IMM, IS_MEM, INPUTS, AUTO_INPUTS, OUTPUTS, AUTO_OUTPUTS, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     class _##OP : public AlphaInstruction {				\
     public:								\
	  INPUTS;							\
	  OUTPUTS;							\
	  AUTO_INPUTS;							\
	  AUTO_OUTPUTS;							\
	  INST_BODY(OP, AlphaInstruction,IN_COUNT, OUT_COUNT, IMM, IS_MEM, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     };						       
     
#define ALPHA_FCMOVE_INST_IMM(OP, IN_COUNT, OUT_COUNT, IMM, IS_MEM, INPUTS, AUTO_INPUTS, OUTPUTS, AUTO_OUTPUTS, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     class _##OP : public AlphaInstruction {				\
     public:								\
	  INPUTS;							\
	  OUTPUTS;							\
	  AUTO_INPUTS;							\
	  AUTO_OUTPUTS;							\
	  INST_BODY(OP, AlphaInstruction,IN_COUNT, OUT_COUNT, IMM, IS_MEM, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     };						       

#define ALPHA_FCONV_INST(OP, IN_COUNT, OUT_COUNT, IMM, IS_MEM, INPUTS, AUTO_INPUTS, OUTPUTS, AUTO_OUTPUTS, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     class _##OP : public AlphaInstruction {				\
     public:								\
	  INPUTS;							\
	  OUTPUTS;							\
	  AUTO_INPUTS;							\
	  AUTO_OUTPUTS;							\
	  INST_BODY(OP, AlphaInstruction,IN_COUNT, OUT_COUNT, IMM, IS_MEM, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     };							
     
#define ALPHA_FLOAD_INST(OP, IN_COUNT, OUT_COUNT, IMM, IS_MEM, INPUTS, AUTO_INPUTS, OUTPUTS, AUTO_OUTPUTS, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     class _##OP : public AlphaLoadInstruction {				\
     public:								\
	  INPUTS;							\
	  OUTPUTS;							\
	  AUTO_INPUTS;							\
	  AUTO_OUTPUTS;							\
	  INST_BODY(OP, AlphaLoadInstruction,IN_COUNT, OUT_COUNT, IMM, IS_MEM, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     };							
     
#define ALPHA_FSTORE_INST(OP, IN_COUNT, OUT_COUNT, IMM, IS_MEM, INPUTS, AUTO_INPUTS, OUTPUTS, AUTO_OUTPUTS, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     class _##OP : public AlphaStoreInstruction {				\
     public:								\
	  INPUTS;							\
	  OUTPUTS;							\
	  AUTO_INPUTS;							\
	  AUTO_OUTPUTS;							\
	  INST_BODY(OP, AlphaStoreInstruction,IN_COUNT, OUT_COUNT, IMM, IS_MEM, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     };							

#define ALPHA_FBRANCH_INST(OP, IN_COUNT, OUT_COUNT, IMM, IS_MEM, INPUTS, AUTO_INPUTS, OUTPUTS, AUTO_OUTPUTS, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     class _##OP : public AlphaBranchInstruction {				\
     public:								\
	  INPUTS;							\
	  OUTPUTS;							\
	  AUTO_INPUTS;							\
	  AUTO_OUTPUTS;							\
	  INST_BODY(OP, AlphaBranchInstruction,IN_COUNT, OUT_COUNT, IMM, IS_MEM, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     };							


///////////////////////////////////////////////////////////////////////////////////
/////////////////  SYSCALLS
///////////////////////////////////////////////////////////////////////////////////                    
#define ALPHA_PAL_INST(OP, IN_COUNT, OUT_COUNT, IMM, IS_MEM, INPUTS, AUTO_INPUTS, OUTPUTS, AUTO_OUTPUTS, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     class _##OP : public AlphaPALInstruction {				\
     public:								\
	  INPUTS;							\
	  OUTPUTS;							\
	  AUTO_INPUTS;							\
	  AUTO_OUTPUTS;							\
	  INST_BODY(OP, AlphaPALInstruction,IN_COUNT, OUT_COUNT, IMM, IS_MEM, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     };							

                    

///////////////////////////////////////////////////////////////////////////////////
/////////////////  DEFINITION FOR CUSTOM INSTRUCTIONS
///////////////////////////////////////////////////////////////////////////////////                    
#define INST(OP, IN_COUNT, OUT_COUNT, IMM, IS_MEM, INPUTS, AUTO_INPUTS, OUTPUTS, AUTO_OUTPUTS, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     class _##OP : public Instruction {					\
     public:								\
	  INPUTS;							\
	  OUTPUTS;							\
	  AUTO_INPUTS;							\
	  AUTO_OUTPUTS;							\
	  INST_BODY(OP, Instruction,IN_COUNT, OUT_COUNT, IMM, IS_MEM, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     };							

#define LINK(a,b)	typedef _##a _##b;


#undef CUSTUM_INST
#define CUSTOM_INST(OP, IN_COUNT, OUT_COUNT, IMM, IS_MEM, INPUTS, AUTO_INPUTS, OUTPUTS, AUTO_OUTPUTS) 
#define IN_VAL(a) inputs[a].downcast<const WaveCacheDataValue>()
#define OUT_VAL(a) outputs[a].downcast<const WaveCacheDataValue>()
#define ZERO_IN ;
#define ZERO_OUT ;
#define ONE_IN(a)  enum{a=0}
#define ONE_OUT(a) enum{a=0}
#define TWO_IN(a, b)  enum{a=0, b=1}
#define TWO_OUT(a, b) enum{a=0, b=1}
#define THREE_IN(a, b, c)  enum{a=0, b=1, c=2}
#define THREE_OUT(a, b, c) enum{a=0, b=1, c=2}
#define FOUR_IN(a, b, c, d)  enum{a=0, b=1,c=2, d=3}
#define FOUR_OUT(a, b, c, d) enum{a=0, b=1,c=2, d=3}

#define WriteIntegerRegister_Rc(a) rc = (a)
#define WriteIntegerRegister_Ra(a) ra = (a)

#define WriteFloatingPointRegister_Rc(a) rc = (a)
#define WriteFloatingPointRegister_Ra(a) ra = (a)

#include"Instructions.def"

#undef LINK
#undef INST
#undef CUSTOM_INST
#undef INPUT
#undef OUTPUT
#undef ONE_IN
#undef ONE_OUT
#undef TWO_IN
#undef TWO_OUT
#undef THREE_IN
#undef THREE_OUT
#undef FOUR_IN
#undef FOUR_OUT

#include"uniform_INST.def"
#define LINK(a, OP)							\
     case OP_ENUM(OP): {						\
     return new ::_##OP(imm, name == NULL ? #OP : name);			\
}									\
	  break;							

#define INST(OP, IN_COUNT, OUT_COUNT, IMM, IS_MEM, INPUTS, AUTO_INPUTS, OUTPUTS, AUTO_OUTPUTS, FIRE_0_SET, FIRE_0_CODE, FIRE_1_SET, FIRE_1_CODE, FIRE_2_SET, FIRE_2_CODE) \
     case OP_ENUM(OP): {						\
	  return new ::_##OP(imm, name == NULL ? #OP : name);		\
     }									\
	  break;							

Instruction *Instruction::Create(WSOpCode inst, const char *name)
{
     return Create(inst, 0, name);
}

Instruction *Instruction::Create(WSOpCode inst, int64 imm, const char *name)
{
     switch (inst) {
#include"Instructions.def"
     default:
	  assert(0);
	  return NULL;
     }
}
