#include"Util/CommandLineParser.h"
#include"Util/SimOutput.h"
#include <string.h>

//#include"regex.hpp"

#include <sys/types.h>
#include <regex.h>

//using boost::regex;

bool CommandLineParser::Option::ParseList(string s, OptionType type, void *l)
{
     string::size_type loc = 0;
     bool done = false;
     bool success = true;
     while (!done) {
	  string::size_type newLoc;
	  newLoc = s.find(",", loc);
	  if (newLoc == string::npos) {
	       newLoc = s.size();
	       done = true;
	  }
	  
	  string item(s, loc, newLoc - loc);
	       
#define PARSE_TYPE(TYPE, type, parse, eat, format, conv) \
     case VECTOR(TYPE): \
	  {\
	       type c; \
	       success = success && ParseItem(item, TYPE, &c);\
	       vector<type> *theList = (vector<type> *)l; \
	       theList->push_back(c); \
          }\
	  break;

	  switch(type) {
	       PARSABLE_TYPES
	  default:
	       SimOutput::Crash("Illegal Type");
	  }
#undef PARSE_TYPE
	       
	  loc = newLoc + 1;
     }
     return success;
}

bool  CommandLineParser::Option::ParseItem(string s, OptionType type, void *storage)
{

#define PARSE_TYPE(TYPE, type, parse, eat, format, conv) \
      case TYPE: \
      {\
	   type *c =(type *)storage; \
	   *c = (type)parse; \
      } \
      break;

     switch (type) {
	  PARSABLE_TYPES
	       default:
	  SimOutput::Crash("Illegal Type");
     }

     return true;

#undef PARSE_TYPE

}

bool CommandLineParser::Option::Parse(CommandLine *cmdLine, int index) {
     bool ret = false;
     switch (type) {

#define PARSE_TYPE(TYPE, type, parse, eat, format, conv) \
     case TYPE:\
	  {\
	       ret = ParseItem((*cmdLine)[index + 1], TYPE, storage);\
	  }\
	  break;

	  PARSABLE_TYPES;
#undef PARSE_TYPE

#define PARSE_TYPE(TYPE, type, parse, eat, format, conv) \
     case VECTOR(TYPE):\
	  {\
	       ret = ParseList((*cmdLine)[index + 1], VECTOR(TYPE), storage);\
	  }\
	  break;
	  PARSABLE_TYPES;
#undef PARSE_TYPE
     default:
	  SimOutput::Crash("Illegal Type");
     }
     

     return ret;
}

bool CommandLineParser::Option::Parse(string value) {
     bool ret = false;
     switch (type) {

#define PARSE_TYPE(TYPE, type, parse, eat, format, conv) \
     case TYPE:\
	  {\
	       ret = ParseItem(value, TYPE, storage);\
	  }\
	  break;

	  PARSABLE_TYPES;
#undef PARSE_TYPE

#define PARSE_TYPE(TYPE, type, parse, eat, format, conv) \
     case VECTOR(TYPE):\
	  {\
	       ret = ParseList(value, VECTOR(TYPE), storage);\
	  }\
	  break;
	  PARSABLE_TYPES;
#undef PARSE_TYPE
     default:
	  SimOutput::Crash("Illegal Type");
     }
     

     return ret;
}

const char* CommandLineParser::Option::ToString(string &s) {
     char buf[1024];
     s.clear();
     switch(type) {
#define PARSE_TYPE(TYPE, type, parse, eat, format, conv)				\
     case TYPE:										\
	  sprintf(buf, "%s " format, GetName().c_str(), conv(*((type *)storage)));		\
	  s += buf;									\
	  break;

	  PARSABLE_TYPES
#undef PARSE_TYPE

#define PARSE_TYPE(TYPE, type, parse, eat, format, conv)					\
      case VECTOR(TYPE):									\
 	  {											\
 	       uint i;										\
 	       bool isFirst = true;								\
 	       vector<type> *t = (vector<type> *)storage;					\
 	       sprintf(buf, "%s ", GetName().c_str()); s+=buf;					\
	       if (t->size() == 0) {								\
		    s += "<empty>";								\
	       }										\
 	       for (i = 0; i < t->size(); i++) {						\
 		    sprintf(buf, "%s" format, isFirst ? "" : "," , conv((*t)[i]));		\
 		    s += buf;									\
 		    isFirst = false;								\
 	       }										\
 	  }											\
 	  break;

	  PARSABLE_TYPES
#undef PARSE_TYPE
     default:
	  SimOutput::Crash("Illegal option type.");
     }
     return s.c_str();
}

void CommandLineParser::CollectAllOptions(vector<Option> &allOptions)
{
     uint i;
     allOptions.insert(allOptions.begin(), options.begin(), options.end());
     
     for(i = 0; i < subParsers.size(); i++) {
	  subParsers[i]->CollectAllOptions(allOptions);
     }
}

string CommandLineParser::MunchPattern(const string &pattern)
{
     return pattern;
}

bool CommandLineParser::Matches(const string & pattern, Option &option)
{

     regex_t r;
     regmatch_t match[1];
     int ret;
     string munchedPattern = MunchPattern(pattern);

     if ((ret = regcomp(&r, munchedPattern.c_str(), 0)) != 0) {
	  char error[128];
	  regerror(ret, &r, error, 128);
	  SimOutput::Crash("Faulty pattern: %d %s\n", ret, error);
     }
     regexec(&r, option.GetName().c_str(), 1, match, 0);
     return (match[0].rm_so == 0 &&
	     match[0].rm_eo == (signed)option.GetName().size());
     
}

bool CommandLineParser::DoMatches(CommandLine *cmdLine, uint index, vector<Option> & allOptions)
{
     bool foundMatch = false;
     uint i;
     for(i = 0; i < allOptions.size(); i++) {
	  if (Matches((*cmdLine)[index], allOptions[i])) {
	       //if ((*cmdLine)[index] == allOptions[i].name) {
	       foundMatch = true;
	       allOptions[i].Parse(cmdLine, index);
	  }
     }
     return foundMatch;
}

bool CommandLineParser::DoMatches(const string& key, const string &value, vector<Option> & allOptions)
{
     bool foundMatch = false;
     uint i;

     for(i = 0; i < allOptions.size(); i++) {
	  if (Matches(key, allOptions[i])) {
	       foundMatch = true;
	       allOptions[i].Parse(value);
	  }
     }
     return foundMatch;
}



bool CommandLineParser::ParseFile(const string & fileName, 
				  vector<Option> & allOptions, 
				  CmdLineParseMode mode)
{
     File optionFile;
     if (optionFile.Open(fileName) != File::SUCCESS) {
	  SimOutput::Crash("Can't open cmdline file %s\n", fileName.c_str());
     }
     string line;
     
     while(optionFile.Read(line) == File::SUCCESS) {
	  string space("\f\n\r\t\v ");
	  uint keyStart, keyEnd, valueStart, valueEnd;
	  keyStart   = line.find_first_not_of(space, 0);
	  keyEnd     = line.find_first_of(space, keyStart);
	  valueStart = line.find_first_not_of(space, keyEnd);
	  valueEnd   = line.find_first_of(space, valueStart);

	  string key(line, keyStart, keyEnd - keyStart);
	  string value(line, valueStart, valueEnd - valueStart);
	  
	  
	  if (!DoMatches(key, value, allOptions)) {
	       if (mode == MODE_STRICT) {
		    return false;
	       }
	  }
     }
     
     return true;
}

bool CommandLineParser::ParseCommandLine(CommandLine *cmdLine, CmdLineParseMode mode)
{
     uint i;
     vector<Option> allOptions;
     CollectAllOptions(allOptions);

     for(i = 0; i < cmdLine->size(); i++) {
	  if ((*cmdLine)[i] == stopOption) {
	       break;
	  } else if ((*cmdLine)[i] == loadFileOption) {
	       if (!ParseFile((*cmdLine)[i], allOptions, mode)) {
		    if (mode == MODE_STRICT) {
			 return false;
		    }
	       }
	  } else {
	       if (!DoMatches(cmdLine, i, allOptions)) {
		    if (mode == MODE_STRICT) {
			 return false;
		    }
	       }
	  }
	  i++;
     }

     return true;
}

void CommandLineParser::GetUsageString(string &buf)
{
     uint i;
     uint parser;
     buf.clear();

     for(i = 0; i != options.size(); i++) {
	  buf += options[i].GetName();
	  buf += ":  ";
	  buf += options[i].description;
	  buf += "\n";

     }     
     
     for(parser = 0; parser < subParsers.size(); parser++) {
	  string s;
	  subParsers[parser]->GetUsageString(s);
	  buf += s;
     }
}


char **CommandLineParser::GetUnparsedArgv(CommandLine *cmdLine) 
{
     char **ret;
     uint i;
     uint count = 0;
     bool foundStop = false;
     ret = new char*[cmdLine->args.size()];
     for(i = 0; i < cmdLine->args.size(); i++) {
	  if (foundStop) {
	       ret[count++] = strdup(cmdLine->args[i].arg.c_str());
	  }
	  
	  if (cmdLine->args[i].arg == stopOption) {
	       foundStop = true;
	  }
     }
     ret[count] = NULL;
     return ret;
}

void CommandLineParser::PrintUsage(File &f)
{

     string t;
     GetUsageString(t);
     f.Print("Usage:  \n%s\n", t.c_str());
}

void CommandLineParser::PrintOptionsAndValues(File &f)
{
     uint i;
     string s;
     uint parser;

     for(i = 0; i != options.size(); i++) {
	  string s;
	  f.Print("%s\n", options[i].ToString(s));
     }
     
     for(parser = 0; parser < subParsers.size(); parser++) {
	  subParsers[parser]->PrintOptionsAndValues(f);
     }
}

void CommandLineParser::AddPrefix(const string &prefix)
{
     string n;
     uint i;
     for(i = 0; i < options.size(); i++) {
	  options[i].fullName = options[i].name;

	  if(options[i].fullName[0] == '-') {
	       options[i].fullName.erase(options[i].fullName.begin());
	  }

	  options[i].fullName.insert(0, prefix);
	  options[i].fullName.insert(0, (string)"-");	  
     }
}
