#ifndef LRU_LIST_INCLUDED
#define LRU_LIST_INCLUDED

#include<map>
using std::map;
#include<list>
using std::list;

template<class LRU_TYPE>
class LRUList {
     map<LRU_TYPE, typename list<LRU_TYPE>::iterator> locationMap;
     list<LRU_TYPE> orderedList;
 public:
     typedef typename list<LRU_TYPE>::iterator iterator;
     
     inline void Use(LRU_TYPE t) {
	  typename map<LRU_TYPE, typename list<LRU_TYPE>::iterator>::iterator mapLoc;
	  mapLoc = locationMap.find(t);

	  if (mapLoc != locationMap.end()) {
	       if (mapLoc->second != orderedList.begin()) {
		    orderedList.erase(mapLoc->second);
		    orderedList.push_front(t);
		    locationMap[t] = orderedList.begin();
	       }
	  } else {
	       orderedList.push_front(t);
	       locationMap[t] = orderedList.begin();
	  }
     }

     inline LRU_TYPE LRUItem() {
	  return orderedList.back();
     }
   
     inline LRU_TYPE MRUItem() {
	  return orderedList.front();
     }

     inline void Remove(LRU_TYPE t) {
	  typename map<LRU_TYPE, typename list<LRU_TYPE>::iterator>::iterator mapLoc;
	  mapLoc = locationMap.find(t);	  
	  if (mapLoc != locationMap.end()) {
	       orderedList.erase(mapLoc->second);
	       locationMap.erase(mapLoc);
	  }
     }

     inline void Clear() {
       orderedList.clear();
       locationMap.clear();
     }
};

#ifdef TEST_LRULIST

#include "File.h"
#include <vector>
using std::vector;

struct Test {
     int num;
     Test(int i = 0) {
	  num = i;
     }
     bool operator< (const Test v) const {return num < v.num;}
};

void MyAssert(bool c)
{
     if (c) {
	  File::Stdout.Print("PASSED.\n");
     } else {
	  File::Stdout.Print("FAILED.\n");
     }
}

int main() 
{
     LRUList<Test> structList;
     LRUList<Test*> ptrList;
     vector<Test> list;
     list.reserve(5);
     uint i;

     for(i = 0; i < 5; i++) {
	  list.push_back(Test(i));
	  structList.Use(list[i]);
	  ptrList.Use(&list[i]);
     }
     
     MyAssert(structList.LRUItem().num == 0);
     MyAssert(structList.MRUItem().num == 4);

     MyAssert(ptrList.LRUItem()->num == 0);
     MyAssert(ptrList.MRUItem()->num == 4);
     
     MyAssert(ptrList.LRUItem() == &list[0]);
     MyAssert(ptrList.MRUItem() == &list[4]);
     
     structList.Use(Test(3));
     MyAssert(structList.LRUItem().num == 0);
     MyAssert(structList.MRUItem().num == 3);

     structList.Use(Test(0));
     MyAssert(structList.LRUItem().num == 1);
     MyAssert(structList.MRUItem().num == 0);

     structList.Remove(Test(3));
     MyAssert(structList.LRUItem().num == 1);
     MyAssert(structList.MRUItem().num == 0);

     structList.Remove(Test(0));
     MyAssert(structList.LRUItem().num == 1);
     MyAssert(structList.MRUItem().num == 4);

     structList.Remove(Test(1));
     MyAssert(structList.LRUItem().num == 2);
     MyAssert(structList.MRUItem().num == 4);

}
#endif

#endif
