#include"Util/Regression.h"
#include"Util/SimOutput.h"
#include"Util/StatelessCallback.h"

#include <signal.h>
Regression::State Regression::success = Regression::UNKNOWN;
vector<StatelessCallback*> Regression::crashCalls;

string Regression::reason = "";

void PrintRegressionStateAndExit(int theSignal)
{
     static bool alreadyCrashing = false;
     char *s;
     if (alreadyCrashing) {
	  return;
     }
     alreadyCrashing = true;
     switch(theSignal) {
     case SIGABRT:
	  s = "SIGABRT";
	  break;
     case SIGINT:
	  s = "SIGINT";
	  break;
     case SIGSEGV:
	  s = "SIGSEGV";
	  break;
     default:
	  s = "UNKNOWN SIGNAL";
	  break;
     }


     uint i;
     for(i = 0;i < Regression::crashCalls.size(); i++) {
	  (*Regression::crashCalls[i])();
     }

     Regression::Fail(s);
     Regression::PrintSuccess();
     exit(1);
}

void Regression::Initialize() {	 
     struct sigaction action;
     action.sa_handler = PrintRegressionStateAndExit;
     sigaction(SIGABRT, &action, NULL);
     sigaction(SIGINT, &action, NULL);
     sigaction(SIGSEGV, &action, NULL);
}

void Regression::Succeed()
{
     switch(success) {
     case UNKNOWN:
	  success = SUCCESS;
	  break;
     case SUCCESS:
     case FAILURE:
	  break;
     }
}

void Regression::Fail(char *aReason, ...)
{
     if(aReason == NULL || aReason == string("")) {
     } else {
	  char buf[1000];
	  va_list ap;
	  va_start(ap, aReason);
	  vsprintf(buf, aReason, ap);
	  va_end(ap);

	  reason += string(buf) + string("\n");
     }
     switch(success) {
     case UNKNOWN:
     case SUCCESS:
	  success = FAILURE;
	  break;
     case FAILURE:
	  break;
     }
}

void Regression::PrintSuccess()
{
     switch(success) {
     case UNKNOWN:
	  SimOutput::Print("Regression UNKNOWN\n");
	  break;
     case SUCCESS:
	  SimOutput::Print("Regression PASSED\n");
	  break;
     case FAILURE:
	  SimOutput::Print("Regression FAILED: %s\n", reason.c_str());
	  break;
     }
}

void Regression::AddCrashCallback(StatelessCallback* callback)
{
     crashCalls.push_back(callback);
}

void Regression::CancelCrashCallback(StatelessCallback* callback)
{
     uint i;
     vector<uint> toDelete;

     for(i = 0; i < crashCalls.size();i++) {
	  if (crashCalls[i] == callback) {
	       toDelete.push_back(i);
	  }
     }

     for (i = 0; i < toDelete.size();i++) {
	  crashCalls.erase(crashCalls.begin() + toDelete[i]);
     }

}
