#ifndef REGRESSION_INCLUDED
#define REGRESSION_INCLUDED

#include <string.h>

#include<vector>
using std::vector;

#include<string>
using std::string;

void PrintRegressionStateAndExit(int theSignal);

class StatelessCallback;

/**
 * This handles regression outcome state.  The idea is that you can call
 * Succeed() 1 or more times and succeed, but if you ever call Failure(), you
 * fail.  In addition, you must call one or the other.  The goal is that, no
 * matter what happens, the program should print a success or failure message.
 * Hence, this catches the signals commonly used to kill processes.
 *
 * You should call Initialize() near the beginning of your simulation.  This
 * will ensure that the regression result gets printed if the simulator crashes
 * or is killed.
 *
 */ 
class Regression {
     static vector<StatelessCallback *> crashCalls;

     friend void PrintRegressionStateAndExit(int theSignal);
public:
     enum State {FAILURE, SUCCESS, UNKNOWN};
     static State success;
     static string reason;

     static void Initialize();
     static void Succeed();
     static void Fail(char *aReason = NULL, ...);
     static void PrintSuccess();
     static void AddCrashCallback(StatelessCallback* callback);
     static void CancelCrashCallback(StatelessCallback* callback);
};

#endif
