#ifndef SIMOUTPUT_DEFINED
#define SIMOUTPUT_DEFINED

#include "File.h"
#include <stdio.h>

/** @file SimOutput put is an attempt at a nicer inteface for doing printf.  It
 *  provides several different ways of printing stuff out.  Some of them, like
 *  Crash(), provide additional semanticas as well.  It also provides a level
 *  indirection between stdio, stderr, and where the output goes. 
 *
 *  FIXME: This class is sort of lame.
 */
class SimOutput {
public:

     static File err;
     static File out;
     static File debug;
     static File trace;

#define Warn(fmt, args...) _Warn(__FILE__, __LINE__, fmt, ## args)
#define Crash(fmt, args...) _Crash(__FILE__, __LINE__, fmt, ## args)

#ifndef SUPRESS_DIAGNOSTIC_OUTPUT

#define Debug(fmt, args...) _Debug(__FILE__, __LINE__, fmt, ## args)
#define Trace(fmt, args...) _Trace(__FILE__, __LINE__, fmt, ## args)

#else

#define Debug(fmt, args...) Noop()
#define Trace(fmt, args...) Noop()

#endif

     static void Print(const char *format, ...);
     static void PrintError(const char *format, ...);
     static void _Debug(const char *file, int line, const char *format = "", ...);
     static void _Warn (const char *file, int line, const char *format = "", ...);
     static void _Crash(const char *file, int line, const char *format = "", ...);
     static void _Trace(const char *file, int line, const char *format = "", ...);
     static inline void Noop() {};

     static void SetErr  (File &f) {err = f;}
     static void SetPrint(File &f) {out = f;}
     static void SetDebug(File &f) {debug = f;}
     static void SetTrace(File &f) {trace = f;}

     static void SetErr  (FILE *f) {err.SetFile(f);}
     static void SetPrint(FILE *f) {out.SetFile(f);}
     static void SetDebug(FILE *f) {debug.SetFile(f);}
     static void SetTrace(FILE *f) {trace.SetFile(f);}
};

#endif
