#ifndef __QLOCK_HEADER
#define __QLOCK_HEADER

#include <memthreads/mtypes.h>
#include <memthreads/spinlock.h>

#define TQLEN MAXTHREADS

/*
 * This simple circular queue has a fixed capacity of TQLEN elements.
 * Further, head is the array index before the first element, and tail
 * is the array index of the last element -- think about this; it's
 * unusual.
 */

typedef struct _thread_queue {
  volatile wsint64 head, tail;
  volatile tid_t queue[TQLEN+1];
} thread_queue_t;

#define TQ_INCR(VAL) (((VAL) < (TQLEN)) ? (VAL)+1 : 0)

#define TQ_EMPTY(TQ) ((TQ)->head == (TQ)->tail)

#define TQ_INITIALIZE(TQ) ((TQ)->head = (TQ)->tail = 0)

#define TQ_ENQUEUE(TQ, VAL) ((TQ)->queue[(TQ)->tail = TQ_INCR((TQ)->tail)] = VAL)
#define TQ_DEQUEUE(TQ) ((TQ)->queue[(TQ)->head = TQ_INCR((TQ)->head)])
#define TQ_PEEK(TQ) ((TQ)->queue[TQ_INCR((TQ)->head)])

typedef struct _qlock {
  volatile wsint64 slock;
  volatile wsint64 isBusy;
  volatile thread_queue_t tqueue;
} qlock_t;

void qlock_initialize(volatile qlock_t *qlock);
void qlock_acquire(volatile qlock_t *qlock);
void qlock_release(volatile qlock_t *qlock);

#endif
