#include<stdio.h>
#include"inst.h"
#include <stdlib.h>
#include <assert.h>
#include <glib.h>
#include <string.h>
#include "basicBlock.h"
#include "list.h"
#include <assert.h>
#include "dfg.h"
#include "buildInst.h"

#include "controlDistribution.h"
#include "ptrSet.h"
#include "lcc.h"

#include "rippleChunk.h"

PBasicBlock basicBlockNew(char *label, 
			  int setId)
{
     PBasicBlock n;
     static int id = 0;
     char buf[1024];

     n = (PBasicBlock)calloc(1, sizeof(BasicBlock));
  
     if (setId == -1) 
	  n->id = id++;
     else {
	  n->id = setId;
	  id = MAX(setId, id) + 1; // ensure uniqueness
     }

/*       g_tree_insert(FUNC_OWNER_PROGRAM(owner)->basicBlocksById,  */
/*  		   (void *)n->id, n); */

     if (label == NULL) {
	  sprintf(buf, "B%d", n->id);
	  n->label = strdup(buf);
     } else {
	  n->label = strdup(label);
     }
     
     n->assignedVariables = setNew(MAX_VARIABLES);
     n->consumedVariables = setNew(MAX_VARIABLES);
     n->entrancePC = -1;

     return n;
}

void basicBlockFree(PBasicBlock block)
{
     GList * l;
     char *label;
     PInst inst;
     PCfgEdge p;
     free(block->label);

     forEach(block->exitLabels, l, label) {
	  if (label) 
	       free(label);
     }

     forEach(block->insts, l, inst) {
	  instFree(inst);
     }

     forEach(block->exits, l, p) {
	  cfgEdgeFree(p);
     }

     forEach(block->entrances, l, p) {
	  cfgEdgeFree(p);
     }

     g_list_free(block->insts);
     g_list_free(block->exits);
     g_list_free(block->entrances);
     g_list_free(block->indirectExits);
     g_list_free(block->potentialFunctionArgs);
}
 

void basicBlockClearVisitedFlags (GList *blocks) 
{
     GList *l;
     GList *i;
     PBasicBlock currentBlock;
     PInst inst;

     forEach(blocks, l, currentBlock) {
	  currentBlock->visited = 0;
	  currentBlock->startedHere = 0;    

	  forEach(currentBlock->insts, i, inst) {
	       PInst inst = (PInst)i->data;
	       inst->startedHere = 0;
	  }
     }
}


GList * basicBlockLastInstLink(PBasicBlock b) 
{
     return g_list_last(b->insts);
}

PInst basicBlockExitInst(PBasicBlock b) 
{
     GList *l;
     PInst inst;

     forEach_r(b->insts, l, inst) {
	  if (instIsCondBranch(inst)) {
	       return inst;
	  }
     }

     return basicBlockLastInst(b);
}

PInst basicBlockLastInst(PBasicBlock b) 
{
     if ((PInst)g_list_last(b->insts)) {
	  return (PInst)g_list_last(b->insts)->data;
     } else {
	  return NULL;
     }
}


char* basicBlockName_DOT(PBasicBlock block, char *buf)
{
     strcpy(buf,block->label);
     dotifyString(buf);
     return buf;
}

void basicBlockPrintAnnotatedMap(PBasicBlock block, FILE * file)
{
     GList *i;
     PInst inst;

     fprintf(file, "\tBlockID: %d\n", block->id);

     forEach(block->insts, i, inst) {
	  instPrintAnnotatedMap(inst, file);
     }     

     fprintf(file, "\tEnd Block\n");
}

void basicBlockPrint_DOT(FILE *f, PBasicBlock block, bool_t includeDfg)
{
     GList *l;
     char blockName[1024], 
	  blockName2[1024], 
	  last[1024], 
	  tail[1024], 
	  dst[1024];
     PInst inst;
     PCfgEdge edge;
     char *color;
#undef SUBGRAPHS
     //#define SUBGRAPHS
     
     
#ifdef SUBGRAPHS
     static int clusterNum= 0;
     fprintf(f, "subgraph cluster%d { label=\"%s\";\ncolor=blue;\nlabel=\"%s\";\n",   
	     clusterNum++,  
	     basicBlockName_DOT(block, blockName),   
	     basicBlockName_DOT(block, blockName2));  
#endif
     if (g_list_length(block->insts) == 0) {
	  color= "red";
     } else if (g_list_length(block->insts) == 1 &&
		((PInst)block->insts->data)->type == OP_mnop) {
	  color = "green";	  
     } else if (block->isLoopExitBlock) {
	  color = "grey";
     } else if (block->isFunctionCallBlock) {
	  color = "blue";
     } else {
	  color = "black";
     }

     fprintf(f, "%s [height=0.05,shape=box,label=%s,color=%s];\n", 
	     basicBlockName_DOT(block, blockName),  
	     basicBlockName_DOT(block, blockName2), color);
     
  
     if (includeDfg) {
	  InstType currentType = OP_TYPE_MAX;
	  bool_t inRank = FALSE;
	  strcpy(last, "");

	  forEach(block->insts, l, inst) {  

	       if (inst->type == currentType) {
	       } else if (inst->type == OP_TYPE_ALPHA_OP) {
		    if (inRank) {
			 fprintf(f, "}");
			 inRank = FALSE;
		    }
	       } else {
		    if (inRank) {
			 fprintf(f, "}");
		    }
		    inRank = TRUE;
		    fprintf(f, "{rank = same; ");
	       }
	       instPrintAsBox_DOT(f, inst, last);     

	       currentType = inst->type;
	  }  
	  if (inRank) {
	       fprintf(f, "}");
	  }
     }

/*      fprintf(f, "%s [height=0.05,shape=box,label=%s];\n",  */
/* 	     basicBlockName_DOT(block, blockName), */
/* 	     basicBlockName_DOT(block, blockName2)); */

#ifdef SUBGRAPHS
     fprintf(f, "}\n");
#endif

     forEach(block->exits, l, edge) {
	  if (edge->dst) {
	       fprintf(f, "%s  -> %s [style=bold,weight=%d];\n", 
		       basicBlockName_DOT(block, tail), 
		       basicBlockName_DOT(edge->dst, dst), 1);
	  }   
     }


     if (includeDfg) {
	  forEach(block->insts, l, inst) { 
	       instPrintEdges_DOT(f, inst, FALSE); 
	  } 
     }
}

void basicBlockPrint(FILE *f, PBasicBlock block, bool_t verbose)
{
     PInst inst;
     PCfgEdge edge;
     GList *l, *l2;
     PArg arg;
     int v;
     PBasicBlock d;

     fprintf(f, "======================\n");
     if (block->isFunctionCallBlock) {
	  fprintf(f, "FUNCTION CALL\n");
     }
     fprintf(f, "b%d: %d %s [", 
	     block->id, 
	     block->localID, 
	     block->label);
/*      fprintf(f, block->jumpInst ? "Calls " : ""); */

/*      forEach(block->indirectExits, l, func) { */
/* 	  fprintf(f, "%s ", EXIT_FUNCTION(func)->name); */
/*      } */
     
     fprintf(f, " with args: ");
     forEach(block->potentialFunctionArgs, l, arg) {
	  fprintf(f, "%d ", arg->reg);
     }
     fprintf(f, "] :\n");
  
     forEach(block->insts, l, inst) {
	  instPrint(f, inst);
     }

     forEach(block->exits, l, edge) {
	  if (edge->dst) {
	       fprintf(f, "    -> %s\n", edge->dst->label);
	  }
     }
  
     if (1 || verbose) {

	  PControlDistributionEdge e;
	  PtrSetIndex i2;
	  PBasicBlock b;

	  if (block->isEpochHead) {
	       fprintf(f, "EPOCH: ");
	       forEachPtrElement(block->epoch, i2, b) {
	           fprintf(f, "%s ", b->label);
	       }
	       fprintf(f, "\n");
	  }

	  return;

	  if (block->exitBlockFor != NULL) {
	       fprintf(f, "LOOP EXIT FOR %d\n", block->exitBlockFor->id);
	  }


	  fprintf(f, "LOOP EXITS: ");
	  if (block->loop) {
	       forEachPtrElement(block->loop->exitBlocks, i2, b) {
		    fprintf(f, "%d ", b->id);
	       }
	  }
	  fprintf(f, "\n");
	 

	  fprintf(f, "LOOP BODY: ");
	  if (block->loop) {
	       forEachPtrElement(block->loop->nodes, i2, b) {
		    fprintf(f, "%d ", b->id);
	       }
	  }
	  fprintf(f, "\n");
	 
	  fprintf(f, "Control Dependent Blocks: \n    " );
	  forEach(block->controlDependentNodes, l2, d) {
	       fprintf(f, "%s ", d->label);
	  }
	  fprintf(f, "\n");

	  fprintf(f, "Strict Control Dependent Blocks: " );
	  if (block->strictControlDependentNodeSet) {
	       forEachElement(block->strictControlDependentNodeSet, v) {
		    fprintf(f, "%d ", 
			    FUNC_BASIC_BLOCK_BY_LOCAL_ID(block->owner, v)->id);
	       }
	  }
	  fprintf(f, "\n");

	  fprintf(f, "Control Dependent On Blocks: \n    " );
	  if (block->controlDependentOnNodes) {
	       forEach(block->controlDependentOnNodes, l2, d) {
		    fprintf(f, "%s ", d->label);
	       }
	       fprintf(f, "\n");
	  }

	  if (block->dominance[DOMINANCE_DIRECTION_PRE].dominatedNodes != NULL) {
	       fprintf(f, "Dominance Closure (pre)\n");
	       forEachPtrElement(block->dominance[DOMINANCE_DIRECTION_PRE].dominatedNodes, i2, b) {
		    fprintf(f, "%s ", b->label);
	       }
	  }
	  fprintf(f, "\n");

	  if (block->dominance[DOMINANCE_DIRECTION_POST].dominatedNodes != NULL) {
	       fprintf(f, "Dominance Closure (post)\n");
	       forEachPtrElement(block->dominance[DOMINANCE_DIRECTION_POST].dominatedNodes, i2, b) {
		    fprintf(f, "%s ", b->label);
	       }
	  }
	  fprintf(f, "\n");
	  
	  fprintf(f, "Predominance frontier: \n   " );
	  if (block->dominance[DOMINANCE_DIRECTION_PRE].dominanceFrontier != NULL) {
	       forEachElement(block->dominance[DOMINANCE_DIRECTION_PRE].dominanceFrontier, v) {
		    fprintf(f, "%s ", 
			    FUNC_BASIC_BLOCK_BY_LOCAL_ID(block->owner, v)->label);
	       }
	  }
	  fprintf(f, "\n");

	  
/* 	  fprintf(f, "Reachable forward: \n   " ); */
/* 	  if (block->reachableNodes[REACHABLE_FORWARD] != NULL) { */
/* 	       forEachPtrElement(block->reachableNodes[REACHABLE_FORWARD], i2, b) { */
/* 		    fprintf(f, "%s ",  */
/* 			    b->label); */
/* 	       } */
/* 	  } */
/* 	  fprintf(f, "\n"); */

/* 	  fprintf(f, "Reachable backward: \n   " ); */
/* 	  if (block->reachableNodes[REACHABLE_BACKWARD] != NULL) { */
/* 	       forEachPtrElement(block->reachableNodes[REACHABLE_BACKWARD], i2, b) { */
/* 		    fprintf(f, "%s ",  */
/* 			    b->label); */
/* 	       } */
/* 	  } */
/* 	  fprintf(f, "\n"); */


	  fprintf(f, "ControlDependentClosure: ");
	  if (block->controlDependentClosureNodeSet) {
	       forEachElement(block->controlDependentClosureNodeSet, v) {
		    fprintf(f, "%d ",  functionFindBasicBlockByLocalID(block->owner, v)->id);
	       }
	       fprintf(f, "\n");
	  }

	  forEach(block->controlExits, l, e) {
	       fprintf(f, "ControlDistribution %s: ", e->type == CONTROL_DISTRIBUTION_CONDITIONAL ? "cond" : "uncond");
	       forEachElement(e->dsts, v) {
		    fprintf(f, "%d ",  functionFindBasicBlockByLocalID(block->owner, v)->id);
	       }
	       fprintf(f, "\n");
	  }
	  
	  
#if (0)	  
	  fprintf(f, "takenControlDependentNodes: ");
	  forEachElement(block->takenControlDependentNodes, v) {
	       fprintf(f, "%d ", v);
	  }
	  fprintf(f, "\n");

	  fprintf(f, "notTakenControlDependentNodes: ");
	  forEachElement(block->notTakenControlDependentNodes, v) {
	       fprintf(f, "%d ", v);
	       
	  }
	  fprintf(f, "\n");

	  
	  fprintf(f, "takenControlDependentOnNodes: ");
	  forEachElement(block->takenControlDependentOnNodes, v) {
	       fprintf(f, "%d ", v);
	  }
	  fprintf(f, "\n");

	  fprintf(f, "notTakenControlDependentOnNodes: ");
	  forEachElement(block->notTakenControlDependentOnNodes, v) {
	       fprintf(f, "%d ", v);
	  }
	  fprintf(f, "\n");


	  fprintf(f, "no local successors \n   " );
	  forEach(block->noLocalSuccessorInsts, l1, inst) {
	       fprintf(f, "%d ", inst->id);
	  }
	  fprintf(f, "\n");

	  fprintf(f, "no local predecessors \n   " );
	  forEach(block->noLocalPredecessorInsts, l1, inst) {
	       fprintf(f, "%d ", inst->id);
	  }
	  fprintf(f, "\n");

	  fprintf(f, "Control source: %d \n", 
		  block->controlSource != NULL ? block->controlSource->id: -1);
#endif
     }
}

void basicBlockAddInst(PBasicBlock block, PInst inst)
{
     // If the program has already been constructed then we add the inst to the
     // big list of insts in the program.  Otherwise it will get added later.
     if (block->owner && block->owner->owner) {
	  programAddInst(block->owner->owner, inst);
     }
}



void basicBlockAppendInst(PBasicBlock block, PInst inst)
{
     GList *link = NULL;
     link = g_list_append(link, inst);
     assert(inst->basicBlockLink == NULL);
     inst->basicBlockLink = link;

     inst->owner = block;
     block->length++;
     block->insts = g_list_append_link(block->insts, link);
     basicBlockAddInst(block, inst);
}

void basicBlockPrependInst(PBasicBlock block, PInst inst)
{
     GList *link = NULL;
     link = g_list_append(link, inst);
     assert(inst->basicBlockLink == NULL);
     inst->basicBlockLink = link;


     inst->owner = block;
     block->length++;
     block->insts = g_list_prepend_link(block->insts, link);
     basicBlockAddInst(block, inst);
}

void basicBlockInsertInstBeforeInst(PBasicBlock block, PInst before, PInst inst)
{
     GList *link = NULL;
     link = g_list_append(link, inst);
     assert(inst->basicBlockLink == NULL);
     inst->basicBlockLink = link;

     inst->owner = block;
     block->length++;
     block->insts = g_list_insert_link_before(block->insts, before, link);
     basicBlockAddInst(block, inst);
}

void basicBlockInsertInstAfterInst(PBasicBlock block, PInst after, PInst inst)
{
     GList *link = NULL;
     link = g_list_append(link, inst);
     assert(inst->basicBlockLink == NULL);
     inst->basicBlockLink = link;

     inst->owner = block;
     block->length++;
     block->insts = g_list_insert_link_after(block->insts, after, link);
     basicBlockAddInst(block, inst);
}

void basicBlockRemoveInst(PBasicBlock block, PInst inst)
{
     GList *l = g_list_find(block->insts,inst);
     assert(l);
     assert(!inst->isTokenizeNop);
     block->insts = g_list_remove_link(block->insts, inst->basicBlockLink);
     g_list_free_1(inst->basicBlockLink);
     inst->basicBlockLink = NULL;
     instFree(inst);
}


void  basicBlockMerge(PBasicBlock top, PBasicBlock bottom)
{
     GList *i, *j;
     PCfgEdge between, enterEdge;
     PInst inst;
     GList *c;
     bool_t first = TRUE; 
     assert(g_list_length(top->exits) == 1 && g_list_length(bottom->entrances) == 1);
     between = (PCfgEdge)top->exits->data;

     assert(!between->isBackEdge);


     c = g_list_copy(top->entrances);
     forEach(c, i, enterEdge) {
	  cfgEdgeMoveDst(enterEdge, bottom);
     }
     g_list_free(c);

     cfgRemoveEdge(between);
     cfgRemoveBlock(top);
     top->isDead =TRUE;

     bottom->insts = g_list_concat(top->insts, bottom->insts);

     forEach(bottom->insts, i, inst) {
	  
	  inst->owner = bottom;
	  if (first) {
	       inst->isFirstInBlock = TRUE;
	       first = FALSE;
	  } else {
	       inst->isFirstInBlock = FALSE;
	  }
     }

     if (top->isLoopExitBlock) {
	  bottom->isLoopExitBlock = TRUE;
     }

     if (top->isLoopTop || bottom->isLoopTop) {
	  bottom->isLoopTop = TRUE;
     }
     
     if (top->entrancePC != -1) {
	  bottom->entrancePC = top->entrancePC;
     }
     
     bottom->extraInstructions  += top->extraInstructions;
}


PCfgEdge basicBlockGetNotTakenExitEdge(PBasicBlock block)
{
     GList *l;
     PCfgEdge e;

     forEach(block->exits, l, e) {
	  if (e->type == CFG_EDGE_NOT_TAKEN) {
	       return e;
	  }
     }
     assert(0);
     return NULL;
}

PCfgEdge basicBlockGetTakenExitEdge(PBasicBlock block)
{
     GList *l;
     PCfgEdge e;

     forEach(block->exits, l, e) {
	  if (e->type == CFG_EDGE_TAKEN) {
	       return e;
	  }

     }
     return NULL;
}

PBasicBlock basicBlockGetNotTakenExit(PBasicBlock block)
{
     return basicBlockGetNotTakenExitEdge(block)->dst;
}

PBasicBlock basicBlockGetTakenExit(PBasicBlock block)
{
     return basicBlockGetTakenExitEdge(block)->dst;
}


void basicBlockSetPC (PBasicBlock block, PC pc) 
{
     block->entrancePC = pc;
}

GList *my_g_list_find(GList *l, void *p) 
{
     while (l) {
	  if (l->data == p)
	       return l;
	  l = l->next;
     } 
     return NULL;
}

void basicBlockFindUsedVariables(PBasicBlock block)
{
     GList *l;
     PInst inst;
     ArgIndex i;
     PArg arg;
     
     forEach(block->insts, l, inst) {
	  forEachSrc(inst, i, arg) {
	       if (arg->reg != -1) {
		    setAdd(block->consumedVariables, arg->reg);
	       }
	  }
	  forEachDst(inst, i, arg) {
	       if (arg->reg != -1) {
		    setAdd(block->assignedVariables, arg->reg);
	       }
	  }
     }
}

void basicBlockSetJumpFlags(PBasicBlock block)
{
     GList *l;
     PInst n;

     forEach(block->insts, l, n) {
	  if (instNeedsJumpPorts(n)) {
	       PC *pc = (PC *)calloc(sizeof(PC), 1);
	       g_tree_insert(BB_OWNER_PROGRAM(block)->functionCallBlocksByPC, pc, block);
	       block->needsJumpPorts = 1;
	       block->jumpInst = n;
	  }
 
	  if (instNeedsReturnPorts(n)) {
	       block->jumpInst = n;
	       block->needsJumpPorts = 1;
	       block->needsReturnPorts = 1;
	  }
     }
}


static bool_t isBackEdgeDependence(PBasicBlock b, PInst src, PInst dst)
{
     GList *l;
     PInst i;
	  
     /* We are going to search through the instructions in the basic block in
	order.  If we hit dst first, then the dependency is between iterations
	of a loop.  If we hit the src first, then it is a dependency within one
	instance of the basic block.*/

     forEach(b->insts, l, i) {
	  if (i == dst) {      
	       return TRUE;
	  } 
	  if (i == src) {
	       return FALSE;
	  } 
     }
     assert(0);
     return FALSE;
}

/** 
 * Find the instructions in the block that are not dependent on anything in
 * this basic block.
 */
void basicBlockComputeNoLocalPredecessorInsts(PBasicBlock block)
{

     GList *l1, *l3;
     PInst inst;
     PArg src;
     PInst srcInst;

     forEach(block->insts, l1, inst) {
	  int instParentFound = FALSE;
	  ArgIndex i;

	  forEachSrc(inst, i, src) {
	       PArg producer;
	       int argParentFound = FALSE;
	       forEach(src->producers, l3, producer) {
		    srcInst = producer->owner;
		    if (srcInst->owner == block && !isBackEdgeDependence(block, srcInst, inst)) {
			 instParentFound = TRUE;
			 argParentFound = TRUE;
		    } 
	       }
	       if (!argParentFound && src->reg != -1 && src->reg != 31) {
		    block->noLocalProducerArgs = g_list_append(block->noLocalProducerArgs, src);
	       }
	  }

	  if (!instParentFound) {
	       block->noLocalPredecessorInsts = g_list_append(block->noLocalPredecessorInsts, inst);
	       inst->hasNoLocalPredecessor = TRUE;
	  }
     }
}

/** 
 * Find the instructions in the block that have no dependent instructions in
 * this basic block.
 */
void basicBlockComputeNoLocalSuccessorInsts(PBasicBlock block) 
{
#if (0) // THis needs to be ported to the new definition of srcs and dsts.     
     PInst inst;
     GList *l1, *l3;
     PInst inst;
     PArg dst;
     int v;
     PInst dstInst;

     forEachLink(block->insts, l1) {
	  int internalChildFound = FALSE;
	  inst = (PInst)l1->data;

	  //	  fprintf(stderr, "Checking %d: \n", inst->id);

	  for (v = OUTPUT_A; v <= OUTPUT_C; v++) {
	       dst = &inst->dsts[v];
/* 	       if (dst->outputType == OUTPUT_CONTROL_TAKEN || */
/* 		   dst->outputType == OUTPUT_CONTROL_NOT_TAKEN) { */
/* 		    assert(0); */
/* 		    continue; */
/* 	       } */

	       forEachLink(dst->consumers, l3) {
		    dstInst = ((PArg)l3->data)->owner;
		    if (dstInst->owner == block && !isBackEdgeDependence(block, inst, dstInst)) {
			 internalChildFound = TRUE;
		    } 
	       }
	  }
	  if (!internalChildFound) {
	       block->noLocalSuccessorInsts = g_list_append(block->noLocalSuccessorInsts, inst);
	  }
     }
#endif
     assert(0);     
}

#if(0)
void basicBlockAddControlDistributionEdge(PBasicBlock src, PBasicBlock dst)
{
     src->controlExits = g_list_append(src->controlExits, dst);
     dst->controlEntrances = g_list_append(dst->controlEntrances, src);
}
#endif

void basicBlockSanity(PBasicBlock block)
{
     GList *l;
     PCfgEdge edge;

     forEach(block->exits, l, edge) {
	  assert(edge->dst);
	  assert(edge->src);
     }
     
}


void basicBlockBuildRippleChunks(PBasicBlock b)
{
     GList *l;
     PInst inst;
     PRippleChunk n = NULL;

     forEach(b->insts, l, inst) {
	  if (instNeedsEpochTags(inst)) {

	       if (!cfgOptions.useRipples ||
		   instIsType(inst, F_STORE) ||
		   inst->type == OP_TYPE_FUNCTION_CALL) {
		    // If we aren't supposed to use ripples, then just make them all
		    // singleton.  Otherwise, stores and function calls are singletons.
		    if (n != NULL) {
			 b->rippleChunks = g_list_append(b->rippleChunks, n);
		    }

		    n = rippleChunkNew(b);     
		    rippleChunkAddInst(n, inst); 

		    b->rippleChunks = g_list_append(b->rippleChunks, n);
		    
		    n = NULL;
	       } else if (inst->isCanonicalEpochAdvance) { 
		    // Epoch advance starts a new ripple.
		    if (n != NULL) {
			 b->rippleChunks = g_list_append(b->rippleChunks, n);
		    }
		    n = rippleChunkNew(b);     
		    rippleChunkAddInst(n, inst); 
	       } else {
		    if (n == NULL) {
			 n = rippleChunkNew(b);     
		    }
		    rippleChunkAddInst(n, inst); 
	       }
	  }
     }
     if (n != NULL) {
	  b->rippleChunks = g_list_append(b->rippleChunks, n);
     }

}
