#ifndef CFG_INCLUDED
#define CFG_INCLUDED

typedef struct _CFG_EDGE *PCfgEdge;

#include <glib.h>
#include "function.h"
#include "basicBlock.h"
#include "host.h"
#include "inst.h"
#include "dominance.h"
#include "ptrSet.h"

#define REACHABLE_BACKWARD 0
#define REACHABLE_FORWARD 1

typedef enum {CFG_EDGE_TAKEN,
	      CFG_EDGE_NOT_TAKEN,
	      CFG_EDGE_NORMAL,
	      CFG_EDGE_MAX} CfgEdgeType;

typedef struct CFG_OPTIONS {
     bool_t includeExitEnterBlocks;
     bool_t verbose;
     bool_t onlyCFG;
     bool_t serialize;
     bool_t performEpochNumbering;
     
     bool_t dynamicInstructionsByEpoch;
     bool_t dumpFinal;
     bool_t supressEpochOptimize;
     bool_t supressCollapsing;

     bool_t doTokenMemory;

     bool_t retchedInfiniteRecursionHack;

     bool_t useRipples;
     //     bool_t includeWaveCoordinate;
     bool_t realFunctionCalls;
     char *compileSpecFile;
     bool_t outputAssembly;
     bool_t compressWaveAdvance;
} CfgOptions, *PCfgOptions;

typedef struct _CFG_EDGE {
     PBasicBlock dst;  
     PSet controlDependentNodes;
     PSet controlDependentClosureNodeSet;
     PBasicBlock src;
     bool_t isBackEdge;
     bool_t isLoopExitEdge;

     CfgEdgeType type;

} CfgEdge;

extern CfgOptions cfgOptions;

void cfgBuild(PFunction func);
void cfgPrint_DOT(FILE *f, PFunction func, bool_t includeDfg);
PCfgEdge cfgAddEdge(PBasicBlock from, PBasicBlock to, CfgEdgeType type) ;
void cfgPrintExport(PFunction func, FILE *f);
void cfgPartitionControlDependences(PFunction func);
void cfgLabelLoops(PFunction func);

PCfgEdge cfgEdgeNew(PBasicBlock owner, PBasicBlock target, CfgEdgeType type);
void cfgEdgeFree(PCfgEdge e);
bool_t cfgStrictlyDominates(PBasicBlock a, PBasicBlock b, DominanceDirection direction);
bool_t cfgDominates(PBasicBlock a, PBasicBlock b, DominanceDirection direction);
void cfgRemoveEdge(PCfgEdge e);
void cfgRemoveBlock(PBasicBlock block);
void cfgEdgeMoveSrc(PCfgEdge e, PBasicBlock newSrc) ;
void cfgComputeRechableSets(PFunction func, int dir, bool_t allowBackEdges);

void cfgAddUnifiedLoopExits(PFunction func);
void cfgReachableTearDown(PFunction func);
void cfgFindLoops(PFunction func);
void cfgAddExitBlocks(PFunction func);
void cfgUpdateLoopBodies(PFunction func);
void cfgLabelBackEdges(PFunction func);

void cfgExpandLoops(PFunction func);

#endif
