#ifndef DFG_INCLUDED
#define DFG_INCLUDED

#include <glib.h>

typedef struct _DFG_STATS *PDfgStats;

#include "function.h"
#include "inst.h"

#define EPOCH_BASIC_BLOCK_SIZE_HISTOGRAM_BUCKETS 30
#define EPOCH_OP_SIZE_HISTOGRAM_BUCKETS 200

typedef struct _DFG_STATS {
     int instCount;
     int branchCount;

     int dataEdgeCount;
     int controlEdgeCount;
     int switchEdgeCount;

     int maxControlOutDegree;
     int maxDataOutDegree;
     int maxSwitchOutDegree;
     int maxTotalOutDegree;
     
     int instructionDistribution[OP_TYPE_MAX];

     int epochBasicBlockSizeHistogram[EPOCH_BASIC_BLOCK_SIZE_HISTOGRAM_BUCKETS];
     int epochInstSizeHistogram[EPOCH_OP_SIZE_HISTOGRAM_BUCKETS];
     
     int waveSizeHistogram[1000];
     
     int zeroLatencyInstructionCount;
     int extraInstructionCount;
     int overheadInstructions;
} DfgStats;

void dfgBuild(PFunction func);
void dfgBuild_OLD(GList * func);
void dfgDismantal(PFunction func);
void dfgAddEdge(PArg from, PArg to);
void dfgAddBranchControlEdges(PFunction func);
void dfgCollectStats(PFunction func, PDfgStats stats);
void dfgSmartAddEdge (PInst src, OutputArgType srcType, PInst dst, InputArgType dstType);
void dfgSmartRemoveEdge(PInst src, OutputArgType srcType, PInst dst, InputArgType dstType);
void dfgRemoveEdge(PArg src, PArg dst);
void dfgRemoveInstEdges(PInst inst);
void dfgUpdateSource(PArg src);
void dfgPrintExport(PFunction func, FILE *f) ;
void dfgPrintStats(FILE *f, PDfgStats stats);

#endif

