#include "epoch.h"
#include <stdio.h>
#include "function.h"
#include "basicBlock.h"
#include "inst.h"
#include "wsArch.h"
#include <assert.h>

void epochTagSPrintRaw(char *s, Epoch prev, Epoch cur, Epoch next, RippleSize rippleSize)
{
     static char b1[1024], b2[1024], b3[1024];
     
     strcpy(b1, EPOCH_ORDER_NAME(prev));
     strcpy(b2, EPOCH_ORDER_NAME(cur));
     strcpy(b3, EPOCH_ORDER_NAME(next));
     
     sprintf(s, "< %s %s %s >.%d", b1, b2, b3, rippleSize);
}


void epochTagSPrint(char *s, PInst inst) {
     epochTagSPrintRaw(s, 
		       inst->epochOrderPrev, 
		       inst->epochOrder, 
		       inst->epochOrderNext,
		       inst->lastStoreSequenceNumber);
}

void epochTagPrintRaw(FILE *f, Epoch prev, Epoch cur, Epoch next, RippleSize rippleSize)
{
     char buf[1024];
     epochTagSPrintRaw(buf, prev, cur, next, rippleSize);
     fprintf(f, buf);
}

void epochTagPrint(FILE *f, PInst inst) 
{
     char buf[1024];
     epochTagSPrint(buf, inst);
     fprintf(f, buf);
}

bool_t epochHoistValues(PFunction func)
{
     GList *i, *j, *l;
     PBasicBlock block;
     PInst inst;
     bool_t hoistedSomething = FALSE;

     forEach(func->basicBlocks, i, block) {
	  forEach(block->insts, l, inst) {
	       if (inst->type == OP_wa &&
		   !inst->isCanonicalEpochAdvance) {
		    PArg output = instGetDstByType(inst, OUTPUT_C);
		    PInst rho = NULL;
		    PInst epochAdvance = NULL;
		    GList *k;
		    PArg consumer;
			 
			bool_t	HaveRho = FALSE;
			PArg	producer2;
			GList	*k2;
    		PArg expendableDataInput2 = instGetSrcByType(inst, INPUT_A);

			forEach(expendableDataInput2->producers,k2,producer2) {
				if(producer2->owner->type == OP_rho)
					HaveRho = TRUE;
				}
		  
			if ( HaveRho )	continue; 
		    if (output->reg == EPOCH_REGISTER ||
			output->reg == EPOCH_COORDINATION_REGISTER) {
			 continue;
		    }

		    forEach(output->consumers, k, consumer) {
			 if (consumer->owner->type == OP_rho) {
			      rho = consumer->owner;
			 }
		    }
			 
		    if (rho != NULL) {
			 PArg takenOutput = instGetDstByType(rho, OUTPUT_A);


			 if (g_list_length(takenOutput->consumers) == 1) {
			      PArg a = (PArg)g_list_nth_data(takenOutput->consumers, 0);
			      if (a != NULL && a->owner == inst) {
				   epochAdvance = inst;
			      }
			 }
		    }
			 
		    if (epochAdvance != NULL) {
			 PArg notTakenOutput = instGetDstByType(rho, OUTPUT_B);
			 PArg valueInput = instGetSrcByType(epochAdvance, INPUT_A);
			 PArg producer,  consumer;
			 GList *c1, *c2;
			      
			 c1 = g_list_copy(valueInput->producers);

			 forEach(c1, j, producer) {
			      if (producer->owner != rho) {

				   if(g_list_length(output->consumers)== 1) {
					hoistedSomething = TRUE;
					dfgRemoveEdge(producer, valueInput);
				   }
				   
				   c2 = g_list_copy(notTakenOutput->consumers);
				   
				   forEach(c2, k, consumer) {
					hoistedSomething = TRUE;
					dfgRemoveEdge(notTakenOutput, consumer);
					dfgAddEdge(producer, consumer);
				   }
				   
				   g_list_free(c2);
			      }
			 }
			 g_list_free(c1);
		    }
	       }
	  }
     }

     return hoistedSomething;
}

bool_t  epochRemoveUseless(PFunction func)
{
     
     GList *i,  *l;
     GList *c;
     PBasicBlock block;
     PInst inst;
     bool_t hoistedSomething = FALSE;

     forEach(func->basicBlocks, i, block) {
	  c = g_list_copy(block->insts);
	  forEach(c, l, inst) {
			
	       if(inst->type == OP_wa) {
		    // check all of my producers and make sure none of them
		    // is a rho
		    PArg output = instGetDstByType(inst, OUTPUT_C);
		    PArg consumer;
		    PInst expendable = NULL, important = NULL;
		    bool_t	HaveRho = FALSE;
		    PArg	producer2;
		    GList	*k2;
		    PArg inputA = instGetSrcByType(inst, INPUT_A);

		    // Mark hack
		    forEach(inputA->producers,k2,producer2) {
			 if(producer2->owner->type == OP_rho) {
			      HaveRho = TRUE;
			 }
		    }
		   
		    // is this an EPOCH -> EPOCH instruction? 
		    if ( ! HaveRho )
			 if (instCountConsumers(inst) == 1) {
			      consumer = (PArg)g_list_nth_data(output->consumers, 0);
			      if (consumer->owner->type == OP_wa ) {
				   expendable = inst;
				   important = consumer->owner;
			      }
			 }
		   
		    // yes, so re-wire ALL inputs directly to the follow on EPOCH 
		    if (important != NULL) {
			 GList  *k;
			 GList *c;
			 PArg producer;

			 PArg expendableDataInput = instGetSrcByType(expendable, INPUT_A);

			 PArg importantDataInput = instGetSrcByType(important, INPUT_A);

			 // acquire all my producing instructions
			 
			 c = g_list_copy(expendableDataInput->producers);
			 forEach(c, k, producer) {

			      // delete edge from producer to me
			      dfgRemoveEdge(producer, expendableDataInput);

			      // add edge from producer to the follow on code
			      // hmm does this really work?
			      dfgAddEdge(producer, importantDataInput);
			 }
			 g_list_free(c);
			
			 dfgRemoveInstEdges(expendable);
			 basicBlockRemoveInst(inst->owner, inst);

			 hoistedSomething = TRUE;
			      
		    }
	       }
	  }
	  g_list_free(c);
     }
     return hoistedSomething;
}

void epochOptimize(PFunction func)
{
     
     //    int c;
     bool_t hoistedSomething = TRUE;

     while(hoistedSomething) {
	  hoistedSomething = FALSE;
	  hoistedSomething = hoistedSomething || epochHoistValues(func);
	  hoistedSomething = hoistedSomething || epochRemoveUseless(func);
     }

#if (0)
     if (block->isEpochHead) {
	  PBasicBlock b;
	  PtrSetIndex m;
	  int reg;
	  PPtrSet defined = ptrSetNew();
	  
	  forEachPtrElement(block->epoch, m, b) {
	       PArg src;
	       ArgIndex n;
	       forEachSrc(inst, n, src) {
		    ptrSetAdd(defined, (void*)src->reg);
	       }
	  }
     }
     
     fprintf(stderr, "Block %d uses : ", block->id);
     forEachPtrElement(defined, m, reg) {
	  fprintf(stderr, "%d ", reg);
     }
     fprintf(stderr, "\n");
     
     block->registersUsed = defined;
#endif
}
	  


void epochOptimizeRemoveCoordinatorsInEmptyBlocks(PFunction func)
{
     GList *i, *j, *k;
     PBasicBlock block;
     PInst inst;

     forEach(func->basicBlocks, i, block) {
	  
	  if (g_list_length(block->insts) == 1) {
	       inst = (PInst)block->insts->data;
	  } else {
	       continue;
	  }
	  
/* 	  if (inst->type == OP_TYPE_EPOCH_COORDINATOR) { */
/* 	       PArg epochOutput = instGetDstByType(inst, OUTPUT_EPOCH); */

/* 	       if (g_list_length(epochOutput->consumers) == 0) { */
/* 		    PArg coordinationOutput = instGetDstByType(inst, OUTPUT_EPOCH_COORDINATION); */
/* 		    PArg coordinationInput = instGetSrcByType(inst, INPUT_EPOCH_COORDINATION); */
		    
/* 		    PArg consumer; */
/* 		    PArg producer; */

/* 		    forEach(coordinationOutput->consumers, j, consumer) { */
/* 			 forEach(coordinationInput->producers, k, producer) { */
/* 			      dfgAddEdge(producer, consumer); */
/* 			 } */
/* 		    } */

/* 		    dfgRemoveInstEdges(inst); */
/* 		    basicBlockRemoveInst(block, inst); */
/* 		    instFree(inst); */
/* 	       } */
/* 	  } */
     }
}


void epochOptimizeCoordinators(PFunction func)
{
     epochOptimizeRemoveCoordinatorsInEmptyBlocks(func);
}

void epochMarkUnneededCoordinators(PFunction func){
     GList *i, *j;//, *k;
     PBasicBlock block;
     PInst inst;


     forEach(func->basicBlocks, i, block) {
	  PBasicBlock dstBlock = NULL;
	  bool_t zapThem = FALSE;;

	  forEach(block->insts, j, inst) {
/* 	       if (inst->type == OP_TYPE_EPOCH_COORDINATOR) { */
/* 		    PArg epochCoordinationOutput = instGetDstByType(inst, OUTPUT_EPOCH_COORDINATION); */
/* 		    PArg consumer =  (PArg)epochCoordinationOutput->consumers->data; */
/* 		    if (g_list_length(epochCoordinationOutput->consumers) == 1 && consumer->owner->type == OP_TYPE_EPOCH_COORDINATOR) { */
/* 			 dstBlock = consumer->owner->owner; */
/* 			 zapThem = TRUE; */
/* 		    } */
/* 	       } else */ 
	       if(dstBlock != NULL && inst->type == OP_wa) {
		    PArg output = instGetDstByType(inst, OUTPUT_C);
		    PArg consumer =  (PArg)output->consumers->data;
		    if (!(g_list_length(output->consumers) == 1 && 
			  consumer->owner->owner == dstBlock &&
			  consumer->owner->type == OP_wa 
			  //			  && g_list_length(consumer->producers) == 1
			  )) {
			 zapThem = FALSE;
			 break;
		    }
	       }
	  }

	  if (zapThem) {
	       fprintf(stderr, "Zapping %d\n", block->id);
	       forEach(block->insts, j, inst) {
		    if (/* inst->type == OP_TYPE_EPOCH_COORDINATOR || */ inst->type == OP_wa) {
			 inst->latencyType |= LATENCY_ZAPPED_EPOCH_MANAGEMENT;
		    }
	       }
	  }
     }

}
