#include "types.h"
#include "execFormat.h"
#include "assert.h"

#define uint8 guint8
#define uint16 guint16
#define uint32 guint32
#define uint64 guint64
#define int8 gint8
#define int16 gint16
#define int32 gint32
#define int64 gint64

#	define	cint8(xxx)		((int8)(xxx))
#	define	cint16(xxx)		((int16)(xxx))
#	define	cint32(xxx)		((int32)(xxx))
#	define	cint64(xxx)		((int64)(xxx))
#	define	cuint8(xxx)		((uint8)(xxx))
#	define	cuint16(xxx)	((uint16)(xxx))
#	define	cuint32(xxx)	((uint32)(xxx))
#	define	cuint64(xxx)	((uint64)(xxx))
#include "AlphaParsing.h"


#ifndef Darwin
#include "malloc.h"
#endif

#define ATTRIBUTE_COLUMN 80 
static bool_t *opCodeTakesImmediate = NULL;
static bool_t *opCodeIsMemory = NULL;

#ifndef IN_SURFER

#include "machine.h"

static void initTables()
{
     if (opCodeTakesImmediate != NULL) {
	  return;
     }
     nice_calloc(opCodeTakesImmediate, WSOpCodeMax *sizeof(bool_t));
     nice_calloc(opCodeIsMemory, WSOpCodeMax *sizeof(bool_t));

#include"uniform_INST.def"

#define true TRUE
#define false FALSE

#define OP_ENUM(a) OP_##a
     
#define INST(OPCODE, FLAGS, IN_COUNT, OUT_COUNT, IMM, IS_MEM, INPUTS, AUTO_INPUTS, OUTPUTS, AUTO_OUTPUTS, FIRE_0_IN_SET, FIRE_0_OUT_SET, FIRE_0_CODE, FIRE_1_IN_SET, FIRE_1_OUT_SET, FIRE_1_CODE, FIRE_2_IN_SET, FIRE_2_OUT_SET, FIRE_2_CODE) \
     opCodeTakesImmediate[OP_ENUM(OPCODE)] = IMM;			\
     opCodeIsMemory[OP_ENUM(OPCODE)] = IS_MEM;		       
     
#define LINK(a, b) \
     opCodeTakesImmediate[OP_ENUM(b)] = opCodeTakesImmediate[OP_ENUM(a)]; \
     opCodeIsMemory[OP_ENUM(b)]       = opCodeIsMemory[OP_ENUM(a)];

#include "Instructions.def"
     opCodeTakesImmediate[OP_TYPE_ALPHA_OP] = false;	\
     opCodeIsMemory[OP_TYPE_ALPHA_OP] = false;		\
     opCodeTakesImmediate[OP_TYPE_FUNCTION_CALL] = false;	\
     opCodeIsMemory[OP_TYPE_FUNCTION_CALL] = false;

#undef INST
#undef CUSTOM_INST
#undef LINK

}

static int execFormatWriteInput(PArg src, FILE *f, bool_t writeAssembly, bool_t *firstItem)
{
     EncodedInput eSrc;
     GList *l;
     PArg producer;
     bool first = true;
     int charCount =0;
     eSrc.type = src->inputType;
     eSrc.producerCount = g_list_length(src->producers);
     eSrc.producers = NULL;


     if (writeAssembly) {

	  if (!(*firstItem)) {
	       charCount += fprintf(f, ", ");
	  } else {
	       *firstItem = FALSE;
	  }
	  if (src->isZero &&
	      src->reg == -1) {
	       charCount += fprintf(f, "{}");
	       return charCount;
	  }
	  if (g_list_length(src->producers) != 1) {
	       charCount += fprintf(f, "{ ");
	  }
     } else {
	  if (fwrite(&eSrc, sizeof(eSrc), 1, f) != 1) {
	       assert(0);
	  }
     }

     forEach(src->producers, l, producer) {
	  EncodedProducer eProducer;
	  eProducer.inst = producer->owner->id;
	  eProducer.type = producer->outputType;
	       
	  if (writeAssembly) {
	       charCount += fprintf(f, "%sv%d", first ? "" : ", ",  producer->id);
	       first = false;
	  } else {
	       if (fwrite(&eProducer, sizeof(eProducer), 1, f) != 1) {
		    assert(0);
	       }
	  }
     }
     

     if (writeAssembly) {
	  if (g_list_length(src->producers) != 1) {
	       charCount += fprintf(f, "}");
	  }
     }
     return charCount;
}

static int execFormatWriteOutput(PArg dst, FILE *f, bool_t writeAssembly, bool_t *firstItem)
{
     EncodedOutput eDst;
     GList *l;
     PArg consumer;
     int charCount = 0;
     eDst.type = dst->outputType;
     eDst.consumerCount = g_list_length(dst->consumers);
     eDst.consumers = NULL;
     
     if (writeAssembly) {
	  if (!(*firstItem)) {
	       charCount +=  fprintf(f, ", ");
	  } else {
	       *firstItem = FALSE;
	  }
	  charCount += fprintf(f, "v%d", dst->id);
     } else {
	  if (fwrite(&eDst, sizeof(eDst), 1, f) != 1) {
	       assert(0);
	  }
	  forEach(dst->consumers, l, consumer) {
	       EncodedConsumer eConsumer;
	       eConsumer.inst = consumer->owner->id;
	       eConsumer.type = consumer->inputType;
	       
	       if (fwrite(&eConsumer, sizeof(eConsumer), 1, f) != 1) {
		    assert(0);
	       }
	  }
     }

     return charCount;

}

static void writeFunctionExport(FILE *f, PFunction func, char *labelName)
{

     fprintf(f, ".export ^%s as %s ", labelName, labelName);

}

static void execFormatWriteInstruction(PInst inst, FILE *f, bool_t writeAssembly)
{


     EncodedInstruction eInst;
     ArgIndex i;
     int j;
     PArg src, dst;
     PArg outC;
     PArg outA;
     PArg inA;
     int dstCount = 0;
     int srcCount = 0;
     bool_t firstItem = TRUE;
     int outputColumn = 0;
     eInst.type = inst->type;
     eInst.id = inst->id;
     eInst.pc = inst->pc;
     eInst.latencyType = inst->latencyType;
     
     if (inst->owner->insts->data == inst) { // This means this is the first
					     // instruction in its basic block.
	  eInst.extraInstructions = inst->owner->extraInstructions;
     } else {
	  eInst.extraInstructions = 0;
     }

     eInst.stubID = 0;
     eInst.isForLocalStub =0;
     if (inst->owner->owner->stubID != 0) {
	  if (inst->isForLocalInvocation) {
	       eInst.stubID = inst->owner->owner->localStubID;
	       eInst.isForLocalStub = 1;
	  } else {
	       eInst.stubID = inst->owner->owner->stubID;
	  }
     }
     if (inst->regNum != -1) {
	  eInst.regNum = inst->regNum;
     } else {
	  if (inst->type == OP_src) {
	       assert(0);
	  } else if (inst->type == OP_ir ||
		     inst->type == OP_func_start) {
	       if (eInst.stubID != 0) {
		    eInst.regNum = inst->regNum;
	       } else {
		    outA = instGetDstByType(inst, OUTPUT_A);
		    eInst.regNum = outA->reg;
	       }
	  } else if (inst->type == OP_is_pc) {
	       inA = instGetSrcByType(inst, INPUT_A);
	       eInst.regNum = inA->reg;
	  } else {
	       eInst.regNum = -1;
	  }
     }



     eInst.isCanonicalEpochAdvance = inst->isCanonicalEpochAdvance;
     eInst.waveAdvanceCount = inst->waveAdvanceCount;
     if (inst->type == OP_TYPE_ALPHA_OP) {
	  enum md_opcode op;
	  MD_SET_OPCODE(op, inst->binInst);
	  eInst.immediate = LiteralField(inst->binInst);
	  eInst.opcode = SSToWSMap[op];
	  assert(eInst.opcode != -1);
     } else {
	  if (inst->type == OP_wa &&
	      inst->isCanonicalEpochAdvance) {
	       eInst.opcode = OP_cwa;
	  } else {
	       eInst.opcode = inst->type;
	  }
     }
     
     eInst.waveAdvanceAmount = inst->waveAdvanceAmount;

     eInst.epochOrderPrev = inst->epochOrderPrev; 
     eInst.epochOrder =	    inst->epochOrder;	   
     eInst.epochOrderNext = inst->epochOrderNext; 
     eInst.lastStoreSequenceNumber   = inst->lastStoreSequenceNumber;
     //     eInst.rippleSize     = 1;//inst->rippleSize;
     
     eInst.isForFunctionCall = inst->isForFunctionCall;

     eInst.isZeroStore = 0;
     if (inst->type == OP_TYPE_ALPHA_OP) {
	  int op;
	  MD_SET_OPCODE(op, inst->binInst);

	  if (MD_OP_FLAGS(op) & F_STORE) {
	       PArg data = instGetSrcByType(inst, INPUT_A);
	       if (data->reg == -1) {
		    eInst.isZeroStore = 1;
	       }
	  }
	  
	  // The correct flags to check for are derived by carefully grepping
	  // through machine.def to see what means whats.
	  if (MD_OP_FLAGS(op) & F_MEM ||
	      op == LDA || 
	      op == LDAH) {
	       eInst.immediate = MemoryDisplacementField(inst->binInst);
	  } else if (MD_OP_FLAGS(op) & F_DIRJMP) {
	       eInst.immediate = BranchDisplacementField(inst->binInst);
	  } else {
	       eInst.immediate = LiteralField(inst->binInst);
	  }
	  if (op == RETN ||
	      op == JMP ||
	      op == JSR ||
	      op == JSR_COROUTINE ||
	      op == AMASK) {
	       eInst.immediate = 0;
	  }
     } else {
	  eInst.immediate = inst->immediate;
     }

     if (inst->type == OP_TYPE_ALPHA_OP) {
	  eInst.binInst = (AlphaInst)inst->binInst;
     } else {
	  memset(&eInst.binInst, sizeof(AlphaInst), 0);
     }

     forEachDst(inst, i, dst) {
	  dstCount++;
     }

     forEachSrc(inst, i, src) {
	  srcCount++;
     }
     
     eInst.inputCount = srcCount;
     eInst.outputCount = dstCount;
     
     eInst.inputs = NULL;
     eInst.outputs = NULL;
    
     if (writeAssembly) {
	  if (inst->instructionLabel != NULL) {
	       if (inst->exportLabel) {

		    writeFunctionExport(f, inst->owner->owner, inst->instructionLabel);
		    fprintf(f," .attrs(");
		    fprintf(f, "pc=%lld, ", (uint64)inst->owner->owner->localEntry->entrancePC);
		    fprintf(f, "localInvocationPC=%lld, ", (uint64)inst->owner->owner->localEntry->entrancePC);
		    fprintf(f, "isMain=%lld, ", (uint64)inst->owner->owner->isMain);
		    fprintf(f, "stubID=%lld", (uint64)inst->owner->owner->localStubID);
		    fprintf(f,")\n");
	       }
	       fprintf(f,"%s:\n", inst->instructionLabel);
	       outputColumn = 0;
	  }
	  if (cfgOptions.doTokenMemory) {
	       if (opCodeIsMemory[eInst.opcode]) {
		    outputColumn += fprintf(f,"\t%s_TOK ", WSOpCodeName[eInst.opcode]);
	       } else {
		    outputColumn += fprintf(f,"\t%-20s ", WSOpCodeName[eInst.opcode]);
	       }
	  } else {
	       outputColumn += fprintf(f,"\t%-20s ", WSOpCodeName[eInst.opcode]);
	  }
     } else {
	  if (fwrite(&eInst, sizeof(eInst), 1, f) != 1) {
	       assert(0);
	  }
     }
     

     for(j = OUTPUT_A; j <= OUTPUT_C; j++) {
	  dst = instGetDstByType(inst, j);
	  if (dst && dst->reg != -1) {
	       outputColumn += execFormatWriteOutput(dst, f, writeAssembly, &firstItem);
	  }
     }

     for(j = INPUT_A; j <= INPUT_C; j++) {
	  src = instGetSrcByType(inst, j);
	  if (src && 
	      (src->reg != -1 ||
	       src->isZero)) {
	       outputColumn += execFormatWriteInput(src, f, writeAssembly, &firstItem);
	  } 
     }

     if (opCodeTakesImmediate[eInst.opcode]) {
	  if (writeAssembly) {
	       if (inst->symbolicImmediate != NULL) {
		    outputColumn +=  fprintf(f, "%s^%s ", firstItem ? "" : ", ", inst->symbolicImmediate);
	       } else {
		    outputColumn +=  fprintf(f, "%s%llu ", firstItem ? "" : ", ", eInst.immediate);
	       }
	       firstItem = FALSE;
	  }
     }
     eInst.outAisZero = 0;
     eInst.inAisZero = 0;
     eInst.inBisZero = 0;
     
     if (inst->type == OP_TYPE_ALPHA_OP) {
	  src = instGetSrcByType(inst, INPUT_A);
	  if (src) {eInst.inAisZero = src->isZero;}
	  src = instGetSrcByType(inst, INPUT_B);
	  if (src) {eInst.inBisZero = src->isZero;}
	  dst = instGetDstByType(inst, OUTPUT_A);
	  if (dst) {eInst.outAisZero = dst->isZero;}
     }
     eInst.epochOrderPrev = inst->epochOrderPrev; 
     eInst.epochOrder =	    inst->epochOrder;	   
     eInst.epochOrderNext = inst->epochOrderNext; 
     eInst.rippleSize     = inst->rippleSize;
     
     if (inst->owner->owner->isMain) {
	  eInst.reportValue = inst->reportValue;
     } else {
	  eInst.reportValue = 0;
     }

     if (writeAssembly) {
	  int i;
	  if (opCodeIsMemory[eInst.opcode]) {
	       if (cfgOptions.doTokenMemory) {
	       } else {
		    outputColumn += fprintf(f, " <%d, %d , %d>.%d ",
					    eInst.epochOrderPrev,
					    eInst.epochOrder,
					    eInst.epochOrderNext,
					    //				       rippleSize,
					    eInst.lastStoreSequenceNumber
					    );
	       }
	       firstItem = FALSE;
	  }
	  for(i = outputColumn; i < ATTRIBUTE_COLUMN; i++) {
	       fprintf(f, " ");
	  }
	  {
	       bool_t first = true;
	       char buf[2000];
	       char *w = buf;
#define ATTRIBUTE(a, d) do{if (eInst.a != d) {w+=sprintf(w, "%s%s=%lld", first ? "" : ", ", #a, (uint64)eInst.a); first = false;}} while (0)
	       ATTRIBUTE(isZeroStore, 0);
	       ATTRIBUTE(stubID, 0);
	       ATTRIBUTE(pc, (uint64)-1);
	       ATTRIBUTE(latencyType, 0);
	       ATTRIBUTE(extraInstructions, 0);
	       ATTRIBUTE(regNum, -1);
	       ATTRIBUTE(inAisZero, 0);
	       ATTRIBUTE(inBisZero, 0);
	       ATTRIBUTE(outAisZero, 0);
	       ATTRIBUTE(reportValue, 0);
	       ATTRIBUTE(isForFunctionCall, 0);
	       ATTRIBUTE(waveAdvanceAmount, 0);
	       ATTRIBUTE(isForLocalStub, 0);
#undef ATTRIBUTE
	       if (w != buf ) {
		    fprintf(f, "  .attrs(%s)", buf);
	       }
	       fprintf(f, "; %ld\n", eInst.id);
	  }
     }
}

static void execFormatWriteFunction(PFunction func, FILE *f, bool_t writeAssembly)
{
     EncodedFunction eFunc;
     GList *i, *j;
     PBasicBlock block;
     PInst inst;
     int instCount = 0;

     eFunc.nameLength = strlen(func->name) + 1;
     forEach(func->basicBlocks, i, block) {
	  instCount += g_list_length(block->insts);
     }
     eFunc.stubID = func->stubID;
     eFunc.instructionCount = instCount;

     eFunc.remoteInvocationPC = func->pc;
     if (func->localEntry) {
	  eFunc.localInvocationPC  = func->localEntry->entrancePC;
     } else {
	  eFunc.localInvocationPC = 0;
     }

     eFunc.name = NULL;
     eFunc.instructions = NULL;
     eFunc.isMain = func->isMain;

     if (writeAssembly) {
	  bool_t first = TRUE;
	  fprintf(f,";;; begin %s\n", func->name);
	  
	  writeFunctionExport(f, func, func->name);
	  fprintf(f," .attrs(");
	  fprintf(f, "pc=%lld", (uint64)eFunc.remoteInvocationPC); first = false;
#define ATTRIBUTE(a, d) do{if (eFunc.a != d) {fprintf(f, "%s%s=%lld", first ? "" : ", ", #a, (uint64)eFunc.a); first = false;}} while (0)       
	  ATTRIBUTE(stubID, 0);
	  ATTRIBUTE(isMain, 0);
#undef ATTRIBUTE
	  fprintf(f,")\n");
	  fprintf(f, "%s:\n", func->name);
     } else {
	  if (fwrite(&eFunc, sizeof(eFunc), 1, f) != 1) {
	       assert(0);
	  }
     }

     if (writeAssembly) {
     } else {
	  if (fwrite(func->name, 
		sizeof(char), 
		eFunc.nameLength, f) != eFunc.nameLength) {
	       assert(0);
	  }
     }
	  
#if (0)
     {
	  uint i;
	  GList *workList = NULL;
	  PBasicBlock cur;
	  PCfgEdge e;
     
	  workList = g_list_append(workList, func->enter);
	  if (func->localEntry) {
	       workList = g_list_append(workList, func->localEntry);
	  }
	  while(workList != NULL) {
	       GList *l;
	       GList *i;
	       cur = g_list_top(workList);
	       workList = g_list_pop(workList);
	       if (writeAssembly) {
		    fprintf(f, ";; %s\n",cur->label );		    
	       }
	       forEach(cur->insts, i, inst) {
		    execFormatWriteInstruction(inst, f, writeAssembly);
	       }
	       forEach(cur->exits, l, e) {
		    if (!e->isBackEdge) {
			 workList = g_list_append(workList, e->dst);
		    }
	       }
	  }

     }
#endif

     forEach(func->basicBlocks, i, block) {
	  if (writeAssembly) {
	       fprintf(f, ";; %s\n",block->label );		    
	  }
	  forEach(block->insts, j, inst) {
	       execFormatWriteInstruction(inst, f, writeAssembly);
	  }
     }
     fprintf(f,";;; end %s\n", func->name);


}


void execFormatWriteProgram(PProgram program, FILE *f, bool_t writeAssembly)
{
     GList *i;
     PFunction func;
     EncodedProgram eProgram;
     int version = EXEC_FORMAT_VERSION;
						
     initTables();
     
     if (writeAssembly) {
	  fprintf(f, ".section .code\n");
	  fprintf(f, ".version %d\n", version);
     } else {
	  if (fwrite(&version, sizeof(int), 1, f) != 1) {
	       assert(0);
	  }
     }
     
     eProgram.functionCount = g_list_length(program->functions);
     eProgram.functions = NULL;

     if (writeAssembly) {
     } else {
	  if (fwrite(&eProgram, sizeof(eProgram), 1, f) != 1) {
	       assert(0);
	  }
     }

     forEach(program->functions, i, func) { 
	  execFormatWriteFunction(func, f, writeAssembly);
     }
}
#endif


#ifndef nice_calloc
#define nice_calloc(var,  s) \
do {\
   if (!(var = (typeof(var)) calloc (s, 1))) {\
      fprintf (stderr, "File: %s Line: %d  Can't allocate memory.\n", __FILE__, __LINE__); \
      fprintf (stderr, "line:  %d\n", __LINE__);\
                assert(0);\
        }\
} while (0)
#endif

#define nice_fread(a, s, f) \
if (fread(a, s, 1, f) != 1) {\
     assert(0);\
}

#define nice_calloc_fread(a, s, f) \
do { \
     nice_calloc(a, (s));\
     nice_fread(a, s, f);\
} while (0)

static void execFormatReadInput(FILE *f, EncodedInput *input)
{
     nice_fread(input, sizeof(EncodedInput), f);

     if (input->producerCount > 0) {
	  nice_calloc_fread(input->producers, 
			    sizeof(EncodedProducer) * input->producerCount, 
			    f);
     }
}

static void execFormatReadOutput(FILE *f, EncodedOutput *output)
{
     nice_fread(output, sizeof(EncodedOutput), f);

     if (output->consumerCount > 0) {
	  nice_calloc_fread(output->consumers, 
			    sizeof(EncodedConsumer) * output->consumerCount, 
			    f);
     }
}

static void execFormatReadInstruction(FILE *f, EncodedInstruction *inst)
{
     int i;
     nice_fread(inst, sizeof(EncodedInstruction), f);

     nice_calloc(inst->outputs, sizeof(EncodedOutput) * inst->outputCount);

     for(i = 0; i < inst->outputCount; i++) {
	  execFormatReadOutput(f, &inst->outputs[i]);
     }
     nice_calloc(inst->inputs, sizeof(EncodedInput) * inst->inputCount);

     for(i = 0; i < inst->inputCount; i++) {
	  execFormatReadInput(f, &inst->inputs[i]);
     }

}

static void execFormatReadFunction(FILE *f, EncodedFunction *func)
{
     uint i;

     nice_fread(func, sizeof(EncodedFunction), f);
     nice_calloc_fread(func->name, sizeof(char) * func->nameLength, f);
     
     nice_calloc(func->instructions, sizeof(EncodedInstruction) * func->instructionCount);

     for(i = 0; i < func->instructionCount; i++) {
	  execFormatReadInstruction(f, &func->instructions[i]);
     }

}

EncodedProgram *execFormatReadProgram(FILE *f)
{
     EncodedProgram *eProgram;
     uint i;
     int version;
     
     i = fread(&version, sizeof(int), 1, f);
     assert(version == EXEC_FORMAT_VERSION);

     nice_calloc_fread(eProgram, sizeof(EncodedProgram), f);
     nice_calloc(eProgram->functions, sizeof(EncodedFunction) * eProgram->functionCount);

     for(i = 0; i < eProgram->functionCount; i++) {
	  execFormatReadFunction(f, &eProgram->functions[i]);
     }
     return eProgram;
}

static void  execFormatPrintInput(EncodedInput *input, FILE *f)
{
     int i;

     fprintf(f, "\t\t\tINPUT %s[%d] <- ", 
	     INPUT_ARG_TYPE_NAME[input->type], 
	     input->producerCount);
     
     for(i = 0; i < input->producerCount; i++) {
	  fprintf(f, "%7ld<%s> ", 
		  input->producers[i].inst, 
		  OUTPUT_ARG_TYPE_NAME[input->producers[i].type]);
     }
     fprintf(f, "\n");
}

static void execFormatPrintOutput(EncodedOutput *output, FILE *f)
{
     int i;

     fprintf(f, "\t\t\tOUTPUT %s[%d] -> ", 
	     OUTPUT_ARG_TYPE_NAME[output->type], 
	     output->consumerCount);
     
     for(i = 0; i < output->consumerCount; i++) {
	  fprintf(f, "%7ld<%s> ", 
		  output->consumers[i].inst, 
		  INPUT_ARG_TYPE_NAME[output->consumers[i].type]);
     }
     fprintf(f, "\n");
}

void execFormatPrintInstruction(EncodedInstruction *inst, FILE *f)
{
     int i;

     fprintf(f, "\t\tINST%7ld %15s ",
	     inst->id, 
	     OP_TYPE_NAME[inst->type]);

     if (inst->type == OP_TYPE_ALPHA_OP) {
#ifdef IN_SURFER
	  fprintf(f, "bin: 0x%x@0x%llx ", inst->binInst, inst->pc);
#else
	  fprintf(f, "bin: 0x%x@0x%llx ", inst->binInst.word, inst->pc);
#endif
     }
     
     fprintf(f, "in[%4d] out[%4d]\n", 
	     inst->inputCount, 
	     inst->outputCount);

     for(i = 0; i < inst->inputCount; i++) {
	  execFormatPrintInput(&inst->inputs[i], f);
     }

     for(i = 0; i < inst->outputCount; i++) {
	  execFormatPrintOutput(&inst->outputs[i], f);
     }
     
    fprintf(f, "%s = %d\n", "type", inst->type);
    fprintf(f, "%s = %ld\n", "id", inst->id);
    fprintf(f, "%s = %lld\n", "pc", inst->pc);
    fprintf(f, "%s = 0x%x\n", "binInst", inst->binInst.word); 
    fprintf(f, "%s = %d\n", "opcode", inst->opcode); 
    fprintf(f, "%s = %lld\n", "immediate", inst->immediate);
    fprintf(f, "%s = %d\n", "extraInstructions", inst->extraInstructions); 
    fprintf(f, "%s = %d\n", "regNum", inst->regNum);
    fprintf(f, "%s = %d\n", "isCanonicalEpochAdvance", inst->isCanonicalEpochAdvance);
    fprintf(f, "%s = %d\n", "waveAdvanceCount", inst->waveAdvanceCount);
    fprintf(f, "%s = %d\n", "epochOrderPrev", inst->epochOrderPrev);
    fprintf(f, "%s = %d\n", "epochOrder", inst->epochOrder);
    fprintf(f, "%s = %d\n", "epochOrderNext", inst->epochOrderNext);
    fprintf(f, "%s = %d\n", "rippleSize", inst->rippleSize);
    fprintf(f, "%s = %d\n", "latencyType", inst->latencyType);
    fprintf(f, "%s = %d\n", "inputCount", inst->inputCount);
    fprintf(f, "%s = %d\n", "outputCount", inst->outputCount);
    fprintf(f, "%s = %d\n", "isLocalInvocationInstruction", inst->isLocalInvocationInstruction);
    fprintf(f, "%s = %ld\n", "controllingWaveCoordinate", inst->controllingWaveCoordinate);
    fprintf(f, "%s = %d\n", "isZeroStore", inst->isZeroStore);
    fprintf(f, "%s = %d\n", "stubID", inst->stubID);
}

static void execFormatPrintFunction(EncodedFunction *func, FILE *f)
{
     uint i;
     fprintf(f, "\tFUNCTION: %s@(0x%llx,0x%llx) [%4d]\n", func->name, func->localInvocationPC, func->remoteInvocationPC, func->instructionCount);

     for (i = 0;i < func->instructionCount; i++) {
	  execFormatPrintInstruction(&func->instructions[i], f);
     }

     
}

void execFormatPrintProgram(EncodedProgram *program, FILE *f)
{
     uint i;
     fprintf(f, "PROGRAM: [%4d]\n",  program->functionCount);
     for (i = 0; i < program->functionCount; i++) {
	  execFormatPrintFunction(&program->functions[i], f);
     }
}
