#ifndef EXEC_FORMAT_INCLUDED
#define EXEC_FORMAT_INCLUDED
#include"WSOpCode.h"

#ifdef __cplusplus
extern "C" {
#endif

#ifdef IN_SURFER
#define bool int
#include "Core/Global.h"
#else 
#include "program.h"
#define bool bool_t
#endif

#include <stdio.h>
#include "types.h"
#include "instDefs.h"


#ifdef IN_SURFER
typedef uint32 AlphaInst;
#else 
#define SKIP_ATOM_STUFF
#include "alphaInst.h"
#endif

#define EXEC_FORMAT_VERSION 9

     
#define GUTS_ITEM(type, name) type name

     
 /* Input/OutputArgType are defined in instDefs.h as is InstType.  
  */

typedef struct _ENCODED_CONSUMER {


#define ENCODED_CONSUMER_GUTS \
     GUTS_ITEM(InstructionID, inst);		\
     GUTS_ITEM(InputArgType, type)

     ENCODED_CONSUMER_GUTS;
} EncodedConsumer;

typedef struct _ENCODED_PRODUCER {
#define ENCODED_PRODUCER_GUTS \
     GUTS_ITEM(InstructionID, inst);\
     GUTS_ITEM(OutputArgType, type)

     ENCODED_PRODUCER_GUTS;
} EncodedProducer;

typedef struct _ENCODED_OUTPUT {
#define ENCODED_OUTPUT_GUTS \
     GUTS_ITEM(OutputArgType, type);  \
     GUTS_ITEM(SmallCount, consumerCount); \
     GUTS_ITEM(EncodedConsumer *, consumers)

     ENCODED_OUTPUT_GUTS;
} EncodedOutput;

typedef struct _ENCODED_INPUT {
#define ENCODED_INPUT_GUTS \
     GUTS_ITEM(InputArgType, type); \
     GUTS_ITEM(SmallCount, producerCount); \
     GUTS_ITEM(EncodedProducer *,producers)

     ENCODED_INPUT_GUTS;
} EncodedInput;


/**
 * These structures should only be used temporarily to load the executable.  It
 * should immediately be translated into a collection of DecodedInstructions.
 * The reason for this is that DecodedInstruction takes over the memory
 * allocated to create EncodedInstructions.
 */
typedef struct _ENCODED_INSTRUCTION {
#define ENCODED_INSTRUCTION_GUTS   			\
     GUTS_ITEM(InstType, type);                		\
     GUTS_ITEM(InstructionID, id);             		\
     GUTS_ITEM(PC, pc);                        		\
	  /* Only for alpha instructions.  */		\
     GUTS_ITEM(AlphaInst, binInst);            		\
							\
     GUTS_ITEM(WSOpCode, opcode);			\
     GUTS_ITEM(guint64, immediate);			\
	  GUTS_ITEM(int, isZeroStore);			\
							\
	  GUTS_ITEM(int, stubID);			\
	  GUTS_ITEM(int, inAisZero);			\
	  GUTS_ITEM(int, inBisZero);			\
	  GUTS_ITEM(int, outAisZero);			\
	  GUTS_ITEM(int, reportValue);			\
	  GUTS_ITEM(int, isForFunctionCall);		\
	  GUTS_ITEM(int, isForLocalStub);		\
     /* Zero for assembled code */			\
     GUTS_ITEM(int, extraInstructions);        		\
     /* This only for special sources and sinks. */  	\
     GUTS_ITEM(int, regNum);			        \
     /* This only for wave advances.  */		\
     GUTS_ITEM(int, isCanonicalEpochAdvance);	        \
     GUTS_ITEM(int, waveAdvanceCount);	                \
     GUTS_ITEM(int, waveAdvanceAmount);	                \
							\
     /* The next two are For mem-ops only */   		\
     GUTS_ITEM(MemSequence, epochOrderPrev);         	\
     GUTS_ITEM(MemSequence, epochOrder);             	\
     GUTS_ITEM(MemSequence, epochOrderNext);   		\
     GUTS_ITEM(RippleSize, rippleSize);			\
     GUTS_ITEM(MemSequence, lastStoreSequenceNumber);	\
							\
     /* Probably Ignored for assembled code. */	       	\
     GUTS_ITEM(LatencyType, latencyType);      		\
							\
     GUTS_ITEM(SmallCount, inputCount);        		\
     GUTS_ITEM(SmallCount, outputCount);       		\
     GUTS_ITEM(EncodedInput*, inputs);         		\
	  /* ignored for assembled code */		\
     GUTS_ITEM(bool, isLocalInvocationInstruction); 	\
	  /* ignored for now in assembled code.  This is not really a wave coordinate, it is a waveadvance */		\
     GUTS_ITEM(InstructionID, controllingWaveCoordinate);	\
     GUTS_ITEM(EncodedOutput*, outputs)

     ENCODED_INSTRUCTION_GUTS;
} EncodedInstruction;

#undef GUTS_ITEM

typedef struct _ENCODED_FUNCTION {
     Count nameLength;
     Count instructionCount;
     int stubID;
     /* these two should be equal in assembled code */
     PC localInvocationPC;
     PC remoteInvocationPC;
     int isMain;			
     char *name;
     EncodedInstruction *instructions;
} EncodedFunction;

typedef struct _ENCODED_PROGRAM {
     Count functionCount;
     EncodedFunction *functions;
} EncodedProgram;

#ifndef IN_SURFER
     void execFormatWriteProgram(PProgram program, FILE *f, bool writeAssembly);
#endif


EncodedProgram *execFormatReadProgram(FILE *f);
void execFormatPrintProgram(EncodedProgram *program, FILE *f);
void execFormatPrintInstruction(EncodedInstruction *inst, FILE *f);
void execFormatSPrintInstruction(EncodedInstruction *inst, char *s);

#ifdef IN_SURFER
#undef bool
#endif

#ifdef __cplusplus
}
#endif

#endif
