#include "function.h"
#include <glib.h>
#include <stdlib.h>
#include <stdio.h>
#include "basicBlock.h"
#include "inst.h"
#include <string.h>
#include "list.h"
#include "buildInst.h"
#include <assert.h>
#include "cfg.h"
#include "dot.h"
#include "wsArch.h"

#include "rippleChunk.h"
#include "functionDefs.h" 

int stubCount = 1;
PFunction functionNew(char * name, int action)
{
     PFunction n;
     n = (PFunction)calloc(sizeof(Function), 1);
     n->name = strdup(name);
     n->usedVariables = setNew(MAX_VARIABLES);
     n->variablesToDefiningBlocks = g_hash_table_new(g_direct_hash, g_direct_equal);
     
     // The order here is importart.  The implementation is bound to
     // functionAddBasicBlock

     if (cfgOptions.includeExitEnterBlocks) {
	  n->enter = basicBlockNew("enter", -1);
	  n->enter->isArtificial = TRUE;
	  
	  functionAddBasicBlock(n, n->enter);
	  
	  n->exit =  basicBlockNew("exit", -1);
	  n->exit->isArtificial = TRUE;
	  
	  functionAddBasicBlock(n, n->exit);
     }
     n->compileAction = action;
     return n;

}


void functionFree(PFunction f)
{
     
     g_list_free(f->basicBlocks);
     free(f->name);
     free(f);
}

static void functionGutForStubbing(PFunction function)
{
     GList *l;
     PBasicBlock b;
     GList *copy;
     copy = g_list_copy(function->basicBlocks);
     forEach(copy, l, b) {
	  if (b == function->enter ||
	      b == function->exit ||
	      b == function->returnBlock ||
	      b == function->localEntry ||
	      b == function->remoteEntry) {
	  } else {
	       functionRemoveBasicBlock(function, b);
	       cfgRemoveBlock(b);
	       basicBlockFree(b);
	  }
     }
     g_list_free(copy);
     
}

PBasicBlock functionCreateStubBlock(bool_t isLocal)
{
     uint i;
     PInst fanout, memPing;

     PBasicBlock stubBlock = basicBlockNew(isLocal ? "LocalStub" : "RemoteStub", -1);
     stubBlock->handWaveNumber = TRUE;

     if (cfgOptions.doTokenMemory) {
	  memPing = instNew(OP_nop);
     } else {
	  memPing = instNew(OP_mnop_ping);
	  memPing->epochOrderPrev = 1;
	  memPing->epochOrder = 2;
	  memPing->epochOrderNext =1;
	  memPing->lastStoreSequenceNumber = 1;
     }
     basicBlockAppendInst(stubBlock, memPing);
     

     i = 0;
     while(functionInputRegs[i] != -1) {
	  PInst inst = instNew(OP_stub_receive);
	  //	  if (!cfgOptions.doTokenMemory) {
	  if (i == 0) {
	       instAddSrc(memPing, functionInputRegs[i], INPUT_A);
	       instAddDst(memPing, functionInputRegs[i], OUTPUT_A);
	  }
	       //}
	  instAddSrc(inst, functionInputRegs[i], INPUT_A);
	  instAddDst(inst, functionInputRegs[i], OUTPUT_A);
	  inst->regNum = functionInputRegs[i];
	  basicBlockAppendInst(stubBlock, inst);
	  i++;
     }


     // The order that this happens in must correspond to the code that sends
     // values to these insturcitons in
     // surfer/WaveCache/SimpleSimulator.cxx:SimpleSimulator::Applicaiton::ReceiveStubValue and FinishStub
     i = 0;
     while(functionOutputRegs[i] != -1) {
	  PInst inst = instNew(OP_stub_send);
	  instAddSrc(inst, functionOutputRegs[i], INPUT_A);
	  instAddDst(inst, functionOutputRegs[i], OUTPUT_A);
	  inst->regNum = functionOutputRegs[i];
	  basicBlockAppendInst(stubBlock, inst);
	  i++;
     }

     fanout = instNew(OP_stub_send);
     instAddSrc(fanout, RETURN_ADDRESS_REGISTER, INPUT_A);
     instAddDst(fanout, RETURN_ADDRESS_REGISTER, OUTPUT_A);
     fanout->regNum = RETURN_ADDRESS_REGISTER;
     basicBlockAppendInst(stubBlock, fanout);

     if (cfgOptions.doTokenMemory) {
	  PInst copy = instNew(OP_nop);
	  instAddSrc(copy, RETURN_ADDRESS_REGISTER, INPUT_A);
	  instAddDst(copy, CONTROL_REGISTER, OUTPUT_A);
	  basicBlockAppendInst(stubBlock, copy);
     }

     if (isLocal) {
	  GList *l; 
	  PInst inst;
	  forEach(stubBlock->insts, l, inst) {
	       inst->isForLocalInvocation = TRUE;
	  }
     }
     return stubBlock;
}

void  functionMakeStub(PFunction function)
{
     uint i;
     PInst fanout, memPing;
     PBasicBlock localStub;
     PBasicBlock remoteStub;
     
     functionGutForStubbing(function);

     remoteStub = functionCreateStubBlock(FALSE);

     functionAddBasicBlock(function, remoteStub);
     cfgAddEdge(function->enter, remoteStub, CFG_EDGE_NORMAL);
     cfgAddEdge(remoteStub, function->exit, CFG_EDGE_NORMAL);

     if (function->localEntry) {
	  localStub = functionCreateStubBlock(TRUE);
	  functionAddBasicBlock(function, localStub);
	  cfgAddEdge(function->localEntry, localStub, CFG_EDGE_NORMAL);
	  cfgAddEdge(localStub, function->exit, CFG_EDGE_NORMAL);
     }

     function->returnBlock = function->exit;
}

void functionSetLocalIDs(PFunction func)
{
     GList *l;
     PBasicBlock block;

     return;
     // It is important for the SSA calculation that the localID of the a
     // basicBlock never be 0, therefore we increment the length before
     // creating the basic bolck.

     func->lengthBasicBlocks = 0;

     forEach(func->basicBlocks, l,block) {
	  block->localID = ++func->lengthBasicBlocks;
     }
}

void functionAddBasicBlock(PFunction func, PBasicBlock block)
{
     block->owner = func;

     // It is important for the SSA calculation that the localID of the a
     // basicBlock never be 0, therefore we increment the length before
     // creating the basic bolck.

     // Make sure func->exit is always the last basicblock in the list.
     if (func->exit == NULL) {
	  block->localID = ++func->lengthBasicBlocks;
	  func->basicBlocks = g_list_append(func->basicBlocks, block);
     } else {
	  block->localID = func->exit->localID;
	  func->basicBlocks = g_list_insert(func->basicBlocks, block, func->exit->localID - 1);
	  func->exit->localID =  ++func->lengthBasicBlocks;
     }
     
}

#if (0)
void functionAddBasicBlock(PFunction func, PBasicBlock block)
{
     block->owner = func;


     // Make sure func->exit is always the last basicblock in the list.
     if (func->exit == NULL) {
	  func->basicBlocks = g_list_append(func->basicBlocks, block);
     } else {
	  func->basicBlocks = g_list_insert(func->basicBlocks, block, g_list_length(func->basicBlocks) - 2);
     }
	       
}
#endif


void functionSanity(PFunction func)
{
     GList *l, *k;     
     PBasicBlock block;
     PInst inst;

     forEach(func->basicBlocks, l, block) {
	  basicBlockSanity(block);
	  forEach(block->insts, k, inst) {
	       instSanity(inst);
	  }
     }
}

void functionPrintAnnotatedMap(PFunction f, FILE * file) 
{
     GList* l;
     PBasicBlock block;

     fprintf(file, "Start Function: %s\n",f->name);
     forEach(f->basicBlocks, l, block) {
	  basicBlockPrintAnnotatedMap(block, file);
     }
     fprintf(file, "End Function\n");
}

PBasicBlock functionFindBasicBlockByLocalID(PFunction func, int localID)
{
     PBasicBlock b;

     assert(localID >0);
     // The -1 is because localID's start at 1.
     b = (PBasicBlock)g_list_nth_data(func->basicBlocks, localID - 1);

     assert(b->localID = localID);
     
     return b;
}

void functionPrint(FILE *file, PFunction f, bool_t verbose)
{
     GList* l;
     PBasicBlock block;
     PFunction func;
     PArg input;

     fprintf(file, "Function: %s Calls:  ", f->name);

     forEach(f->functionsCalled, l, func) {
     	  fprintf(file, "%s ", func->name);
     }
     fprintf(file, "; Inputs: ");

     forEach(f->uniqueInputs, l, input) {
     	  fprintf(file, "%d ", input->reg);
     }
     fprintf(file, "\n");
     
     forEach(f->basicBlocks, l, block) {
	  basicBlockPrint(stderr, block, verbose);
     }
}

void functionRemoveUselessSpecialSources(PFunction func)
{
     GList *l, *k;     
     PBasicBlock block;
     PInst inst;
     GList *c;

     forEach(func->basicBlocks, l, block) {
	  c = g_list_copy(block->insts);
	  forEach(c, k, inst) {
	       if (inst->type == OP_func_start && instCountConsumers(inst) == 0) {
		    dfgRemoveInstEdges(inst);
		    basicBlockRemoveInst(block, inst);
		    instFree(inst);
	       }
	  }
	  g_list_free(c);
     }
}

static void functionInsertSpecialSources_help(PBasicBlock block)
{
     PInst inst;
     int i = 0;

     while(functionInputRegs[i] != -1) {
	  inst = instNew(OP_func_start);
	  instAddDst(inst, functionInputRegs[i], OUTPUT_A);
	  basicBlockAppendInst(block, inst);
	  i++;
     }
     
     inst = instNew(OP_nop);
     instAddSrc(inst, functionInputRegs[0], INPUT_A);
     instAddDst(inst, CONTROL_REGISTER, OUTPUT_A);
     basicBlockAppendInst(block, inst);

#if (0)
     inst = instNew(OP_wtd);
     instAddSrc(inst, RETURN_ADDRESS_REGISTER, INPUT_A);
     instAddDst(inst, FUNCTION_WAVE_REGISTER, OUTPUT_A);
     basicBlockAppendInst(block, inst);
#endif

     if (cfgOptions.serialize) {
	  inst = instNew(OP_func_start);
	  instAddDst(inst, SERIALIZE_REGISTER, OUTPUT_C);
	  basicBlockAppendInst(block, inst);
     }
}

void functionInsertSpecialSources (PFunction func)
{
     PInst inst;



     assert(g_list_length(func->enter->insts) == 0);
     functionInsertSpecialSources_help(func->enter);
     if (func->localEntry != NULL) {
	  char buf[1000];
	  assert(g_list_length(func->localEntry->insts) == 0);
	  functionInsertSpecialSources_help(func->localEntry);

	  inst = func->localEntry->insts->data;
	  sprintf(buf, "%s_local", func->name);
	  inst->instructionLabel = strdup(buf);
	  inst->exportLabel = TRUE;
     }
}

void functionFixConditionalMoves(PFunction func)
{
     PBasicBlock block;
     PInst inst;
     GList *l, *k;

     forEach(func->basicBlocks, l, block) {
	  forEach(block->insts, k, inst) {
	       
	       // cmove is different because its source has to be an input as
	       // well, and since there can be an immediate, it maybe the 2nd
	       // or the 3rd input
	       if(instIsConditionalMove(inst)) {
		    PArg dst, src;
		    dst = instGetDstByType(inst, OUTPUT_A);
		    src = instGetSrcByType(inst, INPUT_B);
		    if (src == NULL) {
			 instAddSrc(inst, dst->reg, INPUT_B);
		    } else {			
			 instAddSrc(inst, dst->reg, INPUT_C);
		    }
	       }
	  }
     }
}

void functionCheckForNastiness(PFunction func)
{
     PBasicBlock block;
     PInst inst;
     GList *l, *k;

     forEach(func->basicBlocks, l, block) {
	  forEach(block->insts, k, inst) {
	       /* decode the instruction */
	       enum md_opcode op;
	       MD_SET_OPCODE(op, inst->binInst);
	       
	       if (op == JSR_COROUTINE ||
		   op == JMP) {
		    fprintf(stderr, "JMP or COROUTINE found:  ");
		    instPrintDebug(inst, stderr);
		    fprintf(stderr, "\n");
	       }
	  }
     }
}

     

void functionClearPlacement(PFunction f)
{

     PBasicBlock block;
     PInst inst;
     GList *l, *k;

     forEach(f->basicBlocks, l, block) {
	  forEach(block->insts, k, inst) {
	       inst->locx = -1;
	       inst->locy = -1;
	       inst->isPlaced = FALSE;
	  }
     }
}

void functionCleanUpFunction(PFunction func)
{
     PBasicBlock block;
     PArg dst;
     PInst inst;     
     GList *l1, *l2, *toRemove = NULL;
     int foundRealDst = 0;
     ArgIndex i;

     // This function needs rethinking
     assert(0);
     functionCleanUpArgPorts(func);

     forEach(func->basicBlocks, l1, block) {

	  forEach(block->insts, l2, inst) {

	       if (toRemove != NULL) { // We can't delete anything until we have iterated past it.
		    block->insts = g_list_remove_link(block->insts, toRemove); 
	       }
	       
	       foundRealDst = 0;
	       forEachDst(inst, i, dst) {
		    if (dst->reg != 31 && dst->reg != 32 + 31) {
			 foundRealDst = 1;
		    }
	       }

	       if (!foundRealDst) {
		    toRemove = l2;
	       } else {
		    toRemove = NULL;
	       }

	  }

	  if (toRemove != NULL) { // We can't delete anything until we have iterated past it.
	       block->insts = g_list_remove_link(block->insts, toRemove); 
	  }
     }
}

void functionAddReturnBlock(PFunction func)
{

     PCfgEdge e;
     GList *copy;

     PBasicBlock block;
     PInst inst;

     GList *l, *k, *j;

     copy = g_list_copy(func->exit->entrances);
     forEach(copy, l, e) {
	  if (e->src == func->enter) {
	       cfgRemoveEdge(e);
	       break;
	  }
     }     
     g_list_free(copy);

     func->returnBlock = func->exit;

     forEach(func->basicBlocks, l, block) {
	  forEach(block->insts, k, inst) {
	       enum md_opcode op;
	       MD_SET_OPCODE(op, inst->binInst);	      
	       if (op == RETN) {
		    bool_t edgeAlreadyPresent = FALSE;
		    GList *copy;
		    PCfgEdge e;
		    
		    copy = g_list_copy(block->exits);
		    forEach(copy, j, e) {
			 if (e->dst == func->returnBlock) {
			      edgeAlreadyPresent = TRUE;
			 } else {
			      cfgRemoveEdge(e);
			 }
		    }
		    g_list_free(copy);

		    if (!edgeAlreadyPresent) {
			 cfgAddEdge(block, func->returnBlock, CFG_EDGE_NORMAL);
		    }

		    break;
	       }
	  }
     }
}

void functionAddReturnSinks(PFunction func)
{

     PInst newInst, wtd, dtw1, dtw2, nop,mnop_ping;
     int i = 0;
     
     uint functionFinishWave = getUniqueVariable();
     uint functionFinishWaveForReturn = getUniqueVariable();
     uint returnAddressCopy = getUniqueVariable();
     uint functionWaveReturnAddress = getUniqueVariable();

     if (cfgOptions.doTokenMemory) {
	  PInst wait = instNew(OP_wait);
	  instAddSrc(wait, RETURN_ADDRESS_REGISTER, INPUT_A);
	  instAddSrc(wait, CONTROL_REGISTER, INPUT_B);
	  instAddDst(wait, RETURN_ADDRESS_REGISTER, OUTPUT_A);
	  basicBlockAppendInst(func->returnBlock, wait);
     }

     wtd = instNew(OP_wtd);
     instAddSrc(wtd, FUNCTION_WAVE_REGISTER, INPUT_A);
     instAddDst(wtd, functionFinishWave, OUTPUT_A);
     basicBlockAppendInst(func->returnBlock, wtd);

     nop = instNew(OP_nop);
     instAddSrc(nop, RETURN_ADDRESS_REGISTER, INPUT_A);
     instAddDst(nop, returnAddressCopy, OUTPUT_A);
     basicBlockAppendInst(func->returnBlock,  nop);
     
     dtw1 = instNew(OP_dtw);
     instAddSrc(dtw1, FUNCTION_WAVE_REGISTER, INPUT_A);
     instAddSrc(dtw1, functionFinishWave, INPUT_B);
     instAddDst(dtw1, functionFinishWaveForReturn, OUTPUT_A);
     basicBlockAppendInst(func->returnBlock,  dtw1);

     dtw2 = instNew(OP_dtw);
     instAddSrc(dtw2, FUNCTION_WAVE_REGISTER, INPUT_A);
     instAddSrc(dtw2, returnAddressCopy, INPUT_B);
     instAddDst(dtw2, functionWaveReturnAddress, OUTPUT_A);
     basicBlockAppendInst(func->returnBlock,  dtw2);
     
     i = 0;
     while (functionOutputRegs[i] != -1) {
	  uint reg = functionOutputRegs[i];
	  newInst = instNew(OP_isret);
	  newInst->immediate = i;
	  newInst->regNum = functionOutputRegs[i];
	  if (reg == 0) {
	       //newInst->reportValue = TRUE;
	  }
	  if (reg == FUNCTION_WAVE_REGISTER) {
	       instAddSrc(newInst, functionFinishWaveForReturn, INPUT_A);
	       instAddSrc(newInst, functionWaveReturnAddress, INPUT_B);
	  } else {
/* 	       if (reg == MNOP_PING_DELAY_REGISTER) { */
/* 		    mnop_ping = instNew(OP_mnop_ping); */
/* 		    instAddSrc(mnop_ping, reg, INPUT_A); */
/* 		    instAddDst(mnop_ping, reg, OUTPUT_A); */
/* 		    basicBlockAppendInst(func->returnBlock, mnop_ping); */
/* 	       } */
	       instAddSrc(newInst, reg, INPUT_A);
	       instAddSrc(newInst, RETURN_ADDRESS_REGISTER, INPUT_B);
	  }
	  basicBlockAppendInst(func->returnBlock, newInst);
	  i++;
     }

     if (cfgOptions.serialize) {
	  assert(0);
	  newInst = instNew(OP_sink);
	  instAddSrc(newInst, SERIALIZE_REGISTER, INPUT_A);
	  basicBlockAppendInst(func->returnBlock, newInst);
     }
     
/*      if (cfgOptions.includeWaveCoordinate) { */
/* 	  newInst = instNew(OP_sink); */
/* 	  instAddSrc(newInst, EPOCH_COORDINATION_REGISTER, INPUT_A); */
/* 	  basicBlockAppendInst(func->returnBlock, newInst); */
/*      } */
     
}

void functionSerializeInstructions(PFunction func)
{
     GList *i, *l;
     PBasicBlock block;
     PInst inst;

     forEach(func->basicBlocks, i, block) {
	  forEach(block->insts, l, inst) {
	       if (inst->type == OP_TYPE_ALPHA_OP) {
		    instAddDst(inst, SERIALIZE_REGISTER, INPUT_SERIALIZE);
		    instAddSrc(inst, SERIALIZE_REGISTER, OUTPUT_SERIALIZE);
	       }
	  }
     }
}

void functionRemoveMarkedDeadBlocks(PFunction func)
{
     GList *l;
     PBasicBlock block;
     GList *c;
     
     c = g_list_copy(func->basicBlocks);
     forEach(c, l, block) {
	  if (block->isDead) {
	       assert(g_list_length(block->entrances) == 0 && g_list_length(block->exits) == 0);
	       func->basicBlocks = g_list_remove(func->basicBlocks, block);
	  }
     }
     g_list_free(c);
}

void functionRemoveBasicBlock(PFunction func, PBasicBlock block)
{
     func->basicBlocks = g_list_remove(func->basicBlocks, block);
}

// Get rid of the unreachable basicblocks the compiler inserts for cache alignment.
void functionRemoveAlignmentBlocks(PFunction func)
{
     GList *i, *j, *k,*copy2;
     PBasicBlock block;
     PInst inst;

     copy2 = g_list_copy(func->basicBlocks);
     forEach(copy2, i, block) {
	  if (g_list_empty(block->entrances) && block->localID != 1 && !block->isArtificial) {
	       bool_t reallyWorthless = TRUE;
	       forEach(block->insts, j, inst) {
		    enum md_opcode op;
		    MD_SET_OPCODE(op, inst->binInst);
		    if (inst->type != OP_TYPE_ALPHA_OP) {
			 reallyWorthless = FALSE;
			 break;
		    }

		    if (op == BIS) {
			 if (EXTRACT_RA(inst->binInst) == 31 && 
			     EXTRACT_RB(inst->binInst) == 31 && 
			     EXTRACT_RC(inst->binInst) == 31) {
			 } else {
			      reallyWorthless = FALSE;
			      break;
			 }
		    } else if (op == LDA || op == LDAH) {
			 if (EXTRACT_RA(inst->binInst) == 31) {
			 } else {
			      reallyWorthless = FALSE;
			      break;
			 }
		    } else if (op == LDQ_U) {
			 if (EXTRACT_RA(inst->binInst) == 31) {
			 } else {
			      reallyWorthless = FALSE;
			      break;
			 }
		    }
	       }

	       if (reallyWorthless) {
		    GList *copy = g_list_copy(block->insts);
		    forEach(copy, k, inst) {
			 basicBlockRemoveInst(block, inst);
			 instFree(inst);
		    }
		    g_list_free(copy);
		    
		    block->isDead = TRUE;
		    cfgRemoveBlock(block);
		    basicBlockFree(block);
	       }

	       //	       fprintf(stderr, "suspicious block: %d\n", block->id);
	       //		    assert(0);
	  }

     }
     g_list_free(copy2);

}

void functionMarkEntries(PFunction func)
{
    
     PBasicBlock remoteInvocationBlock = FUNC_BASIC_BLOCK_BY_LOCAL_ID(func, 2);
     GList *l;
     PInst inst;
     bool_t rightKind = TRUE;
    
     assert(0);

     forEach(remoteInvocationBlock->insts, l, inst) {
	  enum md_opcode op;
	  
	  MD_SET_OPCODE(op, inst->binInst);
	  if (!(op == LDAH || op == LDA || op == LDQ_U)) {
	       rightKind = FALSE;
	       break;
	  }
     }
     
     if(rightKind && g_list_length(remoteInvocationBlock->insts) < 5) {
	  func->remoteEntry = remoteInvocationBlock;
	  func->localEntry =  FUNC_BASIC_BLOCK_BY_LOCAL_ID(func, 3);
     } else {
	  func->localEntry =  FUNC_BASIC_BLOCK_BY_LOCAL_ID(func, 2);
     }
}

void functionFixLocalInvocationInsts(PFunction func)
{
     GList *l;
     PInst inst;
     PBasicBlock remoteInvocationBlock = FUNC_BASIC_BLOCK_BY_LOCAL_ID(func, 2);

     if (func->localEntry == NULL) {
	  return;
     }
	  
     forEach(remoteInvocationBlock->insts, l, inst) {
	  if (instNeedsEpochTags(inst)) {
	       PInst n = instNew(OP_mnop);
	       instAddSrc(n, CONTROL_REGISTER, INPUT_A);
	       basicBlockAppendInst(func->localEntry ,n);
	  }
     }
     forEach(func->localEntry->insts, l, inst) {
	  inst->isForLocalInvocation = TRUE;
     }
}


void functionFixLocalInvocationEpoch(PFunction func)
{
     PBasicBlock remoteInvocationBlock = FUNC_BASIC_BLOCK_BY_LOCAL_ID(func, 2);
     GList *l;
     GList *local;
     PInst localInst = NULL;
     PInst inst;



     if (func->localEntry == NULL) {
	  return;
     }
	  
     local = func->localEntry->insts;
     localInst = local->data;
     while (local != NULL &&
	    !instNeedsEpochTags(localInst)) { 
	  local = local->next;
	  if (local != NULL) {
	       localInst = local->data;
	  }
     }
     
	  
     forEach(func->enter->insts, l, inst) {
	  if (instNeedsEpochTags(inst)) {
	       
	       localInst->epochOrderPrev = inst->epochOrderPrev  ;
	       localInst->epochOrder     = inst->epochOrder     ;
	       localInst->epochOrderNext = inst->epochOrderNext;
	       localInst->rippleSize = inst->rippleSize;
	       localInst->lastStoreSequenceNumber = inst->lastStoreSequenceNumber;

	       local = local->next;
	       localInst = local->data;
	       while (local != NULL &&
		      !instNeedsEpochTags(localInst)) { 
		    local = local->next;
		    if (local != NULL) {
			 localInst = local->data;
		    }
	       }
	       if (local == NULL) {
		    return;
	       }
	  }
     }
     forEach(remoteInvocationBlock->insts, l, inst) {
	  if (instNeedsEpochTags(inst)) {
	       
	       localInst->epochOrderPrev = inst->epochOrderPrev  ;
	       localInst->epochOrder     = inst->epochOrder     ;
	       localInst->epochOrderNext = inst->epochOrderNext;
	       localInst->rippleSize = inst->rippleSize;
	       localInst->lastStoreSequenceNumber = inst->lastStoreSequenceNumber;

	       while (local != NULL &&
		      !instNeedsEpochTags(localInst)) { 
		    local = local->next;
		    if (local != NULL) {
			 localInst = local->data;
		    }
	       }
	       if (local == NULL) {
		    return;
	       }
	  }
     }
}

bool_t functionCompressWaveAdvanceOneInst(PInst inst)
{
     GList *c;
     GList *l;
     ArgIndex k,j,m;
     PArg waDst;
     PArg dst;
     PArg  waConsumer;
     PArg consumer;
     bool_t ret = FALSE;
     GList *copy;
     GList *copy2;

     inst->waveAdvanceAmount++;
     
     forEachDst(inst, k, dst) {
	  copy = g_list_copy(dst->consumers);
	  forEach(copy, c, consumer) {
	       assert(consumer->owner->type == OP_wa && 
		      !consumer->owner->isCanonicalEpochAdvance);
	       
	       forEachDst(consumer->owner, m, waDst) {
		    forEach(waDst->consumers, l, waConsumer) {
			 fprintf(stderr, "Removing %d to %d\n", dst->owner->id, consumer->owner->id);
			 fprintf(stderr, "Adding %d to %d\n",  dst->owner->id, waConsumer->owner->id);
			 dfgAddEdge(dst, waConsumer);
			 dfgRemoveEdge(dst, consumer);
			 ret = TRUE;
		    }
	       }
	  }
	  g_list_free(copy);
     }

     return ret;
}

void functionRemoveUnreachableInstructions(PFunction func)
{
     GList *l, *i;
     PBasicBlock block;
     PInst inst;
     PArg src;
     ArgIndex k;
     GList *c;
     PArg producer;
     bool_t removedSomething = TRUE;
     while(removedSomething) {
	  removedSomething = FALSE;
	  forEach(func->basicBlocks, l, block) {
	       GList *copy = g_list_copy(block->insts);
	       forEach(copy, i, inst) {
		    bool_t hasProducer = FALSE;
		    if (inst->type == OP_func_start ||
			inst->type == OP_ir ||
			inst->type == OP_stub_receive) {
			 continue;
		    }
		    forEachSrc(inst, k, src) {
			 forEach(src->producers, c, producer) {
			      hasProducer = TRUE;
			      break;
			 }
		    }
		    if(!hasProducer) {
			 			      
			 dfgRemoveInstEdges(inst);
			 basicBlockRemoveInst(block, inst);
			 /* 			      fprintf(stderr, "Removed %d\n", inst->id); */
			 instFree(inst);
			 removedSomething = TRUE;

		    }
	       }
	       g_list_free(copy);	  
	  }
     }
}

void functionCompressWaveAdvance(PFunction func)
{
     GList *l, *i;
     PBasicBlock block;
     PInst inst;
     PArg dst;
     ArgIndex k;
     GList *c;
     PArg consumer;
     bool_t compressedSomething = TRUE;
     while(compressedSomething) {
	  compressedSomething = FALSE;
	  forEach(func->basicBlocks, l, block) {
	       forEach(block->insts, i, inst) {
		    bool_t allAreWA = TRUE;
		    bool_t hasOutput = FALSE;
		    if (inst->type != OP_wa) {

			 forEachDst(inst, k, dst) {
			      forEach(dst->consumers, c, consumer) {
				   hasOutput = TRUE;
				   if (consumer->owner->type != OP_wa ||
				       consumer->owner->isCanonicalEpochAdvance) {
					//					fprintf(stderr, "%d has bad output %d\n", inst->id, consumer->owner->id);
					allAreWA = FALSE;
					break;
				   }
			      }
			      if (!allAreWA) {
				   break;
			      }
			 }
		    
			 if (allAreWA && hasOutput) {		       
			      if(functionCompressWaveAdvanceOneInst(inst)) {
				   compressedSomething = TRUE;
			      }
			 }
		    }
	       }
	  }
     }
}

void functionMakeLocalInvocation(PFunction func)
{
     PBasicBlock remoteInvocationBlock = FUNC_BASIC_BLOCK_BY_LOCAL_ID(func, 2);
     PBasicBlock localInvocationBlock = FUNC_BASIC_BLOCK_BY_LOCAL_ID(func, 3);
     GList *l;
     PInst inst;
     bool_t rightKind = TRUE;

     forEach(remoteInvocationBlock->insts, l, inst) {
	  enum md_opcode op;
	  
	  MD_SET_OPCODE(op, inst->binInst);
	  if (!(op == LDAH || op == LDA || op == LDQ_U)) {
	       rightKind = FALSE;
	       break;
	  }
     }
     
     func->remoteEntry = func->enter;
     
     if(rightKind && 
	g_list_length(remoteInvocationBlock->insts) < 5 &&
	g_list_length(remoteInvocationBlock->insts) > 0
	) {
	  PBasicBlock localEntry = basicBlockNew("localEntry", -1);  
	  localEntry->isArtificial = TRUE;
	  functionAddBasicBlock(func, localEntry);
	  basicBlockSetPC(localEntry, localInvocationBlock->entrancePC);
	  cfgAddEdge(localEntry, localInvocationBlock, CFG_EDGE_NORMAL);
	  func->localEntry = localEntry;
     }
}

void functionFixReturns(PFunction func)
{
     GList *l, *i;
     PBasicBlock block;
     PInst inst;

     forEach(func->basicBlocks, l, block) {
	  forEach(block->insts, i, inst) {
 	       PArg arg;
	       enum md_opcode op; 
 	       MD_SET_OPCODE(op, inst->binInst); 
 	       if (op == RETN ||
		   op == JSR ||
		   op == BSR ||
		   op == JSR_COROUTINE
		   ) { 
 		    instAddDst(inst, JUMP_NEXT_PC_REGISTER, INPUT_B); 
		    arg = instGetDstByType(inst, INPUT_A);
		    if (arg && arg->reg == -1) {
			 arg->reg = RETURN_ADDRESS_REGISTER;
		    }
 	       } 
	  }
     }

}

static bool_t functionRegisterIsFunctionOutput(int reg)
{
     uint i = 0; 
     while (functionOutputRegs[i] != -1) {
	  if (reg == functionOutputRegs[i] ) {
	       return TRUE;
	  }
	  i++;
     }
     return FALSE;
}

static bool_t functionRegisterIsFunctionInput(int reg)
{
     uint i = 0; 
     while (functionInputRegs[i] != -1) {
	  if (reg == functionInputRegs[i] ) {
	       return TRUE;
	  }
	  i++;
     }
     return FALSE;
}

void functionInstallIndirectSendFunctionCalls(PFunction func)
{
     GList *l, *k;
     PBasicBlock block;
     PInst inst;
     forEach(func->basicBlocks, l, block) {
	  forEach(block->insts, k, inst) {
	       enum md_opcode op;

	       MD_SET_OPCODE(op, inst->binInst);
	       if (op == BSR || op == JSR) {
		    static int returnLabelCount = 0;
		    char returnLabel[100];
		    int i = 0; 
		    Inst *last = inst;
		    Inst * wtd, *retVal;
		    Inst *indSend, *send;


		    int waveReturn = getUniqueVariable();
		    int returnAddress = getUniqueVariable();
		    int functionWave =  getUniqueVariable();
		    guint64 immediate = 0;

		    sprintf(returnLabel, "ret%d", returnLabelCount++);
		    
		    instAddDst(inst, getUniqueVariable(), OUTPUT_A);
		    instAddDst(inst, JMP_DST_REGISTER, OUTPUT_B);

		    // We can't actually write to RETURN_ADDRESS_REGISTER
		    // because this value is for the caller.  we have to create
		    // a local one.
		    retVal = instNew(OP_cnst);
		    instAddSrc(retVal, CONTROL_REGISTER, INPUT_A);
		    instAddDst(retVal, returnAddress, OUTPUT_A);
		    retVal->symbolicImmediate= strdup(returnLabel);
		    basicBlockInsertInstAfterInst(block, last, retVal);
		    last = retVal;

		    wtd = instNew(OP_wtd);
		    instAddSrc(wtd, CONTROL_REGISTER, INPUT_A);
		    instAddDst(wtd, functionWave, OUTPUT_A);
		    basicBlockInsertInstAfterInst(block, last, wtd);
		    last = wtd;

#if (0)
		    indSend = instNew(OP_is_pc);
		    instAddSrc(indSend, JMP_DST_REGISTER, INPUT_B);
		    instAddSrc(indSend, wtdOut, INPUT_A);
		    indSend->immediate = immediate++;
		    //instAddDst(indSend, bogus , OUTPUT_A); // This is extraneous	
		    basicBlockInsertInstAfterInst(block, last, indSend);
		    last = indSend;
#endif

		    i = 0;
		    while(functionInputRegs[i] != -1) {
			 indSend = instNew(OP_is_pc);
			 indSend->isForFunctionCall = TRUE;
			 instAddSrc(indSend, JMP_DST_REGISTER , INPUT_B);
	       
 			 if (functionInputRegs[i] == FUNCTION_WAVE_REGISTER) {
 			      instAddSrc(indSend, functionWave, INPUT_A); 
 			 } else if (functionInputRegs[i] == RETURN_ADDRESS_REGISTER) {
			      instAddSrc(indSend, returnAddress, INPUT_A);
			 } else {
			      instAddSrc(indSend, functionInputRegs[i], INPUT_A);
			 }
			 
			 indSend->immediate = immediate++;
			 basicBlockInsertInstAfterInst(block, last, indSend);
			 last = indSend;
			 i++;
		    }

		    
		    // Function returns
		    i = 0;
		    while(functionOutputRegs[i] != -1) {
			 send = instNew(OP_ir);
			 if (i == 0) {
			      send->instructionLabel = strdup(returnLabel);
			 }
			 
			 
			 if (functionOutputRegs[i] == FUNCTION_WAVE_REGISTER) {
			     instAddDst(send, waveReturn, OUTPUT_A);
			      /* 			 } else if (functionOutputRegs[i] == 29) { */
			      /* 			      instAddDst(send, getUniqueVariable(), OUTPUT_A); */
			 } else {
			      instAddDst(send, functionOutputRegs[i], OUTPUT_A);
			 }
			 
			 basicBlockInsertInstAfterInst(block, last, send);
			 
			 last = send;
			 i++;
		    }
		    
		    for(i = 0; i < REGISTER_COUNT; i++) {
			 // This catches all the non-output registers and the
			 // one output reg that we skipped.
			 if (!functionRegisterIsFunctionOutput(i) ||
			     i == FUNCTION_WAVE_REGISTER /*||
			     i == 29*/) {
			      Inst *dtw = instNew(OP_dtw);
/* 			      if (cfgOptions.doTokenMemory) { */
/* 				   instAddSrc(dtw, FUNCTION_WAVE_REGISTER , INPUT_A); */
/* 			      } else { */
				   instAddSrc(dtw, waveReturn , INPUT_A);
/* 			      } */
			      instAddSrc(dtw, i , INPUT_B);
			      instAddDst(dtw, i, OUTPUT_A);
			      basicBlockInsertInstAfterInst(block, last, dtw);
			      last = dtw;
			 }
		    }
	       }
	  }
     }
}

void functionInstallFunctionCalls(PFunction func)
{
     GList *l, *i;
     PBasicBlock block;
     PInst inst;

     forEach(func->basicBlocks, l, block) {
	  forEach(block->insts, i, inst) {
	       enum md_opcode op;
	       MD_SET_OPCODE(op, inst->binInst);
	       if (op == BSR || op == JSR) {
		    PInst funcCall;
		    funcCall = instNewFunctionCall(inst);
		    basicBlockInsertInstAfterInst(block, inst, funcCall);
	       }
	  }
     }
}


void functionRemoveUnneededInstructions(PFunction func)
{
     PBasicBlock block;
     PInst inst;

     GList *l, *k;
     GList *copy;

     forEach(func->basicBlocks, l, block) {
	  bool_t foundBranch = FALSE;
	  copy = g_list_copy(block->insts);
	  forEach(copy, k, inst) {
	       enum md_opcode op;
	       MD_SET_OPCODE(op, inst->binInst);

	       if (!(op == BSR || op == JSR) &&
		   instIsType(inst, F_UNCOND) && 
		   instIsType(inst, F_CTRL)  && 
		   !instIsType(inst, F_INDIRJMP)) {
		    //		    fprintf(stderr, "#removing %d\n", inst->id);
		    
		    basicBlockRemoveInst(block, inst);
		    block->extraInstructions++;
	       }
	       
	       if (instIsType(inst, F_CTRL)) {
		    foundBranch = TRUE;
	       }
	       
 	       if (!foundBranch  && inst->type == OP_TYPE_ALPHA_OP) { 
 		    if (instIsNoop(inst)) { 
 			 basicBlockRemoveInst(block, inst); 
 			 block->extraInstructions++; 
 		    } 
 	       } 
	       
	  }
	  g_list_free(copy);
	  
	  if (block->extraInstructions != 0) {
	       fprintf(stderr, "Removed %d insts from %d\n", block->extraInstructions, block->id);
	  }
     }
}

void functionCleanUpArgPorts (PFunction func)
{
     PBasicBlock top;
     PInst inst;
     PInst topInst;
     GList *l1, *toRemove = 0;
     PArg dst;
     int foundConsumer;
     ArgIndex i;
     fprintf(stderr, "Cleaning up %s\n", func->name);
     top = (PBasicBlock)g_list_first(func->basicBlocks)->data;
     topInst = (PInst)g_list_first(top->insts)->data;
     
     forEachLink(top->insts, l1) {
	  if (toRemove != NULL) { // We can't delete anything until we have iterated past it.
	       top->insts = g_list_remove_link(top->insts, toRemove); 
	  }
	  inst = (PInst)l1->data;
	  if (!inst->argPort) { // Call ports are all at the top.
	       break;
	  }
	  
	  foundConsumer = 0;
	  forEachDst(inst, i, dst) {
	       if (!g_list_empty(dst->consumers)) {
		    foundConsumer = 1;
	       }
	  }
	
	  if (!foundConsumer) {
	       toRemove = l1;
	       //	       fprintf(stderr, "Removing %d\n", inst->id);
	  } else {
	       toRemove = NULL;
	  }
     }     

     if (toRemove != NULL) {
	  top->insts = g_list_remove_link(top->insts, toRemove); 
     }

}

#if(0)

void functionAddArgPorts (PFunction func)
{
     // These are the possible inputs to the function
     // I Regs 15 (fp), 16-21 (a0-a5), 26(ra), 27(pv), 29(gp), 30(sp)
     // F Regs 16-21 (args)
     PBasicBlock top;
     PInst inst;
     PInst topInst;
     int i = 0;

     top = (PBasicBlock)g_list_first(func->basicBlocks)->data;
     topInst = (PInst)g_list_first(top->insts)->data;

     while (portRegs[i] != -1) {
	  md_inst_t nInst = buildCopyInst(portRegs[i]);
/*  	  md_print_insn(nInst, 0, stderr); */ 
/*  	  fprintf(stderr, " %#08x\n", nInst); */
	  inst = instNewFromBin(nInst, topInst->pc);
	  inst->argPort = 1;
	  basicBlockPrependInst(top, inst);
	  i++;
     }
}

#endif

void functionFindFunctionInputs(PFunction f) 
{
     PBasicBlock block;
     PInst inst;
     PArg src;
     GList *l, *k;
     ArgIndex i;
     assert(0);
     //     fprintf(stdout, "Function: %s\n", f->name);
     forEach(f->basicBlocks, l, block) {
	  forEach(block->insts, k, inst) {
	       forEachSrc(inst, i, src) {
		    if (!src->initializedOnce) {
			 if (!g_list_find_custom(f->uniqueInputs, src, pArgsCompare)) { 
			      f->uniqueInputs = g_list_append(f->uniqueInputs, src);
			 }
			 f->inputs = g_list_append(f->inputs, src);
		    }
	       }
	  }
     }

}

void functionComputeVariableSets(PFunction func) 
{
     GList *l;     
     PBasicBlock block;
     SetElement i;

     forEach(func->basicBlocks, l, block) {
	  basicBlockFindUsedVariables(block);
	  setUnion(func->usedVariables, func->usedVariables, block->assignedVariables);
	  setUnion(func->usedVariables, func->usedVariables, block->consumedVariables);
     }

     forEachElement(func->usedVariables, i) {
	  PSet assigningBlocks = setNew(func->lengthBasicBlocks + 1);

	  forEach(func->basicBlocks, l, block) {
	       if (setElementOf(block->assignedVariables, i)) {
		    setAdd(assigningBlocks, block->localID);
	       }
	  }
	  g_hash_table_insert(func->variablesToDefiningBlocks, (gpointer)i, assigningBlocks);
     }
}

void functionAddStupidCopies(PFunction func)
{
     GList *l;     
     PBasicBlock block;
     forEach(func->basicBlocks, l, block) {
	  basicBlockSetJumpFlags(block);
     }
}


void functionSetJumpFlags(PFunction func)
{
     GList *l;     
     PBasicBlock block;
     forEach(func->basicBlocks, l, block) {
	  basicBlockSetJumpFlags(block);
     }
}

void functionComputeNoLocalPredecessorInsts(PFunction func)
{
     GList *l;     
     PBasicBlock block;
     forEach(func->basicBlocks, l, block) {
	  basicBlockComputeNoLocalPredecessorInsts(block);
     }
}

void functionComputeNoLocalSuccessorInsts(PFunction func)
{
     GList *l;     
     PBasicBlock block;
     forEach(func->basicBlocks, l, block) {
	  basicBlockComputeNoLocalSuccessorInsts(block);
     }
}

void functionSplitFunctionCallBlocks(PFunction func)
{
     GList *l, *j, *k;
     PBasicBlock new, block;
     PInst inst;
     PCfgEdge e;
     GList *blockCopy;
     GList *instCopy;

     blockCopy = g_list_copy(func->basicBlocks);
     forEach(blockCopy, l, block) {
	  instCopy = g_list_copy(block->insts);
	  forEach(instCopy, j, inst) {

	       // It doesn't make any sense that this would ever happen.  We
	       // should not have installed any OP_TYPE_FUNCTION_CALL's yet.
	       assert(inst->type != OP_TYPE_FUNCTION_CALL);
	       
	       if(instIsType(inst, F_CALL) || 
 		  inst->type == OP_mnop) {      // Why is this here?
		    assert(j->next == NULL);

		    new = basicBlockNew(NULL, -1);
		    new->isFunctionCallBlock = TRUE;
		    functionAddBasicBlock(func, new);
		    basicBlockRemoveInst(block, inst);
		    basicBlockAppendInst(new, inst);

		    forEach(block->exits, k, e) {
			 cfgEdgeMoveSrc(e, new);
		    }

		    cfgAddEdge(block, new, CFG_EDGE_NORMAL);
	       }
	  }
	  g_list_free(instCopy);
     }
     g_list_free(blockCopy);
}

void functionAvoidEpochConfusion(PFunction func)
{
     // Add some extra mem noops to keep the epoch sequence numbers
     // well-defined.
     
     PBasicBlock block;
     GList *l, *i;

     forEach(func->basicBlocks, i, block) {
	  PCfgEdge edgeToBreak;
	  int entranceCount;
	  entranceCount = g_list_length(block->entrances);

	  if (entranceCount > 1) {
	       forEach(block->entrances, l, edgeToBreak) {
		    if (!edgeToBreak->isBackEdge) {
			 PBasicBlock predecessor = edgeToBreak->src;
			 if (edgeToBreak->src == func->localEntry) {
			      break;
			 }
			 if (g_list_length(predecessor->exits) > 1) {
			      PBasicBlock newBlock = basicBlockNew(NULL, -1);
			      assert(edgeToBreak->type == CFG_EDGE_TAKEN ||
				     edgeToBreak->type == CFG_EDGE_NOT_TAKEN);
			      functionAddBasicBlock(func, newBlock);
			      
			      
			      cfgAddEdge(predecessor, newBlock, edgeToBreak->type);
			      cfgEdgeMoveSrc(edgeToBreak, newBlock);
			      edgeToBreak->type = CFG_EDGE_NORMAL;
			 }
		    }
	       }
	  }     
     }
}

void functionAddMemNoopBlocks(PFunction func)
{
     assert(0);
#if (0)
     PBasicBlock block;
     GList  *i;
     // This is never called


     // This adds an 'else' to an if-then, so that it is a diamond.

     forEach(func->basicBlocks, i, block) {

	  assert(g_list_length(block->exits) <= 2);

	  if (g_list_length(block->exits) == 2) {
	       PCfgEdge e1, e2, edgeToBreak = NULL;
	       e1 = (PCfgEdge)block->exits->data;
	       e2 = (PCfgEdge)block->exits->next->data;
	       
	       if (ptrSetElementOf(e2->dst->reachableNodes[REACHABLE_FORWARD][FALSE], e1->dst)) {
		    edgeToBreak = e1;
	       } else if (ptrSetElementOf(e1->dst->reachableNodes[REACHABLE_FORWARD][FALSE], e2->dst)) {
		    edgeToBreak = e2;
	       }
	       
	       if(edgeToBreak) {
		    PBasicBlock new = basicBlockNew(NULL, -1);
		    functionAddBasicBlock(func, new);
		    
		    cfgAddEdge(block, new, );
		    cfgEdgeMoveSrc(edgeToBreak, new);
	       }
	  }
     }
#endif
}

void functionFixEntryEpochAdvances(PFunction func)
{
     // For the wave advance instructions at the beginning of a function, we
     // only really need them for the function inputs, however, there is stupid
     // spill code that wants to read from other registers.  We will just feed
     // them the control arg.

     GList *l;
     PInst inst;

     forEach(func->enter->insts, l, inst) {
	  if (inst->type == OP_wa) {
	       PArg in;
	       in = instGetSrcByType(inst, INPUT_A);
	       if (!functionRegisterIsFunctionInput(in->reg)) {
		    in->reg = CONTROL_REGISTER;
	       }
	  }
     }
     if (func->localEntry != NULL) {
	  forEach(func->localEntry->insts, l, inst) {
	       if (inst->type == OP_wa) {
		    PArg in;
		    in = instGetSrcByType(inst, INPUT_A);
		    if (!functionRegisterIsFunctionInput(in->reg)) {
			 in->reg = CONTROL_REGISTER;
		    }
	       }
	  }
     }
}

void functionAddMemNoOps(PFunction func)
{
     PBasicBlock block;
     PInst inst;
     GList *l, *i;

     forEach(func->basicBlocks, i, block) {
	  bool_t foundOne = FALSE;
	  if (func->enter == block || 
	      func->localEntry == block || 
	      block->isEpochHead || 
	      block->isLoopTop || 
	      block->isLoopExitBlock ||
	      block->isFunctionCallBlock) {
	       continue;
	  }

	  forEach(block->insts, l, inst) {
	       if (instNeedsEpochTags(inst))
		    foundOne = TRUE;
	  }
	  
	  if (!foundOne) {
	       PInst new = instNew(OP_mnop);
	       instAddSrc(new, CONTROL_REGISTER, INPUT_A);
	       basicBlockPrependInst(block, new);
	  }
     }
}

static PPtrSet functionGrowEpoch(PBasicBlock b)
{
     GList *workList = NULL;

     GList *l;
     PBasicBlock block;
     PCfgEdge e;
     PPtrSet epoch = ptrSetNew();
     
     assert(!b->isInEpoch);


#if (0)
     // This is part of the ugly hack described in FunctionCallState::DoFire()
     // As are the several commeented out bits marked with ***** HACK *****
     // The effect is to no longer put function calls in their own wave.
     if (b->isFunctionCallBlock) { 
 	  //	  fprintf(stderr, "Creating function call epoch %d\n", b->id); 
  	  b->containingEpoch = b;  
  	  b->isInEpoch = TRUE;  
 	  ptrSetAdd(epoch, b);  
 	  return epoch; 
     } 
#endif

     workList = g_list_push(workList, b);
     
     while(workList != NULL) {

	  block = (PBasicBlock)g_list_top(workList);	  
	  workList = g_list_pop(workList);
	  
	  if (block != b) {
	       if (block->isInEpoch || block->isLoopTop  || block->isFunctionCallBlock || block == b->owner->exit) { 
		    continue;
	       }
	  }
	  
	  forEach(block->exits, l, e) {
	       if (!e->isLoopExitEdge && !e->dst->isLoopTop) {
 		    workList = g_list_push(workList, e->dst);
	       }
	  }

	  block->containingEpoch = b;
	  block->isInEpoch = TRUE;
	  fprintf(stderr, "  Adding %d\n", block->id);

	  ptrSetAdd(epoch, block);

     }
     return epoch;
}

static void functionRemoveTreeFromEpoch(PPtrSet epoch, PBasicBlock block)
{

     GList *workList = NULL;
     GList *l;
     PCfgEdge e;
     
     workList =g_list_push(workList, block);

     while(workList != NULL) {
	  block = (PBasicBlock)g_list_top(workList);	  
	  workList = g_list_pop(workList);
	  
	  if (ptrSetElementOf(epoch, block)) {
	       ptrSetRemove(epoch, block);
	       block->isInEpoch = FALSE;
	       block->containingEpoch = NULL;
	       fprintf(stderr, "  Removing %d\n", block->id);

	       forEach(block->exits, l, e) {
		    if (!e->isBackEdge) {
			 workList = g_list_push(workList, e->dst);
		    }
	       }
	  }
     }
}

static void functionTrimEpoch(PPtrSet epoch)
{
     PtrSetIndex i;
     PPtrSet copy;
     PBasicBlock block;
     GList *l;
     PCfgEdge e;
    
     copy = ptrSetCopy(epoch);

     forEachPtrElement(copy, i, block) {
	  forEach(block->entrances, l, e) {
	       if (!block->isEpochHead && !ptrSetElementOf(epoch, e->src)) {
		    functionRemoveTreeFromEpoch(epoch, block);
	       }
	  }
     }

     ptrSetFree(copy);
}

static PPtrSet functionFindEpochEdge(PPtrSet epoch)
{
     PPtrSet fringe = ptrSetNew();
     PtrSetIndex i;
     PBasicBlock block;
     PCfgEdge e;
     GList *l;

     forEachPtrElement(epoch, i, block) {
	  forEach(block->exits, l, e) {
	       if (!ptrSetElementOf(epoch, e->dst)) {
		    ptrSetAdd(fringe, e->src);
	       }
	  }
     }

     return fringe;
}

static PPtrSet functionFindEpochFringe(PPtrSet epoch)
{
     PPtrSet fringe = ptrSetNew();
     PtrSetIndex i;
     PBasicBlock block;
     PCfgEdge e;
     GList *l;

     forEachPtrElement(epoch, i, block) {
	  forEach(block->exits, l, e) {
	       if (!ptrSetElementOf(epoch, e->dst)) {
		    ptrSetAdd(fringe, e->dst);
	       }
	  }
     }

     return fringe;
}

static void functionConstructEpochs(PFunction func)
{
     GList *workList = NULL;
     PBasicBlock block;
     workList = g_list_push(workList, func->enter);

     while(workList != NULL) {
	  PPtrSet epoch;
	  PtrSetIndex i;
	  PPtrSet fringe = ptrSetNew();
	  PBasicBlock fringeBlock;

	  
	  block = (PBasicBlock)g_list_top(workList);	  
	  workList = g_list_pop(workList);

	  if (block->isInEpoch) {
	       continue;
	  }
	  
	  fprintf(stderr, "starting EPOCH for %d\n", block->id);

	  block->isEpochHead = TRUE;
	  
	  epoch = functionGrowEpoch(block);
	  block->epoch = epoch;
	  functionTrimEpoch(epoch);

	  block->epochEdge = functionFindEpochEdge(epoch);
	  fringe = functionFindEpochFringe(epoch);


 	  fprintf(stderr, "FRINGE: "); 
	  forEachPtrElement(fringe, i, fringeBlock) { 
 	       fprintf(stderr, "%d ", fringeBlock->id);  
	       workList = g_list_push(workList, fringeBlock); 
	  } 
 	  fprintf(stderr, "\n");  

	  
	  ptrSetFree(fringe);
     }
     
     {
	  GList *i;
	  forEach(func->basicBlocks, i, block) {
	       if (block->isEpochHead) {
		    PBasicBlock b;
		    PtrSetIndex j;
		    fprintf(stderr, "%s:\n", block->label);
		    forEachPtrElement(block->epoch, j, b) { 
			 fprintf(stderr, "\t%s", b->label);
		    }
		    fprintf(stderr, "\n");
	       }
	  }
     }
}


static void functionAssignMemoryOrder(PFunction func, PRippleChunk prev, PRippleChunk cur, PBasicBlock nextBlock, PPtrSet traversed, int lookingToStop)
{
     GList *l;
     PRippleChunk next;
     PCfgEdge e;
     PRippleChunk first = NULL;

     if (cur && (!strcmp(cur->owner->label, "B464") ||
		 !strcmp(cur->owner->label, "B12855")
		 )) {
	  e = NULL;
     }
     /*      fprintf(stderr, "VISITING %d - %d { \n", nextBlock->id, lookingToStop); */
     forEach(nextBlock->rippleChunks, l, next) {
	  if (!first) {
	       first = next;
	  }
     
	  if (cur) {
	       if (prev) {
		    assert(cur && next);
			 
		    if (cur->epochOrderPrev == EPOCH_ORDER_UNDEFINED) {
			 cur->epochOrderPrev = prev->epochOrder;

			 cur->possiblePrevMemOps = ptrSetNew();
			 ptrSetAdd(cur->possiblePrevMemOps, prev);

		    } else if (cur->epochOrderPrev != prev->epochOrder) {
			 cur->epochOrderPrev = EPOCH_ORDER_UNKNOWN;
			 ptrSetAdd(cur->possiblePrevMemOps, prev);
			 //			      fprintf(stderr, "CONFUSED at %d prev\n", cur->id);
		    }
	       } else {
		    cur->epochOrderPrev = EPOCH_ORDER_UNKNOWN;
		    //			 fprintf(stderr, "CONFUSED BADLY at %d prev\n", cur->id);
	       }
		    
	       if (cur->epochOrderNext == EPOCH_ORDER_UNDEFINED) {
		    cur->epochOrderNext = next->epochOrder;
		    cur->possibleNextMemOps = ptrSetNew();
		    ptrSetAdd(cur->possibleNextMemOps, next);
	       } else if (cur->epochOrderNext != next->epochOrder) {
		    cur->epochOrderNext = EPOCH_ORDER_UNKNOWN;
		    ptrSetAdd(cur->possibleNextMemOps, next);
		    //			 fprintf(stderr, "CONFUSED at %d next\n", cur->id);
	       }

	       //		    fprintf(stderr, "SETING %d to < %d %d %d >\n", cur->id, cur->epochOrderPrev, cur->epochOrder, cur->epochOrderNext);


	       if (next->epochOrderPrev == EPOCH_ORDER_UNDEFINED) {
		    next->epochOrderPrev = cur->epochOrder;
		    next->possiblePrevMemOps = ptrSetNew();
		    ptrSetAdd(next->possiblePrevMemOps, cur);
	       } else if (next->epochOrderPrev != cur->epochOrder) {
		    next->epochOrderPrev = EPOCH_ORDER_UNKNOWN;
		    ptrSetAdd(next->possiblePrevMemOps, cur);
	       }
	  }
	  

	  if (lookingToStop) {
	       if (lookingToStop == 1) {
		    /* 			 fprintf(stderr, "STOPPING\n"); */
		    /* 			 fprintf(stderr, "}\n"); */
		    return;
	       } else {
		    /* 			 fprintf(stderr, "FOUND ONE TO STOP\n"); */
		    lookingToStop--;
	       }
	  }
	       
	  prev = cur;
	  cur = next;
     } 

     forEach(nextBlock->exits, l, e) {
	  if (!ptrSetElementOf(traversed, e)) { 
	       if (!lookingToStop) {
		    //			 fprintf(stderr, "TRAVERSED NOT LOOKING %d -> %d\n", e->src->id, e->dst->id);
		    ptrSetAdd(traversed, e);         			 
	       } else {
		    //			 fprintf(stderr, "TRAVERSED     LOOKING %d -> %d\n", e->src->id, e->dst->id);
	       }
	       
	       if (e->isBackEdge) {
		    functionAssignMemoryOrder(func, prev, cur, e->dst, traversed, 2);
	       } else {
		    functionAssignMemoryOrder(func, prev, cur, e->dst, traversed, lookingToStop);
	       }
	  } else {
	       if (lookingToStop) {
		    functionAssignMemoryOrder(func, prev, cur, e->dst, traversed, lookingToStop);
	       }
	  }
     }
     
     if (nextBlock->isFunctionCallBlock && first) {
	  first->epochOrderPrev = EPOCH_ORDER_UNKNOWN;
     }

     if (func->exit == nextBlock) {
	  assert(prev);
	  if (cur->epochOrderPrev == EPOCH_ORDER_UNDEFINED) {
	       cur->epochOrderPrev = prev->epochOrder;
	  } else if (cur->epochOrderPrev != prev->epochOrder) {
	       cur->epochOrderPrev = EPOCH_ORDER_UNKNOWN;
	  }
	  
	  assert(cur);
	  cur->epochOrderNext =  1;//EPOCH_ORDER_UNKNOWN;

	  //	  simPrint(stderr, "SETING %d to < %d %d %d >\n", cur->id, cur->epochOrderPrev, cur->epochOrder, cur->epochOrderNext);

     }
/*      fprintf(stderr, "}\n"); */
			 
}


static void functionComputeIndependenceNumbers(PFunction func, PBasicBlock epochHead, PBasicBlock b, int lastStoreNumber, bool_t first)
{
     if (b) {
	  fprintf(stderr, "     %s\n", b->label);
	  PCfgEdge e;
	  GList *i;
	  GList *l;
	  PRippleChunk chunk;
	
	  forEach(b->rippleChunks, l, chunk) {
	       PInst inst;
	       if (first) {
		    chunk->lastStoreSequenceNumber = 0;
		    lastStoreNumber = 1;
		    first = FALSE;
		    continue;
	       }
	       
	       inst = (PInst)chunk->insts->elements->data;
	       
	       chunk->lastStoreSequenceNumber = MAX(chunk->lastStoreSequenceNumber, lastStoreNumber);
	       
	       if (instIsType(inst, F_STORE)) {
		    assert(chunk->epochOrder > chunk->lastStoreSequenceNumber);
		    lastStoreNumber = chunk->epochOrder;
	       }
	  }
	  
	  forEach(b->exits, i, e) {
	       if (ptrSetElementOf(epochHead->epoch, e->dst) && !e->isBackEdge)  {
		    functionComputeIndependenceNumbers(func, epochHead, e->dst, lastStoreNumber, first);
	       }
	  }
     }
}

static void functionNumberMemOpsInEpoch(PFunction func, PBasicBlock b)
{
     GList *workList = NULL;
     PBasicBlock block;
     int currentNumber = 1;
     PRippleChunk ripple;
     GList *l, *j;
     PPtrSet visited = ptrSetNew();
     
     workList = g_list_push(workList, b);

     while (workList != NULL) {
	  GList *k;
	  PCfgEdge e;

	  block = (PBasicBlock)g_list_head(workList);	  
	  workList = g_list_dequeue(workList);
	  
	  ptrSetAdd(visited, block);
	  
	  forEach(block->rippleChunks, l, ripple) {
	       ripple->epochOrder = currentNumber++;
	  }
	  
	  forEach(block->exits, k, e) {
	       if (ptrSetElementOf(b->epoch, e->dst))  {
		    // This ensures that the ordering topological
		    PCfgEdge entrance;
		    bool_t entrancesVisited = TRUE;
		    forEach(e->dst->entrances, j, entrance) {
			 if (!ptrSetElementOf(visited, entrance->src)) {
			      entrancesVisited = FALSE;
			      break;
			 }
		    }
		    
		    if (!ptrSetElementOf(visited, e->dst) && entrancesVisited) {
			 workList = g_list_enqueue(workList, e->dst);
		    }
	       }
	  }
     }

     assert(b->epoch);

}

void functionSetControllingCoordinator(PFunction func)
{
     PBasicBlock block;
     GList *l, *i;
     PInst inst;
 
     forEach(func->basicBlocks, i, block) {
	  if (block->isEpochHead) {
	       PtrSetIndex p;
	       PBasicBlock eBlock;
	       PInst e;
	       forEach(block->insts, l, e) {
		    if (e->type == OP_wa) {
			 break;
		    }
	       }
	       assert(e);
	       forEachPtrElement(block->epoch, p, eBlock) {
		    forEach(eBlock->insts, l, inst) {
			 inst->controllingEpochCoordinate = e;
		    }	       
	       }
	  }

     }
     
}

void functionLabelCanonicalEpochAdvances(PFunction func)
{
     PBasicBlock block;
     GList *l, *i;
     PInst inst;
     forEach(func->basicBlocks, i, block) {

	  PInst canonical = NULL;
	  int advanceCount = 0;
	  
	  // Mark the first advance as canonical and label all of them with how
	  // many of them there are.  We only need this information so we can
	  // reclaim some resources in the migration in kahuna
	  forEach(block->insts, l, inst) {
	       if (inst->type == OP_wa) {
		    if (canonical == NULL) {
			 canonical = inst;
			 inst->isCanonicalEpochAdvance = TRUE;
		    }
		    advanceCount++;
	       }
	  }

	  forEach(block->insts, l, inst) {
	       if (inst->type == OP_wa) {
		    inst->waveAdvanceCount = advanceCount;
	       }
	  }
     }
}

void functionNumberMemOps(PFunction func)
{
     PBasicBlock block;
     GList *l;
     
     PPtrSet traversed;
	  
     forEach(func->basicBlocks, l, block) {
	  if (block->isEpochHead) {
	       functionNumberMemOpsInEpoch(func, block);
	       fprintf(stderr, "Wave at %s\n", block->label);
	       functionComputeIndependenceNumbers(func, block, block, -1, TRUE);
	  }
     }

     traversed = ptrSetNew();
     
     functionAssignMemoryOrder(func, NULL, NULL, func->enter, traversed, 0);
     
     ptrSetFree(traversed);
}

void functionEpochSanity(PFunction func)
{
     int blockCount = 0;
     int epochSizeTotal = 0;
     PBasicBlock block;
     GList *i;
     PInst inst;
     GList *l;
     int skip = 0;
     forEach(func->basicBlocks, l, block) {
	  if (block == func->localEntry) {
	       continue;
	  }
	  if (block->handWaveNumber) {
	       continue;
	  }
	  if (!block->isDead) {
	       blockCount++;
	  }
	  if (block->isEpochHead) {
	       epochSizeTotal+=ptrSetSize(block->epoch);
	  }
	  if (!block->isInEpoch) {
	       fprintf(stderr, "In %s: ", block->owner->name);
	       assert(0);
	  }
	  
	  forEach(block->insts, i, inst) {
	       if(instNeedsEpochTags(inst)) {
		    
  		    assert(inst->epochOrderPrev != EPOCH_ORDER_UNDEFINED && 
  			   inst->epochOrderNext != EPOCH_ORDER_UNDEFINED);  

	       }
	  }
     }
     //     assert(blockCount - skip == epochSizeTotal);

}

void functionBuildEpochs(PFunction func)
{
     functionConstructEpochs(func);
}

void functionPruneFunctionCallOutputs(PFunction func)
{
     PBasicBlock block;
     GList *l, *i;
     PInst inst;

     forEach(func->basicBlocks, i, block) {
	  forEach(block->insts, l, inst) {
	       if(inst->type == OP_TYPE_FUNCTION_CALL) {
		    instPruneFunctionCall(inst);
	       }
	  }
     }     
}

void functionSetFirstInstBits(PFunction func)
{
     PBasicBlock block;
     GList *l, *i;
     PInst inst;

     forEach(func->basicBlocks, i, block) {
	  forEach(block->insts, l, inst) {
	       inst->isFirstInBlock = TRUE;
	       break;
	  }
     }     
}

void functionCollapseBasicBlocks(PFunction func) {
     PBasicBlock block;
     GList *l, *i;
     PCfgEdge e;

     forEach(func->basicBlocks, i, block) {
	  if (g_list_length(block->exits) == 1) {
	       PCfgEdge possibleBackEdge;
	       bool_t goodDeal = TRUE;
	       bool_t isWierdoGlobalPointerBlock; // We need to protect the
						  // first two blocks after
						  // start because they
						  // represent the two possible
						  // call sites for the
						  // function.

	       e = (PCfgEdge)block->exits->data;

	       isWierdoGlobalPointerBlock = (block == g_list_nth_data(func->basicBlocks, 1) || 
					     block == g_list_nth_data(func->basicBlocks, 2));

	       forEach(block->entrances, l, possibleBackEdge) {
		    if (possibleBackEdge->src == e->dst) {
			 goodDeal = FALSE;
			 break;
		    }
	       }

	       if (g_list_length(e->dst->entrances) == 1 && 
		   !e->isBackEdge &&
		   !e->dst->isFunctionCallBlock &&
		   !e->src->isFunctionCallBlock &&
		   e->src != func->enter &&
		   e->src != func->localEntry &&
		   e->dst != func->exit &&
/* 		   !e->dst->isLoopExitBlock && */
/* 		   !e->src->isLoopExitBlock && */
		   goodDeal &&
		   !isWierdoGlobalPointerBlock
		   ) {
		    assert(!e->isLoopExitEdge);
		    basicBlockMerge(e->src, e->dst);
	       }
	  }
     }
}


static LatencyType functionInstIsZeroLatency(PInst inst)
{
     PArg input;
     PArg output;
     LatencyType latencyType = 0;

     return 0;

     switch(inst->type) {
/*      case OP_TYPE_EPOCH_COORDINATOR: */
/* 	  latencyType |= LATENCY_EPOCH_COORDINATORS; */
/* 	  break; */
     case OP_wa:
	  latencyType |= LATENCY_ALL_EPOCH_ADVANCES;
	  output = instGetDstByType(inst, OUTPUT_A);
	  if (((PArg)output->consumers->data)->owner->type != OP_wa) {
	       latencyType |= LATENCY_EASY_EPOCH_ADVANCES;
	  }
	  break;
     case OP_TYPE_ALPHA_OP:
	  if (instIsType(inst, F_COND)) {
	       latencyType |= LATENCY_BRANCHES;
	  }
     case OP_mnop:
     case OP_src:
     case OP_sink:
     case OP_TYPE_FUNCTION_CALL:
	  latencyType |= LATENCY_ALPHA_OPS;
	  break;
     case OP_rho:
	  input = instGetSrcByType(inst, INPUT_A);
	  latencyType |= LATENCY_ALL_RHO_FUNCS;
	       
	  if (((PArg)input->producers->data)->owner->type != OP_rho) {
	       latencyType |= LATENCY_EASY_RHO_FUNCS;
	  }
	  break;
     case OP_is:
     case OP_isret:
     case OP_is_pc:
     case OP_fanout:
     case OP_wtd:
     case OP_cnst:
     default:
	  break;
     }

     return latencyType;
}


void functionSetInstructionLatencies(PFunction func)
{
     GList *l, *i;
     PBasicBlock block;
     PInst inst;

     forEach(func->basicBlocks, l, block) {
	  forEach(block->insts, i, inst) {
	       inst->latencyType |= functionInstIsZeroLatency(inst);
	  }
     }
}

void functionBuildRippleChunks(PFunction f)
{
     PBasicBlock block;
     GList *l;

     forEach(f->basicBlocks, l, block) {
	  basicBlockBuildRippleChunks(block);
     }
}

void functionSetInstructionEpochTags(PFunction f)
{
     GList *l, *i;
     PBasicBlock block;
     PInst inst;

     forEach(f->basicBlocks, l, block) {
	  forEach(block->insts, i, inst) {
	       if (!block->handWaveNumber) {
		    if (inst->rippleChunk != NULL) {
			 inst->epochOrderPrev = inst->rippleChunk->epochOrderPrev;
			 inst->epochOrder	 = inst->rippleChunk->epochOrder; 
			 inst->epochOrderNext = inst->rippleChunk->epochOrderNext;
			 inst->rippleSize     = inst->rippleChunk->rippleSize;
			 inst->lastStoreSequenceNumber = inst->rippleChunk->lastStoreSequenceNumber;
		    }
	       }
	  }
     }
}

void functionTokenizeMemoryOps(PFunction f)
{
     GList *l, *i;
     PBasicBlock block;
     PInst inst;

     forEach(f->basicBlocks, l, block) {
	  GList * copy = g_list_copy(block->insts);

	  forEach(copy, i, inst) {
	       if (instIsType(inst, F_STORE)) {
		    instAddSrc(inst, CONTROL_REGISTER, INPUT_C);
		    instAddDst(inst, CONTROL_REGISTER, OUTPUT_C);
	       } else if (instIsType(inst, F_LOAD)) {
		    instAddSrc(inst, CONTROL_REGISTER, INPUT_C);
		    PArg loadDstArg = instGetDstByType(inst, OUTPUT_A);
		    if (loadDstArg->reg == 31 ||
			loadDstArg->reg == 63) {
			 basicBlockRemoveInst(block, inst);
		    } else {
			
			 PInst copy = instNew(OP_nop);
			 //PInst copy = instNew(OP_identity);
			 copy->isTokenizeNop = TRUE;
			 instAddSrc(copy, loadDstArg->reg, INPUT_A);
			 instAddDst(copy, CONTROL_REGISTER, OUTPUT_A);
			 basicBlockInsertInstAfterInst(block, inst, copy);


			 fprintf(stderr, "FOOOOO ");
			 instPrint(stderr, inst);
			 fprintf(stderr, "FOOOOO ");
			 instPrint(stderr, copy);		
		    }    
	       }
	  }
     }     
}
