#ifndef FUNCTION_INCLUDED
#define FUNCTION_INCLUDED

#include <glib.h>
#include <stdio.h>
#include "types.h"
#include "set.h"

typedef struct _FUNCTION *PFunction;

//#include "substrate.h"
#include "program.h"
#include "basicBlock.h"
#include "inst.h"


typedef enum {LATENCY_NONE = 0,
              LATENCY_EASY_RHO_FUNCS = 1, 
              LATENCY_ALL_RHO_FUNCS = 2,
              LATENCY_BRANCHES = 4,
              LATENCY_EPOCH_COORDINATORS = 8,
              LATENCY_ALL_EPOCH_ADVANCES = 16,
              LATENCY_EASY_EPOCH_ADVANCES = 32,
              LATENCY_ALPHA_OPS = 64,
              LATENCY_ZAPPED_EPOCH_MANAGEMENT = 128} LatencyType;


extern int stubCount;
typedef struct _FUNCTION {
     GList *basicBlocks;
     char *name;
     int lengthBasicBlocks;
     PProgram owner;

     bool_t isMain;

     GList *functionsCalled;
     GList *inputs; // All input consumer PArgs
     GList *uniqueInputs;  // One input conusmer PArg per consumed register
     
     PSet usedVariables;  // Set of all variables used. 
     GHashTable * variablesToDefiningBlocks;

     
     PBasicBlock localEntry;
     PBasicBlock remoteEntry;

     PBasicBlock enter;
     PBasicBlock exit;
     PBasicBlock returnBlock;
     PC pc;
     //     ExecModelFunctionData execModelData;
     int stubID;
     int localStubID;
     uint compileAction;
} Function;  

extern int functionOutputType[];

#define FUNC_OWNER_PROGRAM(f) ((f)->owner)
#define FUNC_BASIC_BLOCK_BY_LOCAL_ID(func, localID) ((PBasicBlock)g_list_nth_data((func)->basicBlocks, (localID) - 1))
#define FUNC_LAST_BASIC_BLOCK(func) ((PBasicBlock)(g_list_last((func)->basicBlocks)->data))

PFunction functionNew(char * name, int action);
void functionFree(PFunction f);
void functionPrint(FILE *file, PFunction f, bool_t verbose);

void functionPrintControlDistribution(PFunction func, FILE *f);
void functionClearPlacement(PFunction f);
void functionSanity(PFunction func);

void functionAddBasicBlock(PFunction func, PBasicBlock block);
void functionRemoveBasicBlock(PFunction func, PBasicBlock block);
// Compiler transformations 
void functionAddArgPorts (PFunction f);
void functionAddCallPorts (PFunction f);
void functionCleanUpArgPorts (PFunction f);
void functionCleanUpFunction(PFunction func);
void functionFindFunctionInputs(PFunction f) ;
void functionFindFunctionCallParams(PFunction func);
void functionAddJumpPorts (PFunction func);
void functionAddReturnPorts(PFunction func);

void functionComputeNoLocalSuccessorInsts(PFunction func);
void functionComputeNoLocalPredecessorInsts(PFunction func);

void functionSetControlSources(PFunction func);

void functionComputeVariableSets(PFunction func);

PBasicBlock functionFindBasicBlockByLocalID(PFunction func, int localID);

void functionInsertSpecialSources (PFunction func);
void functionRemoveUnneededInstructions(PFunction func);

void functionRemoveUselessSourceInstructions(PFunction func);
void functionCheckForNastiness(PFunction func);

void functionFixLocalInvocationEpoch(PFunction func);
void functionFixLocalInvocationInsts(PFunction func);
void functionMakeLocalInvocation(PFunction func);
void functionRemoveAlignmentBlocks(PFunction func);
void functionSerializeInstructions(PFunction func);
void functionFixConditionalMoves(PFunction func);
void functionSetInstructionLatencies(PFunction func);
void functionAddReturnBlock(PFunction func);
void functionSplitFunctionCallBlocks(PFunction func);
void functionBuildEpochs(PFunction func);
void functionEpochSanity(PFunction func);
void functionFixEntryEpochAdvances(PFunction func);
void functionAddMemNoOps(PFunction func);
void functionAddMemNoopBlocks(PFunction func);
void functionPruneFunctionCallOutputs(PFunction func);
void functionAvoidEpochConfusion(PFunction func);
void functionSetFirstInstBits(PFunction func);

void functionInstallFunctionCalls(PFunction func);
void functionInstallIndirectSendFunctionCalls(PFunction func);
void functionAddReturnSinks(PFunction func);
void functionFixReturns(PFunction func);

void functionRemoveUselessSpecialSources(PFunction func);
void functionNumberMemOps(PFunction func);
void functionPrintAnnotatedMap(PFunction f, FILE * file) ;
void functionCollapseBasicBlocks(PFunction func) ;
void functionSetLocalIDs(PFunction func);

void functionRemoveMarkedDeadBlocks(PFunction func);
void functionMarkEntries(PFunction func);
void functionBuildRippleChunks(PFunction func);

void functionLabelCanonicalEpochAdvances(PFunction func);
void functionSetInstructionEpochTags(PFunction f);

void functionSetControllingCoordinator(PFunction func);

void functionCompressWaveAdvance(PFunction func);
void functionRhoFold(PFunction func);
void functionMakeStub(PFunction func);

#endif


