#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <glib.h>
#include <string.h>

#include "inst.h"
#include "types.h"
#include "machine.h"
#include "list.h"
#include "epoch.h"
#include "ptrSet.h"
#include "wsArch.h"
#include "lcc.h"
#include "dot.h"


#include "instDefs.c"

void instMoveSrc(PArg arg,
		 PInst to)
{
     instRemoveSrc(arg->owner, arg);
     arg->owner = to;
     to->srcs = g_list_append(to->srcs, arg);
}

void instMoveDst(PArg arg,
		 PInst to)
{
     instRemoveDst(arg->owner, arg);
     arg->owner = to;
     to->dsts = g_list_append(to->dsts, arg);
}

PArg instAddDst(PInst inst, 
		int reg, 
		OutputArgType outputArgType)
{
     PArg n;
     n = argNew(reg, INPUT_INVALID, outputArgType, inst);
     inst->dsts = g_list_append(inst->dsts, n);
     return n;
}

PArg instAddSrc(PInst inst, 
		int reg, 
		InputArgType inputArgType)
{
     PArg n;
     n = argNew(reg, inputArgType, OUTPUT_INVALID, inst);
     inst->srcs = g_list_append(inst->srcs, n);
     return n;
     
}

void instRemoveDst(PInst inst,
		   PArg dst)
{
     inst->dsts = g_list_remove(inst->dsts, dst);
}

void instRemoveSrc(PInst inst,
		   PArg src)
{
     inst->srcs = g_list_remove(inst->srcs, src);
}

PArg instGetSrcByType(PInst inst, InputArgType type) 
{
     GList *i;
     PArg src;
     forEach(inst->srcs, i, src) {
	  if (src->inputType == type) {
	       return src;
	  }
     }
     return NULL;
}



PArg instGetDstByType(PInst inst, OutputArgType type) 
{
     GList *i;
     PArg dst;
     forEach(inst->dsts, i, dst) {
	  if (dst->outputType == type) {
	       return dst;
	  }
     }
     return NULL;
}


void instInit(PInst n, InstType type) 
{
     static int instNum = 0;
     
     n->id = instNum++;
     n->type = type;
     n->binInst = -1;
     n->pc = -1;
     n->targetPC = -1;
     n->locx = -1;
     n->locy = -1;
     n->regNum = -1;
     n->epochOrder = EPOCH_ORDER_UNDEFINED;
     n->epochOrderPrev = EPOCH_ORDER_UNDEFINED;
     n->epochOrderNext = EPOCH_ORDER_UNDEFINED;
     n->rippleSize = 0;
}

PInst instNew(InstType type) 
{
     PInst n = (PInst)calloc(sizeof(Inst), 1);
     instInit(n, type);
     return n;
}

void instInitFromBin(PInst n, md_inst_t inst, PC pc)
{
     enum md_opcode op;
     PC target;
     PArg dst;
     ArgIndex i;
     uint whatHappened;

     instInit(n, OP_TYPE_ALPHA_OP);
     MD_SET_OPCODE(op, inst);
     n->regNum = -1;

     if(op == PAL_RDUNIQ ||
	op == PAL_WRUNIQ ||
	op == CALL_PAL ||
	op == PAL_CALLSYS
	) {
	  // for these we ignore them and assume we are in a function that is
	  // going to get stubbed.
     } else {
	  // We have to do this in two stages because using the stock RA, RB, RC
	  // macros doesn't tell us which field in the instruction the instruction
	  // uses, they tell us which register is the in the slot.  If an
	  // instruction has a source that is also a destination, we can't tell the
	  // difference between the two. 
	  //
	  // To remedy this we first redefine RA, RB, and RC to let us track which
	  // slots are being used, and then fix up the src and dst registers. 


	  // A little bit of paranoia.  These asserts ensure that when we restore
	  // RA, RB, RC to their original definitions, we get the right ones.
	  assert(RA == EXTRACT_RA(inst));
	  assert(RB == EXTRACT_RB(inst));
	  assert(RC == EXTRACT_RC(inst));
	  switch (op) {

#define IN_EXEC_MODEL

#define DGPR(N)                (N)
#define DFPR(N)                (N)
#define DNA                    (-1)
#define DUNIQ                    (UNIQ_REGISTER)
#define DFPCR                    (FPCR_REGISTER)
#undef RA
#undef RB
#undef RC
#define RA                   (1)
#define RB                   (2)
#define RC                   (3)

#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3)	\
	       case OP:						\
		    assert(O2 == DNA);				\
	       assert(I3 == DNA);				\
	       if (O1 == RA) {					\
		    instAddDst(n, -1, OUTPUT_A);		\
	       } else if (O1 == RC) {				\
		    instAddDst(n, -1, OUTPUT_A);		\
	       } else if (O1 != DNA) {				\
		    /*assert(0)*/;				\
	       }						\
	       if (I1 == RA && I2 == RB) {			\
		    instAddSrc(n, -1, INPUT_A);			\
		    instAddSrc(n, -1, INPUT_B);			\
		    whatHappened = 0;				\
	       }						\
	       if (I1 == DNA && I2 == RB) {			\
		    instAddSrc(n, -1, INPUT_A);			\
		    whatHappened = 1;				\
	       }						\
	       if (I1 == RA && I2 == DNA) {			\
		    instAddSrc(n, -1, INPUT_A);			\
		    whatHappened = 2;				\
	       }						\
	       if (I1 == DNA && I2 == DNA) {			\
		    whatHappened = 3;				\
	       }						\
	       if ((MD_OP_FLAGS(OP) & F_COND) != 0 ||		\
		   (OP == BR) ||				\
		   (OP == BSR)) {				\
		    target = pc + (SEXT21(TARG) << 2) + 4;	\
	       } else target = -1;				\
	       break;

#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)					\
	       case OP:							\
		    panic("attempted to execute a linking opcode");
#define CONNECT(OP)
#define DECLARE_FAULT(FAULT)			\
	       { fault = (FAULT); break; }
#include "machine.def"
	  default:
	       target = -1;
	  }
     
	  // Restore the old definitions (From alpha.h)
#undef RA
#undef RB
#undef RC
#define RA		EXTRACT_RA(inst)		/* reg source #1 */
#define RB		EXTRACT_RB(inst)		/* reg source #2 */
#define RC		EXTRACT_RC(inst)       		/* reg dest */

	  switch (op) {
#undef DGPR
#undef DFPR

#undef DUNIQ
#undef DFPCR	 
#define DGPR(N)                (N)
#define DFPR(N)                (32 + (N))
#define DNA                    (-1)
#define DUNIQ                    (1+32+32)
#define DFPCR                    (0+32+32)

#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3)		\
	       case OP:							\
		    {							\
			 uint count = 0;				\
			 switch(whatHappened) {				\
			 case 0:					\
			      if (I1 == 31 ||				\
				  I1 == 63) {				\
				   instGetSrcByType(n, INPUT_A)->reg = -1; \
				   instGetSrcByType(n, INPUT_A)->isZero = TRUE; \
			      } else {					\
				   instGetSrcByType(n, INPUT_A)->reg = I1; \
			      }						\
			      if (I2 == 31 ||				\
				  I2 == 63) {				\
				   instGetSrcByType(n, INPUT_B)->reg = -1; \
				   instGetSrcByType(n, INPUT_B)->isZero = TRUE; \
			      } else {					\
				   instGetSrcByType(n, INPUT_B)->reg = I2; \
			      }						\
			      break;					\
			 case 1:					\
			      if (I2 == 31 ||				\
				  I2 == 63) {				\
				   instGetSrcByType(n, INPUT_A)->reg = -1; \
				   instGetSrcByType(n, INPUT_A)->isZero = TRUE; \
			      } else {					\
				   instGetSrcByType(n, INPUT_A)->reg = I2; \
			      }						\
			      break;					\
			 case 2:					\
			      if (I1 == 31 ||				\
				  I1 == 63) {				\
				   instGetSrcByType(n, INPUT_A)->reg = -1; \
				   instGetSrcByType(n, INPUT_A)->isZero = TRUE; \
			      } else {					\
				   instGetSrcByType(n, INPUT_A)->reg = I1; \
			      }						\
			      break;					\
			 case 3:					\
			      break;					\
			 }						\
									\
			 count = 0;					\
			 forEachDst(n, i,dst) {				\
			      dst->reg = O1;				\
			      count++;					\
			 }						\
			 assert(count <= 1);				\
		    }							\
	       								\
	       break;
	       
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)					\
	       case OP:							\
		    panic("attempted to execute a linking opcode");
#define CONNECT(OP)
#define DECLARE_FAULT(FAULT)			\
	       { fault = (FAULT); break; }
#include "machine.def"
	  default:
	       break;
	  }
     }
     
     n->wsOpCode =  SSToWSMap[op];
     
     n->binInst = inst;
     n->pc = pc;
     n->targetPC = target;

     n->isLoad = instIsType(n, F_LOAD); 

}

PInst instNewFromBin(md_inst_t inst, PC pc)
{
     PInst n = (PInst) calloc(sizeof(Inst), 1);
     instInitFromBin(n, inst, pc);
     n->isFreeable = TRUE;
     return n;
}

PInst instNewFunctionCall(PInst callingInst) 
{
     // THis code is finish.
     assert (0);
     return NULL;
/* #if (0) */

/*      PInst new; */
/*      int i; */
/*      PBasicBlock callDstBlock; */
/*      enum md_opcode op; */



/*      new = instNew(OP_TYPE_FUNCTION_CALL); */

/*      new->callingInstruction = callingInst; */
/*      for(i = INPUT_RET; i <= INPUT_FPA5; i++) { */
/* 	  instAddSrc(new, ARG_TYPE_TO_REG[i], i); */
/*      } */

/*      i = 0;  */
/*      while (functionOutputType[i] != -1) { */
/* 	  instAddDst(new, ARG_TYPE_TO_REG[functionOutputType[i]], functionOutputType[i]); */
/* 	  i++; */
/*      } */
     
/*      assert(cfgOptions.performEpochNumbering); */

/*      //     instAddSrc(new, CONTROL_REGISTER, INPUT_CONTROL); */
/*      //     instAddDst(new, CONTROL_REGISTER, OUTPUT_CONTROL); */

/*      if (cfgOptions.serialize) { */
/* 	  instAddSrc(new, SERIALIZE_REGISTER, INPUT_SERIALIZE); */
/* 	  instAddDst(new, SERIALIZE_REGISTER, OUTPUT_SERIALIZE); */
/*      } */

/* /\*      if (cfgOptions.includeWaveCoordinate) { *\/ */
/* /\* 	  instAddSrc(new, EPOCH_COORDINATION_REGISTER, INPUT_EPOCH_COORDINATION); *\/ */
/* /\* 	  instAddDst(new, EPOCH_COORDINATION_REGISTER, OUTPUT_EPOCH_COORDINATION); *\/ */
/* /\*      } *\/ */

/*      MD_SET_OPCODE(op, callingInst->binInst); */

/*      if (op == JSR) { */
/* 	  //	  instAddSrc(new, JMP_DST_REGISTER, INPUT_JMP_DST); */
/* 	  //	  instAddDst(callingInst, JMP_DST_REGISTER, OUTPUT_JMP_DST); */
/* 	  callDstBlock = NULL; */
/*      } else { */
/* 	  callDstBlock = programFindBasicBlockByPC(callingInst->owner->owner->owner, callingInst->targetPC); */
/*      } */

/*      new->addressToCall = callingInst->targetPC; */
/*      if (callDstBlock != NULL) { */
/* 	  new->functionToCall = callDstBlock->owner; */
/*      } */

/*      return new; */
/* #endif */
}

void instPruneFunctionCall(PInst inst)
{
     GList *l;
     PInst sink;
     GList *i;
     PArg dst;
     GList *c1;
     PPtrSet sinkRegisters = ptrSetNew();

     if (inst->functionToCall == NULL) {
	  return;
     }

     forEach(inst->functionToCall->exit->insts, l, sink) {
	  if (sink->type == OP_sink) {
	       PArg input = instGetSrcByType(sink, INPUT_A);
	       //	       fprintf(stderr, "FOUND %d:%d\n", inst->id,input->reg);
	       ptrSetAdd(sinkRegisters, (void *)input->reg);
	  }
     }
     
     ptrSetAdd(sinkRegisters, (void*)CONTROL_REGISTER);
     //     ptrSetAdd(sinkRegisters, (void*)EPOCH_COORDINATION_REGISTER);

     if (cfgOptions.serialize) {
	  ptrSetAdd(sinkRegisters, (void*)SERIALIZE_REGISTER);
     }

     c1 = g_list_copy(inst->dsts);
     forEach(c1, i, dst) {
	  if (!ptrSetElementOf(sinkRegisters, (void*)dst->reg)) {
	       GList *c2;
	       GList *j;
	       PArg consumer;
	       
	       //	       fprintf(stderr, "PRUNING %d:%d ", inst->id, dst->reg);
	       
	       instRemoveDst(inst, dst);

	       c2 = g_list_copy(dst->consumers);
	       forEach(c2, j, consumer) {
		    //		    fprintf(stderr, " %d->", consumer->owner->id);
		    dfgRemoveEdge(dst, consumer);
		    dfgUpdateSource(consumer);
		    //		    fprintf(stderr, " %d", consumer->owner->id);
	       }
	       g_list_free(c2);

//	       fprintf(stderr, "\n");
	  }
     }
     g_list_free(c1);
}

void instFree(PInst inst)
{
     assert(inst);
     if(inst->isFreeable){
	  free(inst);
     }
}

void instSanity(PInst inst)
{
     ArgIndex i;
     PArg arg;

     forEachSrc(inst, i, arg) {
	  argSanity(arg);
     }

     forEachDst(inst, i, arg) {
	  argSanity(arg);
     }

     if (inst->type == OP_wa) {
	  assert(inst->waveAdvanceAmount == 0);
     }
}


void instPrintAnnotatedMap(PInst inst, FILE *f)
{
     fprintf(f, "\t\t");
     fprintf(f, "0x%08llx\t", inst->pc);
     fprintf(f, "0x%08x\t", inst->binInst);
     fprintf(f, "0x%lx\t", inst->id);
     fprintf(f, "\n");
}

void instPrintDebug(PInst inst, FILE *f)
{
     fprintf(f, "i%lu:%d:%s ", inst->id, inst->owner->id, inst->owner->owner->name);
}

void instPrint(FILE * f, PInst inst)
{

     GList *l2;
     PArg producer, consumer;
     PArg src, dst;
     ArgIndex i;
     enum md_opcode op;

     fprintf(f, "%p i%lu:%d:%s 0x%x ", inst->basicBlockLink, inst->id, inst->owner->id, inst->owner->owner->name, inst->latencyType);

     forEachSrc(inst, i, src) {

	  fprintf(f, "%d(", src->reg);

	  forEach(src->producers, l2, producer) {
	       fprintf(f, "%lu ", producer->owner->id);
	  }
	  fprintf(f, ")");
	  
	  fprintf(f, "in%d",src->inputType);
	  fprintf(f, " ");
     }
 
     fprintf(f, " -> ");

     forEachDst(inst, i, dst) {
	  
	  fprintf(f, "%d(", dst->reg);

	  forEach(dst->consumers, l2, consumer) {
	       fprintf(f, "%lu ", consumer->owner->id);
	  }
	  fprintf(f, ")");
	  fprintf(f, "out%d",dst->outputType);
	  fprintf(f, " ");
     }
     
     if (inst->type == OP_TYPE_ALPHA_OP) {
	  MD_SET_OPCODE(op, inst->binInst);
	  //	  fprintf(f, "%s ", MD_OP_NAME(op));
	  md_print_insn(inst->binInst, 0, f);

	  if (instIsType(inst, F_LOAD | F_STORE)) {
	       epochTagPrint(f, inst);
	  }

     } else {
	  char postfix[1024];
	  postfix[0] = '\0';
	  
	  if (inst->type == OP_TYPE_FUNCTION_CALL) {
	       
	       if (inst->functionToCall) {
		    char buf[1024];
		    epochTagSPrint(buf, inst);
		    sprintf(postfix, " %s  %s", inst->functionToCall->name, buf);
	       } else {
		    char buf[1024];
		    epochTagSPrint(buf, inst);
		    sprintf(postfix, " %#llx  %s", inst->addressToCall, buf);
	       }

	  } else if (instNeedsEpochTags(inst)) {
	       char buf[1024];
	       epochTagSPrint(buf, inst);
	       sprintf(postfix, "  %s %s", buf, inst->isCanonicalEpochAdvance ? "**" : "");
	  }
	  fprintf(f, "%s%s ", OP_TYPE_NAME[inst->type], postfix);
     }
     fprintf(f, "\n");


}

char *instDotName(PInst inst, char *buf)
{
     
     sprintf(buf, "inst%lu", inst->id);

     return buf;
}

char *instOutputArgDotName(PArg arg, char *buf, bool_t simple)
{
     if (simple) {
	  sprintf(buf, "inst%lu", arg->owner->id);
     } else {
	  sprintf(buf, "inst%lu:out%d", arg->owner->id, arg->outputType);
     }
     return buf;
}

char *instInputArgDotName(PArg arg, char *buf, bool_t simple)
{
     if (simple) {
	  sprintf(buf, "inst%lu", arg->owner->id);
     } else {
	  sprintf(buf, "inst%lu:in%d", arg->owner->id, arg->inputType);
     }
     return buf;
}

void instPrintAsBox_DOT(FILE * f, PInst inst, char *prev)
{
     char name[1024], buf[256];
     char *shape, *color;
 
     color = "black";
     
     switch(inst->type) {
     case OP_TYPE_ALPHA_OP:
	  if (instIsType(inst, F_LOAD)) {
	       shape = "box";
	       color = "purple";
	  } else if (instIsType(inst, F_STORE)) {
	       shape = "box";
	       color = "forestGreen";
	  } else {
	       shape = "box";
	       color = "yellow";
	  }
	  break;
     case OP_identity:
     case OP_sink:
     case OP_mnop:
	  shape = "box";
	  color = "black";
	  break;
     case OP_src:
	  shape = "box";
	  color = "purple";
	  break;
	  /* 	  case OP_TYPE_TAG_SINK: */
	  /* 	       shape = "diamond"; */
	  /* 	       color = "black"; */
	  /* 	       break; */
     case OP_wa:
	  shape = "diamond";
	  color = "purple";
	  break;
/*      case OP_TYPE_EPOCH_COORDINATOR: */
/* 	  //	  case OP_TYPE_MEM_ORDER_STORE: */
/* 	  color = "forestGreen"; */
/* 	  shape = "circle"; */
/* 	  break; */
	  /* 	  case OP_TYPE_MEM_ORDER_LOAD: */
	  /* 	       color = "blue"; */
	  /* 	       shape = "circle"; */
	  /* 	       break; */
     case OP_rho:
	  shape = "triangle";
	  color = "black";
	  break;
     case OP_TYPE_FUNCTION_CALL:
	  shape = "circle";
	  color = "orange";
	  break;
     case OP_is:
     case OP_fanout:
     case OP_dtw:
     case OP_wtd:
     case OP_cnst:
	  shape = "box";
	  color = "red";
	  break;
     default:
	  shape = "box";
	  color = "black";
     }
	  
     if (dotOptions.instAsmAsLabel) {
	  md_sprint_insn(inst->binInst, 0, buf);
	  fprintf(f, "%s [label=\"%s\",shape=%s,color=%s];\n", instDotName(inst, name), buf, shape, color);
     } else {
	  fprintf(f, "%s [shape=%s,color=%s];\n", instDotName(inst, name), shape, color);
     }

}

char *edgeColors[] = {
     "red",
     "navyBlue",
     "forestGreen",
     "goldenrod",
     "purple",
     "orange",
     "magenta",
     NULL
};


void instPrintEdges_DOT(FILE * f, PInst inst, InstFilter filter)
{
     GList *link2;
     PArg producer, src;
     ArgIndex i;
     char b1[1024], b2[1024];
     char *color;
     char *style;
     forEachSrc(inst, i, src) {

	  forEach(src->producers, link2, producer) {
	       if (filter != NULL) {
		    if (!filter(producer->owner)) {
			 continue;
		    }
	       }

	       switch (producer->outputType) {
	       case OUTPUT_SERIALIZE:
		    color = "orange";
		    style = "";
		    break;
/* 	       case OUTPUT_CONTROL: */
/* 		    color = "black"; */
/* 		    style = ""; */
/* 		    break; */
/* 	       case OUTPUT_CONTROL_TAKEN: */
/* 		    color = "blue"; */
/* 		    style = ""; */
/* 		    break; */
/* 	       case OUTPUT_CONTROL_NOT_TAKEN: */
/* 		    color = "blue"; */
/* 		    style = ",style=dotted"; */
/* 		    break; */
/* 	       case OUTPUT_OUT: */
/* 		    color = "black"; */
/* 		    style = ",style=dotted"; */
/* 		    break; */
/* 	       case OUTPUT_MEM: */
/* 	       case OUTPUT_MEM_ORDER_COORDINATION: */
/* 		    color = "forestGreen"; */
/* 		    style = ""; */
/* 		    break; */
/* 	       case OUTPUT_MEM_ORDER: */
/* 	       case OUTPUT_MEM_STORE_ORDER: */
/* 		    style = ""; */
/* 		    color = "purple"; */
/* 		    break; */
/* 	       case OUTPUT_MEM_LOAD_ORDER: */
/* 		    color = "purple"; */
/* 		    style = ",style=dotted"; */
/* 		    break; */
/* 	       case OUTPUT_JMP_DST: */
/* 		    color = "blue"; */
/* 		    style = ",style=dotted"; */
/* 		    break; */
	       case OUTPUT_A:
	       case OUTPUT_C:
		    //	       case OUTPUT_OTHER:
	       default:
		    color = "red";
		    style = "";
	       }
	       

	       fprintf(f, "%s -> %s [color=%s,weight=1%s];\n", 
		       instOutputArgDotName(producer, b1, 1),
		       instInputArgDotName(src, b2, 1),
		       color, style
		       );
	  }
     }
}


bool_t instHasNoInputs(PInst inst)
{

     ArgIndex i;
     PArg src;
     bool_t foundInput = FALSE;

     forEachSrc(inst, i, src) {
	  if (g_list_length(src->producers) > 0) {
	       foundInput = TRUE;
	       break;
	  }
     }
     return !foundInput;
}

bool_t instIsConditionalMove(PInst inst)
{	       
	       
     enum md_opcode op;
     MD_SET_OPCODE(op, inst->binInst);
     switch (op) {
     case CMOVLBS: 
     case CMOVLBSI:
     case CMOVLBC: 
     case CMOVLBCI:
     case CMOVEQ:  
     case CMOVEQI: 
     case CMOVNE:  
     case CMOVNEI: 
     case CMOVLT:  
     case CMOVLTI: 
     case CMOVGE:  
     case CMOVGEI: 
     case CMOVLE:  
     case CMOVLEI: 
     case CMOVGT:  
     case CMOVGTI: 
     case FCMOVEQ: 
     case FCMOVNE: 
     case FCMOVLT: 
     case FCMOVGE: 
     case FCMOVLE: 
     case FCMOVGT: 
	  return TRUE;
	  break;
     default:
	  return FALSE;
     }
}

bool_t instIsCondBranch(PInst inst) 
{
     return instIsType(inst, F_COND | F_FPCOND);
}

bool_t instNeedsEpochTags(PInst inst)
{
     return instIsType(inst, F_LOAD | F_STORE) || 
	  inst->isCanonicalEpochAdvance || 
	  inst->type == OP_mnop ||
	  inst->type == OP_mnop_ping;     
}


int instIsType(PInst i, int type)
{
     int op;

     if (i->type != OP_TYPE_ALPHA_OP) {
	  return 0;
     } else {
	  MD_SET_OPCODE(op, i->binInst);
	  return (MD_OP_FLAGS(op) & type);
     }
}

bool_t instIsNoop(PInst inst) 
{
     enum md_opcode op;     
     MD_SET_OPCODE(op, inst->binInst);
		    
     if (op == BIS) {
	  if (EXTRACT_RA(inst->binInst) == 31 && 
	      EXTRACT_RB(inst->binInst) == 31 && 
	      EXTRACT_RC(inst->binInst) == 31) {
	       return TRUE;
	  } else {
	       return FALSE;
	  }
     } else if (op == LDA || op == LDAH) {
	  if (EXTRACT_RA(inst->binInst) == 31) {
	       return TRUE;
	  } else {
	       return FALSE;
	  }
     } else if (op == LDQ_U) {
	  if (EXTRACT_RA(inst->binInst) == 31) {
	       return TRUE;
	  } else {
	       return FALSE;
	  }
     }
     
     return FALSE;
}

int instNeedsJumpPorts(PInst i)
{
     int op;
     MD_SET_OPCODE(op, i->binInst);
     return (op == JSR || op == BSR || op == JMP || op == JSR_COROUTINE);
}

int instNeedsReturnPorts(PInst i)
{
     int op;
     MD_SET_OPCODE(op, i->binInst);
     return (op == JSR || op == BSR);
}


int instCountConsumers(PInst inst)
{
     ArgIndex i;
     PArg dst;
     int count = 0;

     forEachDst(inst, i, dst) {
	  count += g_list_length(dst->consumers);
     }

     return count;
}

int instCountProducers(PInst inst)
{
     ArgIndex i;
     PArg src;
     int count = 0;

     forEachSrc(inst, i, src) {
	  count += g_list_length(src->producers);
     }

     return count;
}

void argInit(PArg n, 
	     int reg, 
	     InputArgType inputArgType, 
	     OutputArgType outputArgType,
	     PInst owner)
{
     memset(n, 0, sizeof(Arg));
     assert(inputArgType == INPUT_INVALID || outputArgType == OUTPUT_INVALID);
     assert(owner);
     n->reg = reg;
     n->owner = owner;
     n->inputType = inputArgType;
     n->outputType = outputArgType;
     n->valid = TRUE;
     n->oldReg = -1;
}

PArg argNew(int reg, 
	    InputArgType inputArgType, 
	    OutputArgType outputArgType,
	    PInst owner)
{
     static int idCount = 0;
     PArg n = (PArg)malloc(sizeof(Arg));
     argInit(n, reg, inputArgType, outputArgType, owner);
     n->id = idCount++;
     return n;
}

void argSanity(PArg arg) 
{
     GList *l;
     PArg a;

     if (!arg->valid) {
	  return;
     }

     forEach(arg->consumers, l, a) {
	  assert(a);
     }
     forEach(arg->producers, l, a) {
	  assert(a);
     }
     



     assert(arg->inputType == INPUT_INVALID || 
	    arg->outputType == OUTPUT_INVALID);
     assert(arg->owner);
}

void argTearDown(PArg arg)
{

     g_list_free(arg->consumers);
     g_list_free(arg->producers);

}

void argClearDfgPtrs(PArg arg)
{
     g_list_free(arg->consumers);
     g_list_free(arg->producers);

     arg->consumers = NULL;
     arg->producers = NULL;
}

gint pArgsCompare(gconstpointer a, gconstpointer b)
{
     PArg aa = (PArg) a;
     PArg bb = (PArg) b;

     if (aa->reg < bb->reg) {
	  return -1;
     } else if (aa->reg > bb->reg) {
	  return 1;
     } else {
	  return 0;
     }
}

GList *combineArgListsUniquely(GList * dst, GList * src)
{
     GList *l1;
     PArg source;
     forEach(src, l1, source) {
	  if (!g_list_find_custom(dst, source, pArgsCompare)) {
	       dst = g_list_append(dst, source);
	  }
     }
     return dst;
}

int getUniqueVariable() 
{
     static int v = 100;
     return v++;
}
