 #ifndef OP_INCLUDED
#define OP_INCLUDED


#include "types.h"
#include "machine.h"
#include "list.h"
#include "ptrSet.h"
#include "WSOpCode.h"
#define OUTPUT_CONTROL_DIRECTIONS 3

#include "instDefs.h"


static int ARG_TYPE_TO_REG[];
static int REG_TO_ARG_TYPE[];


typedef struct _INST *PInst;
typedef struct _ARG *PArg;

#include "function.h"
#include "basicBlock.h"

#include "epoch.h"

#include "DecodedInstruction.h"
#include "rippleChunk.h"

#define MAX_VARIABLES 256

#define ARG_VALID(a) ((a)->valid)
//#define ARG_IS_VALUE(a)   ((a)->inputType == INPUT_INVALID ? (a)->outputType < OUTPUT_CONTROL_NOT_TAKEN : (a)->inputType < INPUT_CONTROL_A)

//#define ARG_IS_CONTROL(a) ((a)->inputType == INPUT_INVALID ? (a)->outputType < OUTPUT_CONTROL_NOT_TAKEN : (a)->inputType < INPUT_CONTROL_A)

typedef GList * ArgIndex;

#define forEachSrc(inst, v, src) forEach(inst->srcs, v, src) if(src->valid)
#define forEachDst(inst, v, dst) forEach(inst->dsts, v, dst) if(dst->valid)

#define DST_N(inst, i) ((PArg)g_list_nth_data(inst->dsts, i))
#define SRC_N(inst, i) ((PArg)g_list_nth_data(inst->srcs, i))

//#define forEachSrc(inst, v, src) for((v) = 0, (src) = &(inst)->srcs[(v)]; (v) < INPUT_MAX; (v)++, (src) = &(inst)->srcs[(v)]) if(src->valid)
//#define forEachDst(inst, v, dst) for((v) = 0, (dst) = &(inst)->dsts[(v)]; (v) < OUTPUT_MAX; (v)++, (dst) = &(inst)->dsts[(v)]) if(dst->valid)


#define inputArgTypeIsSane(a)  ((((a) < INPUT_MAX) && ((a) >= INPUT_A)) ? 1 : 0)
#define outputArgTypeIsSane(a) ((((a) < OUTPUT_MAX) && ((a) >= OUTPUT_A)) ? 1 : 0)

#define REG_CONTROL 1024

typedef bool_t (*InstFilter)(PInst inst);

typedef struct _ARG {
     int reg;
     int id;
     OutputArgType outputType;
     GList *consumers;

     InputArgType inputType;
     GList *producers;

     PInst owner;
     int notAlwaysInitialized;
     int initializedOnce;

     bool_t valid;
     bool_t isZero;

     // This is just for ssa
     int oldReg;

     // This is for OP_src outputs.  THis is value it will output.
     qword_t specialSourceValue; 
     int specialSourceEpoch;
     

} Arg;

typedef struct _INST {
     InstructionID id;
     InstType type;
     WSOpCode wsOpCode;
     gint64 immediate;
     char *instructionLabel;
     char *symbolicImmediate;
     bool_t exportLabel;
     bool_t reportValue;
     bool_t isForFunctionCall;
     md_inst_t binInst;
     PC pc;

     bool_t reachesSomethingInteresting;

     GList *srcs;
     GList *dsts;
     
/*       GList *srcs; */
/*       GList *dsts; */

     PC targetPC;

     PBasicBlock owner;

     //     PProcessingElement pe;

     int locx;
     int locy;

     int callPort;
     int argPort;
     int returnPort;

     int isPlaced;
     unsigned long int startedHere;
     int isLoadPredecessor;

     bool_t isFreeable; // Little sanity check to catch free insts that are part of an array

     bool_t isControlHead;
     //     bool_t isControlTail;

     DecodedInstruction decoded;

     bool_t hasNoLocalPredecessor; // All src producers are in other blocks.

     /* stuff for phi functions */
     bool_t isUneededPhiFunction; // This phi func is not needed
     PInst phiFunctionControl;    // This is the branch that provides control for this phi func

     
     // Function call data
     PFunction functionToCall;
     md_addr_t addressToCall;
     PInst callingInstruction;
     
     // Is this instruction in the alternate entrance block (the one for
     // non-local invocations that sets the gp).  THis is the number of times
     // this instruction should become the identity.
     int forceToIdentity;

     // Quick access to our containing link in the list of instructions
     GList * basicBlockLink;

     // number of loads this instruction enables if it is a memory coordination instruction
     int dependentLoadCount;
     int dependentStoreCount;

     int regNum;

     Epoch epochOrderPrev;
     Epoch epochOrder;
     Epoch epochOrderNext;
     int rippleSize;
     int lastStoreSequenceNumber;
     bool_t isEndOfEpoch;
     
     int waveAdvanceAmount;

     bool_t isCanonicalEpochAdvance;
     uint waveAdvanceCount;

     PPtrSet possibleNextMemOps;
     PPtrSet possiblePrevMemOps;

     bool_t isFirstInBlock;

     bool_t isTokenizeNop;

     //     ExecModelInstData execModelData;
     //     StatsInst stats;

     bool_t isLoad;
     bool_t isFoldedRho;
     // Is this an instruction we think can be folded into another?
     int latencyType;

     PRippleChunk rippleChunk;

     bool_t isForLocalInvocation;
     PInst controllingEpochCoordinate;
     bool_t  isUnremovablePhiFunc;
     PInst tagHandShakeMaster;
} Inst;


gint pArgsCompare(gconstpointer a, gconstpointer b);
void argClearDfgPtrs(PArg arg) ;
void argSanity(PArg arg) ;

PArg instGetDstByType(PInst inst, OutputArgType type) ;
PArg instGetSrcByType(PInst inst, InputArgType type) ;


PArg instAddDst(PInst inst, 
		int reg, 
		OutputArgType outputArgType);

void instRemoveSrc(PInst inst,
		   PArg src);
void instRemoveDst(PInst inst,
		   PArg src);

PArg instAddSrc(PInst inst, 
		int reg, 
		InputArgType inputArgType);

void instPrintDebug(PInst inst, FILE *f);
void instPrint(FILE*f, PInst inst);
void instFree(PInst inst) ;
void instParse(char *buf, GList** bb);
void instPrintAsBox_DOT(FILE *f, PInst inst, char *prev);
void instPrintEdges_DOT(FILE * f, PInst inst, InstFilter filter);
//void instPlaceInSubstrate(PInst inst, PInst substrate[SUBSTRATE_SIZE][SUBSTRATE_SIZE], int x, int y);
void instPrintAsm(FILE *f, guint32 wi);

void instInit(PInst n, InstType type);
PInst instNew(InstType type);
PInst instNewFromBin(md_inst_t inst, 
		     PC pc);
void instInitFromBin(PInst n, 
		     md_inst_t inst, 
		     PC pc);
bool_t instIsConditionalMove(PInst inst);
int instIsType(PInst i, int type); 
bool_t instIsCondBranch(PInst inst);
int instNeedsReturnPorts(PInst i);
int instNeedsCallPorts(PInst i);

void instInstallControlInput(PInst inst, InputArgType inputType);
void instInstallControlOutput(PInst inst, OutputArgType outputType) ;

void argInit(PArg n, int reg, InputArgType inputType, OutputArgType outputType, PInst owner);
PArg argNew(int reg, 
	    InputArgType inputArgType, 
	    OutputArgType outputArgType,
	    PInst owner);
void argTearDown(PArg arg);

GList *combineArgListsUniquely(GList *dst, GList *src);
int instNeedsJumpPorts(PInst i) ;

void instSanity(PInst i);

int getUniqueVariable();
bool_t instHasNoInputs(PInst inst);
int instCountConsumers(PInst inst);
int instCountProducers(PInst inst);
PInst instNewFunctionCall(PInst callingInst) ;

bool_t instNeedsEpochTags(PInst inst);
void instPruneFunctionCall(PInst inst);
void instPrintAnnotatedMap(PInst inst, FILE *f);

bool_t instIsNoop(PInst inst) ;

void instMoveSrc(PArg arg,
		 PInst to);
void instMoveDst(PArg arg,
		 PInst to);

#endif

