#include <stdio.h>
#include "cfg.h"
#include "dfg.h"
#include "inst.h"
#include "basicBlock.h"
#include "stdlib.h"
/*  #include "greedyPlace.h" */
/*  #include "inorderPlace.h" */
/*  #include "blockGreedyPlace.h" */
/*  #include "perfectPlace.h" */
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "lcc.h"
/*  #include "loadCentricPlace.h" */
/*  #include "simulation.h" */
#include "parse.h"
#include "types.h"
#include "list.h"
#include "function.h"


#include <sys/types.h>
#include <sys/stat.h>

#ifdef Darwin
#include <fcntl.h>
#define O_LARGEFILE 0
#else
#include <asm/fcntl.h>
#endif

/*  #include "stats.h" */
#include "program.h"

#include "ssa.h"
#include "dot.h"
#include "controlDistribution.h"

#include "mem.h"

#define BUF/* SIZE 256 */

#define INCLUDE_WRITE
#include "execFormat.h"
#undef INCLUDE_WRITE

LccOptions lccOptions;

static void parseCmdLine(int argc, char *argv[]) 
{
/*      extern char* optarg; */
/*      extern int optind; */
     char c;

     lccOptions.traceFile = NULL;
     lccOptions.mapFile = NULL;
     lccOptions.printCfg = FALSE; 
     lccOptions.printCdg = FALSE; 
     lccOptions.printPostDom = FALSE;
     lccOptions.printPreDom = FALSE; 
     lccOptions.analyzeOnly = FALSE;
     lccOptions.cfgOut = NULL;
     lccOptions.dfgOut = NULL;
     lccOptions.mapOut = NULL;
     lccOptions.kenTraceBase = NULL;
     lccOptions.doTiny = FALSE;
     cfgOptions.useRipples = FALSE;
     cfgOptions.verbose = FALSE;
     lccOptions.printControlDistribution = FALSE;
     cfgOptions.dumpFinal = TRUE;
     lccOptions.printEpochControl = FALSE;
     lccOptions.compiledOutput = NULL;
     //     cfgOptions.includeWaveCoordinate = FALSE;

     cfgOptions.includeExitEnterBlocks = TRUE;
     dotOptions.printSimpleInsts = FALSE;
     dotOptions.instAsmAsLabel = FALSE;
     cfgOptions.dynamicInstructionsByEpoch = TRUE;
     cfgOptions.performEpochNumbering = TRUE;
     cfgOptions.compileSpecFile = NULL;

     cfgOptions.doTokenMemory = FALSE;

     cfgOptions.realFunctionCalls = TRUE;
     cfgOptions.outputAssembly = TRUE;
     cfgOptions.supressEpochOptimize =TRUE;

     while (((c = getopt(argc, argv, "RgcC:D:M:TdtoYVEk:K:Is:BpXaeArm:SZOWvF")) != -1)) { 
	  switch (c) {
	  case 'W':
	       cfgOptions.compressWaveAdvance = TRUE;
	       break;
	  case 't':
	       cfgOptions.doTokenMemory = FALSE;
	       break;
 	  case 'F': 
 	       cfgOptions.realFunctionCalls = TRUE; 
 	       break; 
	  case 'k':
	       lccOptions.kenTraceBase = strdup(optarg);
	       break;
	  case 's':
	       cfgOptions.compileSpecFile = strdup(optarg);
	       break;
	  case 'K':
	       lccOptions.compiledOutput = strdup(optarg);
	       break;
	  case 'Y': 
	       cfgOptions.supressEpochOptimize =TRUE;
	       break;
	  case 'R': 
	       cfgOptions.useRipples = TRUE;
	       break;
	  case 'I': 
	       cfgOptions.retchedInfiniteRecursionHack = TRUE;
	       break;
	  case 'V': 
	       cfgOptions.verbose = TRUE;
	       break;
	  case 'v':
	       lccOptions.printDfg = TRUE; 
	       lccOptions.printCfg = TRUE; 
	       break;
	  case 'A':
	       cfgOptions.outputAssembly = TRUE;
	       break;
	  case 'a':
	       lccOptions.analyzeOnly = TRUE;
	       break;
	  case 'm':
	       lccOptions.mapFile = strdup(optarg);
	       break;
	  case 'g': 
	       lccOptions.printDfg = TRUE; 
	       break; 
	  case 'c': 
	       lccOptions.printCfg = TRUE; 
	       break; 
	  case 'd': 
	       lccOptions.printCdg = TRUE; 
	       break; 
	  case 'o': 
	       lccOptions.printPostDom = TRUE; 
	       break;
	  case 'r': 
	       lccOptions.printPreDom = TRUE; 
	       break; 
	  case 'E': 
	       cfgOptions.includeExitEnterBlocks = TRUE; 
	       break; 
	  case 'e': 
	       lccOptions.printEpochControl = TRUE; 
	       break; 
	  case 'p': 
	       lccOptions.printControlDistribution = TRUE; 
	       break;
	  case 'C':
	       lccOptions.cfgOut = strdup(optarg);
	       break;
	  case 'D':
	       lccOptions.dfgOut = strdup(optarg);
	       break;
	  case 'M':
	       lccOptions.mapOut = strdup(optarg);
	       break;
	  case 'T':
	       lccOptions.doTiny = TRUE;
	       break;
	  case 'S':
	       cfgOptions.serialize = TRUE;
	       break;
	  case 'O':
	       cfgOptions.onlyCFG = TRUE;
	       break;
	  case 'X':
	       cfgOptions.performEpochNumbering = TRUE;
	       break;
	  case 'Z':
	       cfgOptions.dynamicInstructionsByEpoch = TRUE;
	       break;
	  case 'B':
	       cfgOptions.supressCollapsing = TRUE;
	       break;
	  }
     }

     //     cfgOptions.onlyCFG = !lccOptions.printDfg;
}
FILE* openOutputFile(char *f)
{
     FILE *outFile;
     
     outFile = fopen(f, "w");
     if (!outFile) {
	  perror("lcc");
	  exit(1);
     }
     
     return outFile;
}


FILE *openSpecFile(char *specFileName)
{
     FILE *specFile;
     
     specFile = fopen(specFileName, "r");
     if (!specFile) {
	  perror("lcc");
	  exit(1);
     }
     
     return specFile;
}

FILE *openMapFile() 
{
     FILE *mapFile;
     int mapFileDesc = -1; 

     mapFileDesc = open(lccOptions.mapFile, O_RDONLY|O_LARGEFILE);
     if (mapFileDesc == -1) {
	  error("Couldn't open map file: %s\n", lccOptions.mapFile);
	  perror("lcc");
	  exit(1);
     }

     mapFile = fdopen(mapFileDesc, "r");
     if(mapFile == NULL) {
	  error("Couldn't open map file: %s\n", lccOptions.mapFile);
	  perror("lcc");
	  exit(1);
     }

     return mapFile;
}

void printOutput(PProgram prog, FILE *f)
{
     PFunction func;
     GList *l;

     
     if (lccOptions.printDfg) {
	  forEach(prog->functions, l, func) {
	       cfgPrint_DOT(f, func, TRUE);
	  }
     }

     if (lccOptions.printCfg) {
	  forEach(prog->functions, l, func) {
	       cfgPrint_DOT(f, func, FALSE);
	  }
     }

/*      if (lccOptions.printCdg) { */
/* 	  forEach(prog->functions, l, func) { */
/* 	       ssaPrintCDG(func, f); */
/* 	  } */
/*      } */
 
     if (lccOptions.printPreDom) { 
	  forEach(prog->functions, l, func) {
	       dominancePrintDominatorTree(func, f, DOMINANCE_DIRECTION_PRE);
	  }
     }
     
     
     if (lccOptions.printPostDom) { 
	  forEach(prog->functions, l, func) {
	       dominancePrintDominatorTree(func, f, DOMINANCE_DIRECTION_POST);
	  }
     }

/*      if (lccOptions.printControlDistribution) { */
/* 	  forEach(prog->functions, l, func) { */
/* 	       controlDistributionPrint(func, f); */
/* 	  } */
/*      } */


     if (lccOptions.printEpochControl) { 
	  forEach(prog->functions, l, func) {
	       memPrintEpochCoordinationGraph_DOT(stdout, func);
	  }
     } 

}

int main (int argc, char *argv[]) 
{
     FILE *mapFile;
     FILE *compileSpec = NULL;
     PProgram program;
     PDfgStats stats;

     InitSSToWSMap();
     parseCmdLine(argc, argv);

     nice_calloc(stats, sizeof(DfgStats));

     md_init_decoder();
     
     if (cfgOptions.compileSpecFile != NULL) {
	  compileSpec = openSpecFile(cfgOptions.compileSpecFile);
     }

     mapFile = openMapFile();

     program =  programLoad(mapFile, compileSpec, 0);
     
     programPreTraceProcessing(program);
     programPostTraceProcessing(program, 1);
     
     programCollectDfgStats(program, stats);
     dfgPrintStats(stderr, stats);

     programPrintPerFunctionStats(program, stderr);
     printOutput(program, stdout);

     if (lccOptions.kenTraceBase != NULL) {
	  FILE *cfg;
	  FILE *dfg;
	  FILE *amap;
	  char buf[1000];
	  PFunction func;
	  GList *l;
	  
	  strcpy(buf, lccOptions.kenTraceBase);
	  cfg  = fopen(strcat(buf, ".cfg"), "w");
	  strcpy(buf, lccOptions.kenTraceBase);
	  dfg  = fopen(strcat(buf, ".dfg"), "w");
	  strcpy(buf, lccOptions.kenTraceBase);
	  amap = fopen(strcat(buf, ".amap"), "w");
	  assert(cfg && dfg && amap);

	  forEach(program->functions, l, func) {
	       cfgPrintExport(func, cfg);
	  }
	  
	  forEach(program->functions, l, func) { 
	       dfgPrintExport(func, dfg); 
	  } 

	  forEach(program->functions, l, func) { 
	       functionPrintAnnotatedMap(func, amap); 
	  } 
	  fclose(cfg);
	  fclose(dfg);
	  fclose(amap);
     }

     if (lccOptions.compiledOutput) {
	       FILE *out;
	  out = fopen(lccOptions.compiledOutput, "w");
	  assert(out);
	  execFormatWriteProgram(program, out, cfgOptions.outputAssembly);
     }
     
     return 0;
}

