#include "mem.h"
#include "function.h"
#include <stdio.h>
#include "wsArch.h"
#include "basicBlock.h"
#include "dot.h"
#include <assert.h>


void memInsertEpochAdvance(PFunction func)
{

     int reg;
     GList *l;
     PBasicBlock block;

     forEach(func->basicBlocks, l, block) {
	  if (block->isEpochHead || 
	      block == func->localEntry) { 
	       //	       if (!block->isFunctionCallBlock) {
	       PInst n;
	       PInst firstAdded = NULL;

	       block->hasEpochExitsInstalled = TRUE;
		    
	       for(reg = 0; reg < REGISTER_COUNT ; reg++) {
		    if (reg != EPOCH_COORDINATION_REGISTER &&
			reg != EPOCH_REGISTER) {
			 n = instNew(OP_wa);
			      
			 /* 			      if (cfgOptions.includeWaveCoordinate) { */
			 /* 				   instAddSrc(n, EPOCH_REGISTER, INPUT_EPOCH); */
			 /* 			      } */
			 instAddSrc(n, reg, INPUT_A);
			 instAddDst(n, reg, OUTPUT_A);	 
			      
			 if (block == func->enter || 
			     block == func->localEntry ||
			     block->isFunctionCallBlock) { // || block->isLoopTop) {
			      basicBlockAppendInst(block, n);
			      if (firstAdded == NULL) {
				   firstAdded = n;
			      }
			 } else {
			      basicBlockPrependInst(block, n);
			 }
			 n->epochOrder = 0;
		    }
		    //		    }

/* 		    if (cfgOptions.includeWaveCoordinate) { */
/* 			 n = instNew(OP_TYPE_EPOCH_COORDINATOR); */
/* 			 instAddSrc(n, EPOCH_COORDINATION_REGISTER, INPUT_EPOCH_COORDINATION); */
/* 			 instAddDst(n, EPOCH_COORDINATION_REGISTER, OUTPUT_EPOCH_COORDINATION); */
/* 			 instAddDst(n, EPOCH_REGISTER, OUTPUT_EPOCH); */
/* 			 if (block == func->enter) { // || block->isLoopTop) { */
/* 			 basicBlockInsertInstBeforeInst(block, firstAdded, n);	       */
/* 			 } else { */
/* 			      basicBlockPrependInst(block, n); */
/* 			 } */
/* 		    }  */
	       }
	  }
     }
}

static bool_t memEpochInstFilter(PInst inst)
{
     PArg input;
     
     if (//inst->type == OP_TYPE_EPOCH_COORDINATOR ||
	 inst->type == OP_TYPE_FUNCTION_CALL) {
	  return TRUE;
     }

     if (inst->type == OP_rho) {
	  input = instGetSrcByType(inst, INPUT_A);
     } else if (inst->type == OP_wa) {
	  if (inst->isCanonicalEpochAdvance) {
	       return TRUE;
	  }
     }

     
     if (input != NULL) {
	  if (input->reg == EPOCH_COORDINATION_REGISTER) {
	       return TRUE;
	  } else {
	       return FALSE;
	  } 
     } else {
	  return FALSE;
     }
}
	 
void memPrintEpochCoordinationGraph_DOT(FILE *f, PFunction func) 
{
     GList *l;
     PBasicBlock block;
     char graphName[1024];
     sprintf(graphName, "%s: CFG", func->name);

     dotOpenGraph(f, graphName);
     
     forEach(func->basicBlocks, l, block) {
	  GList *l;
	  char blockName[1024], 
	       blockName2[1024], 
	       last[1024], 
	       tail[1024], 
	       dst[1024];
	  PInst inst;
	  PCfgEdge edge;
#undef SUBGRAPHS
	  //#define SUBGRAPHS
#ifdef SUBGRAPHS

	  static int clusterNum= 0;

     
     
	  fprintf(f, "subgraph cluster%d { label=\"%s\";\ncolor=blue;\nlabel=\"%s\";\n",   
		  clusterNum++,  
		  basicBlockName_DOT(block, blockName),   
		  basicBlockName_DOT(block, blockName2));  
#endif
     
	  fprintf(f, "%s [height=0.05,shape=box,label=%s%s];\n", 
		  basicBlockName_DOT(block, blockName),  
		  basicBlockName_DOT(block, blockName2), block->isLoopExitBlock ? ",color=red" : ""); 
     
  
	  {

	       strcpy(last, "");

	       forEach(block->insts, l, inst) {  
		    if (memEpochInstFilter(inst)) {
			 instPrintAsBox_DOT(f, inst, last);     
		    }  
	       }
	  }

	  /*      fprintf(f, "%s [height=0.05,shape=box,label=%s];\n",  */
	  /* 	     basicBlockName_DOT(block, blockName), */
	  /* 	     basicBlockName_DOT(block, blockName2)); */

#ifdef SUBGRAPHS
	  fprintf(f, "}\n");
#endif

	  forEach(block->exits, l, edge) {
	       if (edge->dst) {
		    fprintf(f, "%s  -> %s [style=bold,weight=%d];\n", 
			    basicBlockName_DOT(block, tail), 
			    basicBlockName_DOT(edge->dst, dst), 1);
	       }   
	  }


	  forEach(block->insts, l, inst) { 
	       if (memEpochInstFilter(inst)) {
		    instPrintEdges_DOT(f, inst, memEpochInstFilter); 
	       }
	  } 
     }

     dotCloseGraph(f);
}
