#include <glib.h>
#include "basicBlock.h"
#include "inst.h"
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include "types.h"
#include "list.h"
#include "inst.h"

#define debug(a, ...) ; 
#include "program.h"
//#include "stats.h"
//#include "histogram.h"
typedef enum {PARSE_NOWHERE, PARSE_IN_FUNCTION, PARSE_IN_BLOCK} ParsingStates;

#define parseHex(a) (strtol(a, NULL,0))
#define parseHexU(a) (strtoul(a, NULL,0))

//#define DEBUG
void parseLine(char *buf, 
	       PProgram program,
	       int *instructionsParsed)
{

     
     static PFunction currentFunction = NULL;
     static PBasicBlock currentBlock = NULL;
     static ParsingStates state = PARSE_NOWHERE;

     char *bufParts[4];
     int i;
     int p = 0;
     int len = strlen(buf);
     char *functionTag;
     char *blockTag;
     char *instPC;
     char *instWord;

     for(i = 0; i < 4; i++) {
	  bufParts[i] = 0;
     }

     bufParts[p++] = buf;
     for (i = 0; i < len;  i++) {
    
	  if (buf[i] == '\t') {
	       bufParts[p++] = buf + i + 1;
	       buf[i] = '\0';
	  }

	  if (buf[i] == '\n') {
	       buf[i] = '\0';
	  }
     }

     p = 0;
     functionTag = bufParts[p++];
     blockTag    = bufParts[p++];
     instPC      = bufParts[p++];
     instWord    = bufParts[p++];

     if (!strncmp("Start Function:", functionTag, strlen("Start Function:"))) {
	  char *functionName;
	  int action;
	  SpecTableEntry *specTable;
	  assert(state == PARSE_NOWHERE);
	  state = PARSE_IN_FUNCTION;
	  functionName = functionTag + strlen("Start Function: ");

	  specTable = programGetCompileAction(program, functionName);
	  if (specTable) {
	       action = specTable->action;
	  } else {
	       action = DO_NOTHING;
	  }
	  currentFunction = functionNew(functionName, action);
	  if (specTable) {
	       currentFunction->isMain = specTable->isMain;
	  }
	  if (action == DO_STUB) {
	       currentFunction->stubID = stubCount++;
	       currentFunction->localStubID = stubCount++;
	  }
	       //	  fprintf(stderr,"Created function: %s; action = %d\n", functionName, action);
	  debug("Created function: %s\n", functionName);
	  return;
     } else if (!strncmp("End Function", functionTag, strlen("End Function"))) {
	  assert(state == PARSE_IN_FUNCTION);
	  state = PARSE_NOWHERE;
	  assert(currentFunction);
	  programAddFunction(program, currentFunction);
	  currentFunction = NULL;
	  //	  incrementHistogram(stats->functionInstructionLengthHisto, currentFunction->lengthInstructions);
	  //	  incrementHistogram(stats->functionBasicBlockLengthHisto, currentFunction->lengthBasicBlocks);
	  debug("\n");
	  return;
     } else if (strncmp("", functionTag, strlen(""))) {
	  assert (0);
     }

     if (blockTag == NULL) {
	  return;
     }

     if (!strncmp("BlockID:", blockTag, strlen("BlockID:"))) {
	  int blockID;
	  char blockName[1024];
	  assert(state == PARSE_IN_FUNCTION);
	  state = PARSE_IN_BLOCK;
	  assert(currentFunction);

	  blockID = atoi(blockTag + strlen("BlockID:"));
	  sprintf(blockName, "B%d", blockID); 
	  currentBlock = basicBlockNew(blockName, blockID);
	  debug("Created block %d\n", blockID);
	  return;	  
     } else if (!strncmp("End Block", blockTag, strlen("End Block"))) {
	  assert(state == PARSE_IN_BLOCK);
	  state = PARSE_IN_FUNCTION;
	  assert(currentFunction);
	  assert(currentBlock);
	  functionAddBasicBlock(currentFunction, currentBlock);
	  currentBlock = NULL;
	  debug("\n");
	  return;
     } else if (strncmp("", blockTag, strlen(""))) {
	  assert(0);
     }

     if (state == PARSE_IN_BLOCK) {
	  PInst n;
	  PC pc;
	  assert(currentFunction);
	  assert(currentBlock);

	  pc = strtoll(instPC, NULL, 0);

	  if (g_list_empty(currentBlock->insts)) {
	       basicBlockSetPC(currentBlock, pc);
	  }
	  if (program->instructions) {
	       n = program->instructions + *instructionsParsed;
	  } else {
	       n = (PInst)calloc(sizeof(Inst), 1);
	  }
	  
	  
	  instInitFromBin(n,
			  (md_inst_t)parseHexU(instWord),
			  pc);
	  
	  (*instructionsParsed)++;
	  
	  debug("Created instruction: ");

	  basicBlockAppendInst(currentBlock, n);

     } else {
	  assert(0);
     }
     
}

