#ifndef PROGRAM_INCLUDED
#define  PROGRAM_INCLUDED

#include <glib.h>
#include "types.h"

typedef struct _PROGRAM *PProgram;

#include "function.h"
#include "basicBlock.h"
#include "inst.h"

#include "dfg.h"

typedef struct _PROGRAM {
  
     GTree *basicBlocksById;
     GTree *basicBlocksByPC;
     GTree *functionCallBlocksByPC;
     GTree *instructionsByID;
     GTree *functionsByPC;
     GList *compileSpec;
     GList *functions;   // List of functions
     int instCount;      // number of instructions in executable
     PInst instructions; // array of instructions for main text segment   
     PInst initInstructions;         // array of instructions for init segment
     PInst finiInstructions;         // array of instructions for fini segment
} Program;

#define DO_COMPILE 1
#define DO_NOTHING 2
#define DO_STUB 3

typedef struct _SPEC_TABLE_ENTRY {
     char *functionName;
     int action;
     bool_t isMain;
} SpecTableEntry;


PProgram programNew();
void programFree(PProgram p);

PBasicBlock programFindBasicBlockByPC(PProgram program, PC pc);
PBasicBlock programFindBasicBlockById(PProgram program, int id);
PBasicBlock programFindBlockByLabel(PProgram program, char* targetLabel);

SpecTableEntry * programGetCompileAction(PProgram program, char *function);

PProgram programLoad (FILE *f, FILE *spec,
		  int instCount);

void programAddInst(PProgram prog, PInst inst);
void programAddFunction(PProgram prog, PFunction func);

void programBuildDfgCfg(PProgram program);
void programPreTraceProcessing(PProgram program);
void programPostTraceProcessing(PProgram program, bool_t analyzeOnly);
void programCollectDfgStats(PProgram program, PDfgStats stats) ;
PInst programGetInstByID(PProgram progr, int id);
void programFixUpForExecution(PProgram program) ;


PFunction programFindFunctionByName(PProgram program, char *funcName);
PFunction programFindFunctionByPC(PProgram program, qword_t pc);

void programCompressInstructionIDs(PProgram program);

void programRenumberBlocks(PProgram program) ;

void programPrintPerFunctionStats( PProgram program, FILE *f);

void programBuildRippleChunks(PProgram program);
#endif
