#include "basicBlock.h"
#include "inst.h"
#include "function.h"
#include "set.h"
#include "list.h"
#include "rhoFunctions.h"
#include "wsArch.h"

#include <assert.h>

bool_t rhoReachesSomethingInteresting_HELP(PInst inst, PPtrSet visited)
{
     ArgIndex i;
     GList *l;
     PArg dst, consumer;
     //     PtrSetIndex k;
     //     PInst slave;

     if (ptrSetElementOf(visited, inst))
	  return FALSE;

     ptrSetAdd(visited, inst);
     
     if (	 inst->type != OP_rho &&
	 	 inst->type != OP_identity &&
		 inst->type != OP_dtw &&
		 inst->type != OP_wa &&
		 inst->type != OP_fanout
		 ) {
	  return TRUE;
     }

     forEachDst(inst, i, dst) {
	  forEach(dst->consumers, l, consumer) {
	       if (consumer->owner->reachesSomethingInteresting || 
		   rhoReachesSomethingInteresting_HELP(consumer->owner, visited)) {
		    consumer->owner->reachesSomethingInteresting = TRUE;
		    return TRUE;
	       }
	  }
     }

     return FALSE;

}

void rhoResetInterestingFlags(PFunction func)
{
     
     GList *l, *j;//, *i;
     PBasicBlock block;
     PInst inst;
     forEach(func->basicBlocks, l, block) {
	  forEach(block->insts, j, inst) {
	       inst->reachesSomethingInteresting = FALSE;
	  }
     }
}

bool_t rhoReachesSomethingInteresting(PInst inst)
{

     PPtrSet visited = ptrSetNew();
     bool_t ret;

     ret = rhoReachesSomethingInteresting_HELP(inst, visited);
     ptrSetFree(visited);

     return ret;
     
}



static void controlDependenceDFSPostOrder(PBasicBlock block, GList **l)
{
     GList *i;
     PBasicBlock b;
     
     forEach(block->controlDependentNodes, i, b) {
	  controlDependenceDFSPostOrder(b, l);
     }
     
     *l = g_list_append(*l, block);
}

#if(0)
static GList *buildControlDependenceDFSPostOrder(PFunction func)
{
     GList *l = NULL;
     
     controlDependenceDFSPostOrder(func->enter, &l);

     return l;
}
#endif

static void rhoHoistInputs(PInst rho)
{
     GList *l, *k;
     PArg producer, consumer;
     PArg input;
     PArg dst;

     input = instGetSrcByType(rho, INPUT_A);
     dst = instGetDstByType(rho, OUTPUT_A);
     forEach(input->producers, k, producer) {
	  forEach(dst->consumers, l, consumer) {
	       dfgAddEdge(producer, consumer);
	  }
     }
}

void rhoCleanUpRhoFunctions(PFunction func)
{
     GList *l, *j, *i;
     PBasicBlock block;
     PInst inst;
     bool_t removedSomething;

     do {
	  removedSomething = FALSE;
	  forEach(func->basicBlocks, l, block) {
	       GList *toRemove = NULL;
	       forEach(block->insts, j, inst) {
		    if (inst->type == OP_rho) {
			 GList *l;
			 PArg consumer;
			 PArg dst;
			 ArgIndex k;
			 bool_t dominatesAllConsumers = TRUE;

			 forEachDst(inst, k, dst) {
			      forEach(dst->consumers, l, consumer) {
				   if (g_list_length(consumer->producers) > 1) {
					dominatesAllConsumers = FALSE;
					break;
				   }
			      }
			      if (dominatesAllConsumers == FALSE) {
				   break;
			      }
			 }

			 if (dominatesAllConsumers) {
			      rhoHoistInputs(inst);
			      dfgRemoveInstEdges(inst);
			      toRemove = g_list_append(toRemove, inst);
			 }
		    }
	       }
	       forEach(toRemove, i, inst) {
		    basicBlockRemoveInst(block, inst);
		    instFree(inst);
		    removedSomething = TRUE;
	       }
	       g_list_free(toRemove);
	  } 

     } while (removedSomething);
}

void rhoSplitInputs(PFunction func)
{

     GList *l, *j;
     PBasicBlock block;
     PInst inst;

     forEach(func->basicBlocks, l, block) {
	  forEach(block->insts, j, inst) {
	       if (inst->type == OP_rho) {
		    PArg takenOutput;
		    PArg notTakenOutput;
		    GList *l;
		    PBasicBlock takenBlock, notTakenBlock; 
		    PArg consumer;
		    PInst branch;
		    GList *copy;

		    branch = basicBlockExitInst(block);
		    if (instIsCondBranch(branch)) {
			 dfgSmartAddEdge(branch, OUTPUT_C, inst, INPUT_B);
		    }
		    
		    takenOutput = instGetDstByType(inst, OUTPUT_A);
		    
		    takenBlock = basicBlockGetTakenExitEdge(block)->dst; 
		    notTakenBlock = basicBlockGetNotTakenExitEdge(block)->dst; 
		    
		    notTakenOutput = instAddDst(inst, takenOutput->reg, OUTPUT_B);

		    copy = g_list_copy(takenOutput->consumers);
		    forEach(copy, l, consumer) {
			 if (notTakenBlock == consumer->owner->owner) {
			      dfgRemoveEdge(takenOutput, consumer);
			      dfgAddEdge(notTakenOutput, consumer);
			 }
		    }
		    g_list_free(copy);
	       }
	  }
     }
}


void rhoCleanUpAndFixRhoFunctions(PFunction func)
{
     rhoSplitInputs(func);
     rhoRemoveDummyCopies(func);
     rhoCleanUpRhoFunctions(func);
}


void rhoInsertRhoFunctions(PFunction func)
{
     GList *l;
     PBasicBlock block;
     PInst lastInst;

     forEach(func->basicBlocks, l, block) {

	  lastInst = basicBlockExitInst(block);
	  if (lastInst == NULL) {
	       continue;
	  }

	  //	  fprintf(stderr, "Examining %d\n",  block->id);
	  if (instIsCondBranch(lastInst)) {	  
/* 	       PBasicBlock taken, notTaken; */
	       int reg;

/* 	       taken = basicBlockGetTakenExitEdge(block)->dst; */
/* 	       notTaken = basicBlockGetNotTakenExitEdge(block)->dst; */
	       
	       for(reg = 0; reg < REGISTER_COUNT; reg++) {
		    PInst n; 
		    if (reg == 31) {
			 continue;
		    }

		    n = instNew(OP_rho);
		    instAddSrc(n, reg, INPUT_A);
		    instAddDst(n, reg, OUTPUT_A);
		    basicBlockAppendInst(block, n);
	       }
	  }
     }
}

void rhoRemoveDummyCopies(PFunction func)
{
     PBasicBlock block;
     PInst inst;
     GList *copy;
     GList *i, *l;
     
     forEach(func->basicBlocks, i, block) {
	  copy = g_list_copy(block->insts);
	  forEach(copy, l, inst) {
	       if (inst->type == OP_identity) {
		    PArg input, output, producer, consumer;
		    GList *k, *j;

		    input = instGetSrcByType(inst, INPUT_A);
		    output = instGetDstByType(inst, OUTPUT_A);

		    forEach(input->producers, k, producer) {
			 forEach(output->consumers, j, consumer) {
			      dfgRemoveEdge(producer, consumer);
			      dfgAddEdge(producer, consumer);
			 }
		    }

		    dfgRemoveInstEdges(inst);
		    basicBlockRemoveInst(block, inst);
		    instFree(inst);
	       }
	  }
	  g_list_free(copy);
     }
}

void rhoRemoveUselessInstructions(PFunction func)
{
     PBasicBlock block;
     PInst inst;

     bool_t removedSomething;
     GList  *l ,*j;
     

     rhoResetInterestingFlags(func);
     
     do {
	  removedSomething = FALSE;
	  //	  fprintf(stderr, "Removed from function %s\n", func->name);
	  forEach(func->basicBlocks, l, block) {
	       GList *copy = g_list_copy(block->insts);
	       
/* 	       fprintf(stderr, "Removed from block  %d\n", block->id); */
/* 	       fprintf(stderr, "List is : \n"); */
/* 	       forEach(copy, j, inst) { */
/* 		    fprintf(stderr, " -> %d\n", inst->id); */
/* 	       } */
	       forEach(copy, j, inst) {

		    enum md_opcode op;
		    MD_SET_OPCODE(op, inst->binInst);

		    bool_t isStupid = FALSE;
		    if (op == TRAPB ||
			op == MB ||
			//			op == WMB ||
			op == ECB ||
			op == EXCB ||
			op == WH64 ||
			op == FETCH ||
			op == FETCH_M) isStupid = TRUE;


		    if (inst->type != OP_TYPE_ALPHA_OP || isStupid
			) {
			 
			 if (isStupid || ( !inst->isUnremovablePhiFunc && 
					   !rhoReachesSomethingInteresting(inst) && 
					   inst->tagHandShakeMaster == NULL)) {
			      
			      dfgRemoveInstEdges(inst);
			      basicBlockRemoveInst(block, inst);
			      /* 			      fprintf(stderr, "Removed %d\n", inst->id); */
			      instFree(inst);
			      removedSomething = TRUE;
			 }
		    }
	       } 

	       g_list_free(copy);	  
	  }
	       
     } while (removedSomething);

}

void rhoInsertDummyCopies(PFunction func)
{
     GList *l;
     PBasicBlock block;
     int reg;

     forEach(func->basicBlocks, l, block) {
	  
	  for(reg = 0; reg < REGISTER_COUNT ; reg++) {
	       PInst n; 	       

	       if (reg == 31) {
		    continue;
	       }

               if (cfgOptions.dynamicInstructionsByEpoch) {
                    if (!block->isLoopTop) {
                         n = instNew(OP_identity);
                         instAddSrc(n, reg, INPUT_A);
                         instAddDst(n, reg, OUTPUT_A);
                         basicBlockPrependInst(block, n);             
                    }          
               }
	  }
     }
}


void rhoFixControlArgs(PFunction func)
{
     GList *l, *i;
     PBasicBlock block;
     PInst inst;

     forEach(func->basicBlocks, i, block) {
	  forEach(block->insts, l, inst) {
	       ArgIndex k;
	       PArg src;
	       bool_t foundRealInput = FALSE;
	       if (inst->type == OP_ir) {
		    continue;
	       }
	       forEachSrc(inst, k, src) {
		    if (src->reg != 31 &&
			src->reg != -1) {
			 foundRealInput = TRUE;
			 break;
		    }
		    if (src->isZero) {
			 src->reg = CONTROL_REGISTER;
			 foundRealInput = TRUE;
			 break;
		    }
	       }
	       
	       if (!foundRealInput && inst->type != OP_func_start) {
		    instAddSrc(inst, CONTROL_REGISTER, INPUT_A);
	       }
	  }
     }
}

void rhoOptimizeRhoFunctions(PFunction func)
{
     GList *l, *i;
     PBasicBlock block;
     bool_t hoistedSomething;
     PInst inst;
     bool_t print = FALSE;
     do {
	  hoistedSomething = FALSE;
	  forEach(func->basicBlocks, i, block) {
	       forEach(block->insts, l, inst) {
		    if (inst->type == OP_rho) {
			 GList *k, *j;
		
			 PArg takenArg = instGetDstByType(inst, OUTPUT_A), 
			      notTakenArg = instGetDstByType(inst, OUTPUT_B);
			 
			 PArg takenConsumer, notTakenConsumer;
			 GList *c1, *c2;
		    
			 c1 = g_list_copy(takenArg->consumers);
			 forEach(c1, k, takenConsumer) {
			      c2 = g_list_copy(notTakenArg->consumers);
			      forEach(c2, j, notTakenConsumer) {
				   if (print) fprintf(stderr, "Checking %lu %lu %lu\n", inst->id, takenConsumer->owner->id, notTakenConsumer->owner->id);
				   if (takenConsumer == notTakenConsumer) {
					GList *m;
					PArg rhoInput = instGetSrcByType(inst, INPUT_A);
					PArg rhoProducer;

					dfgRemoveEdge(takenArg, takenConsumer);
					dfgRemoveEdge(notTakenArg, notTakenConsumer);
					
					forEach(rhoInput->producers, m, rhoProducer) {
					     dfgAddEdge(rhoProducer, notTakenConsumer);
					}
					hoistedSomething = TRUE;
				   }
			      }
			      g_list_free(c2);
			 }
			 g_list_free(c1);
		    }
	       }
	  }
     } while(hoistedSomething);
}

