#include "set.h"
#include <stdlib.h>
#include "util.h"
#include <assert.h>
#include "string.h"

PSet setNew(int universeSize)
{
     PSet s;
     nice_calloc(s, sizeof(Set));
     s->universeSize = universeSize;
     s->maxElementLoc = 0;
     s->epoch = 1;
     nice_calloc(s->elements, sizeof(int) * universeSize);
     return s;
}

void setFree(PSet s)
{
     free(s->elements);
     free(s);
}

void setAdd(PSet s, int id)
{
     assert(id < s->universeSize);
     s->maxElementLoc = MAX(s->maxElementLoc, id);
     s->elements[id] = s->epoch;
}

void setRemove(PSet s, int id) 
{
     assert(id < s->universeSize);
     s->elements[id] = 0;
}

bool_t setElementOf(PSet s, int id)
{
     assert(id < s->universeSize);
     return s->elements[id] == s->epoch;
}

void setCopy(PSet dst, PSet src)
{
     assert(dst->universeSize == src->universeSize);
     memcpy(dst->elements, src->elements, dst->universeSize *sizeof(guint32));
     
     dst->epoch = src->epoch;
     dst->maxElementLoc = src->maxElementLoc;
}

bool_t setCombineUnion(PSet dst, PSet src)
{
     guint32 i;
     guint32 needToCheck;
     bool_t changed = FALSE;

     needToCheck = MAX(dst->maxElementLoc, src->maxElementLoc);

     assert(dst->universeSize == src->universeSize);
     
     for (i = 0; i <= needToCheck; i++) {
	  if (dst->elements[i] == dst->epoch ||
	      src->elements[i] == src->epoch) {

	       if (dst->elements[i] != dst->epoch) {
		    dst->elements[i] = dst->epoch;
		    changed = TRUE;
	       }

	  }
     }
     return changed;
}

void setUnion(PSet dst, PSet src1, PSet src2)
{
     guint32 i;
     guint32 needToCheck;
     
     needToCheck = MAX(dst->maxElementLoc, src1->maxElementLoc);
     needToCheck = MAX(needToCheck, src2->maxElementLoc);

     assert(dst->universeSize == src1->universeSize);
     assert(dst->universeSize == src2->universeSize);

     for (i = 0; i <= needToCheck; i++) {
	  if (src1->elements[i] == src1->epoch ||
	      src2->elements[i] == src2->epoch) {
	       dst->elements[i] = dst->epoch;
	  } else {
	       dst->elements[i] = 0;
	  }
     }
}

void setIntersection(PSet dst, PSet src1, PSet src2)
{
     int i;
     guint32 needToCheck;
     
     needToCheck = MAX(dst->maxElementLoc, src1->maxElementLoc);
     needToCheck = MAX(needToCheck, src2->maxElementLoc);


     assert(dst->universeSize == src1->universeSize);
     assert(dst->universeSize == src2->universeSize);
     
     for (i = 0; i <= needToCheck; i++) {
	  if (src1->elements[i] == src1->epoch &&
	      src2->elements[i] == src2->epoch) {
	       dst->elements[i] = dst->epoch;
	       dst->maxElementLoc = i;
	  } else {
	       dst->elements[i] = 0;
	  }
     }
}

void setSubstract(PSet dst, PSet src1, PSet src2)
{
     int i;
     guint32 needToCheck;
     
     needToCheck = MAX(dst->maxElementLoc, src1->maxElementLoc);
     needToCheck = MAX(needToCheck, src2->maxElementLoc);

     assert(dst->universeSize == src1->universeSize);
     assert(dst->universeSize == src2->universeSize);

     for (i = 0; i < src1->universeSize; i++) {
	  if (src1->elements[i] == src1->epoch &&
	      src2->elements[i] != src2->epoch) {
	       dst->elements[i] = dst->epoch;
	       dst->maxElementLoc = i;
	  } else {
	       dst->elements[i] = 0;
	  }
     }
}


void setEmpty(PSet s)
{
     int i;

     if (s->epoch + 1 == 0)  {// overflow
	  for (i = 0; i < s->universeSize; i++) {
	       s->elements[i] = 0;
	  }
	  s->epoch = 1;
     } else {
	  s->epoch++;
     }
}

int  setCardinality(PSet s)
{
     int i;
     int count = 0;;

     for (i = 0; i <= s->maxElementLoc; i++) {
	  if (s->elements[i] == s->epoch) {
	       count++;
	  }
     }
     
     return count;
}

