#ifndef SET_INCLUDED
#define SET_INCLUDED
#include "host.h"

/* This is an implementation of a set of basicblocks that incorportates the
   optimizations outlined in Cryton's paper about SSA.

   An element i is in the set if elements[i] is equal to the epoch number.
   This allows us to clear the set by incrementing the epoch number.

   If you alays use DEFAULT_EPOCH as the epoch arg, the set will behave like a
   normal set.
*/

typedef struct _SET *PSet;


#include "util.h"
#include <glib.h>

#define forEachElement(set, v)  for(v = 0; v < (set)->universeSize; v++) if (setElementOf(set, v))

typedef guint32 SetElement;

typedef struct _SET {
     guint32 universeSize;
     guint32 *elements;
     guint32 epoch;
     guint32 maxElementLoc;
} Set;


PSet setNew(int universeSize);
void setFree(PSet s);
void setAdd(PSet s, int id);
void setRemove(PSet s, int id);
bool_t setElementOf(PSet s, int id);
void setUnion(PSet dst, PSet src1, PSet src2);
void setIntersection(PSet dst, PSet src1, PSet src2);
void setEmpty(PSet s);
int  setCardinality(PSet s);
bool_t setCombineUnion(PSet dst, PSet src);
void setCopy(PSet dst, PSet src);
void setSubstract(PSet dst, PSet src1, PSet src2);
#endif
