#include "ssa.h"
#include "list.h"
#include "set.h"
#include "util.h"
#include <assert.h>
#include "set.h"
#include "dot.h"
#include "buildInst.h"
#include  "dominance.h"

void ssaPrintCDG(PFunction func, FILE *f)
{
     char srcName[1024], dstName[1024];
     GList *l1, *l2;
     PBasicBlock src;
     PBasicBlock dst;
     char graphName[1024];

     sprintf(graphName, "%s: Control dependence", func->name);
     dotOpenGraph(f, graphName);

     forEach(func->basicBlocks, l1, src) {
	  fprintf(f, "%s;\n",  basicBlockName_DOT(src, srcName));
	  forEach(src->controlDependentNodes, l2, dst) {
	       fprintf(f, "%s -> %s [weight=100];\n",  basicBlockName_DOT(src, srcName), basicBlockName_DOT(dst, dstName)); 
	  }
     }
     dotCloseGraph(f);
}

/* Figure 11 from Cryton */
void ssaPlacePhiFunctions(PFunction func)
{
     int iterCount = 0;
     PBasicBlock block;
     GList *l;
     GList * W;
     SetElement V;
     SetElement Xindex;
     PBasicBlock X;

     forEach(func->basicBlocks, l, block) {
	  block->ssaHasAlready = 0;
	  block->ssaWork = 0;
     }

     W = NULL;
     
     forEachElement(func->usedVariables, V) {
	  PSet AofV;
	  iterCount++;
	  
	  AofV = g_hash_table_lookup(func->variablesToDefiningBlocks, (gpointer)V);
	  
	  forEachElement(AofV, Xindex) {
	       X = functionFindBasicBlockByLocalID(func, Xindex);
	       X->ssaWork = iterCount;
	       W = g_list_append(W, X);
	  }
	  
	  
	  while(!g_list_empty(W)) {
	       SetElement Yindex;
	       PBasicBlock Y;
	       GList *Xlink = g_list_first(W);
	       
	       W = g_list_remove_link(W, Xlink); // Remove the first element
	       X = (PBasicBlock)Xlink->data;
	       g_list_free_1(Xlink);
	       
	       forEachElement(X->dominance[DOMINANCE_DIRECTION_PRE].dominanceFrontier, Yindex) {

		    Y = functionFindBasicBlockByLocalID(func, Yindex);

		    if (Y->ssaHasAlready < iterCount) {
			 PInst newPhi = instNew(OP_phi);
			 int i, inDegree = g_list_length(Y->entrances);
			 for (i = 0; i < inDegree; i++)  
			      instAddSrc(newPhi, V, INPUT_OTHER);
			 instAddDst(newPhi, V, OUTPUT_C);
			 
			 basicBlockPrependInst(Y, newPhi);
			 Y->ssaHasAlready = iterCount;

			 if (Y->ssaWork < iterCount) {

			      Y->ssaWork = iterCount;
			      W = g_list_append(W, Y);
			 }
		    }
	       }
	  }
     }
}

typedef struct {
     GList *list;
} *PStack, Stack;

#define PUSH(s, v) ((s)->list = g_list_push((s)->list, v))
#define POP(s) ((s)->list = g_list_pop((s)->list))
#define TOP(s) (g_list_top((s)->list))


/* Figure 12 from Cryton */
int ssaRenameVariables_WHICH_PRED(PBasicBlock Y, PBasicBlock X)
{
     GList *l;
     PBasicBlock b;
     PCfgEdge bp;
     int c = 0;

     forEach(Y->entrances, l, bp) {
	  b = bp->src;
	  if (b == X) {
	       return c;
	  }
	  c++;
     }
     assert(0);
     return -1;
}

void ssaRenameVariables_SEARCH(PBasicBlock X, GHashTable *S) 
{
     GList *l, *k;
     ArgIndex j;
     PArg V,  dst;
     PCfgEdge exit;
     PBasicBlock Y;
     PInst inst;
     PSet seen = setNew(MAX_VARIABLES);

     forEach(X->insts, l, inst) {
	  if (inst->type != OP_phi) {
	       forEachSrc(inst, j, V) { 
		    if (V->reg != -1) {
			 PStack SofV = g_hash_table_lookup(S, (gpointer)V->reg);
			 assert(SofV);
			 if (!g_list_empty(SofV)) {
			      int newVariable = (int)TOP(SofV);
			      V->reg = newVariable;
			 } else {
			      ; // FIXME: the paper is ambiguous here.  It seems to
			      // assume that you always have something on SofV
			 }
		    }
	       }
	  }

	  setEmpty(seen); // maintaining this set is a bit of paranoid
			  // overkill.  The only time we will get into trouble
			  // is when the same register is a destination twice.
			  // I can't imagine this happening.
	  forEachDst(inst, j, dst) {
	       int newVariable = getUniqueVariable();
	       PStack SofV = g_hash_table_lookup(S, (gpointer)dst->reg);
	       if (dst->reg != -1) {
		    assert(!setElementOf(seen, dst->reg));
		    setAdd(seen, dst->reg);
		    dst->oldReg = dst->reg;
		    dst->reg = newVariable;
		    PUSH(SofV, (gpointer)newVariable);
	       }
	  }
     }

     
     forEach(X->exits, l, exit) {
	  int incomingEdge;
	  PInst inst;
	  Y = exit->dst;
	  
	  incomingEdge = ssaRenameVariables_WHICH_PRED(Y, X);
	  
	  forEach(Y->insts, k, inst) {
	       if (inst->type == OP_phi) {
		    PArg V = SRC_N(inst, incomingEdge);
		    PStack SofV = g_hash_table_lookup(S, (gpointer)V->reg);
		    if (!g_list_empty(SofV)) {
			 int newVariable = (int)TOP(SofV);
			 V->reg = newVariable;
		    } else {
			 // FIXME: the paper is ambiguous here.  It seems to
			 // assume that you always have something on SofV
			 assert (0); 
		    }
	       }
	  }
     }

     forEach(X->dominance[DOMINANCE_DIRECTION_PRE].dominates, l, Y) {
	  ssaRenameVariables_SEARCH(Y, S);
     }

     forEach(X->insts, l, inst) {
	  setEmpty(seen); // More paranoia
	  forEachDst(inst, j, dst) {
	       if (dst->oldReg != -1) {
		    PStack SofV = g_hash_table_lookup(S, (gpointer)dst->oldReg);
		    assert(SofV);
		    assert(!setElementOf(seen, dst->oldReg));
		    setAdd(seen, dst->oldReg);
		    POP(SofV);
	       }
	  }
     }
}

void ssaConvertPhiFunc(PInst inst) 
{
#if (0)
     // This code is for converting phi functions with the dfg in place.
     PBasicBlock block;
     GList *j, *i, *k;
     PCfgEdge path;
     PArg dstArg, srcArg, producer, consumer;
     GList *toDelete = NULL;
     
     dstArg = instGetDstByType(inst, OUTPUT_C);
     
     forEach(inst->srcs, i, srcArg) {
	  
	  forEach(srcArg->producers, j, producer) {
	       forEach(dstArg->consumers, k, consumer) {
		    dfgAddEdge(producer, consumer);
	       }
	  }
     }

     forEach(inst->srcs, i, srcArg) {
	  toDelete = NULL;
	  forEach(srcArg->producers, j, producer) {
	       toDelete = g_list_append(toDelete, producer);
	  }
	  forEach(toDelete, j, producer) {
	       dfgRemoveEdge(producer, srcArg);
	  }
	  g_list_free(toDelete);
     }
     
     toDelete = NULL;
     forEach(dstArg->consumers, k, consumer) {
	  toDelete = g_list_append(toDelete, consumer);
     }

     forEach(toDelete, k, consumer) {
	  dfgRemoveEdge(dstArg, consumer);
     }
     
     g_list_free(toDelete);

     // This is the old rotten code for converting phi-functions into copies.  
#else
     // this code is for doing it without the dfg
     PBasicBlock block;
     GList *l, *i;
     PCfgEdge path;
     PBasicBlock prev;
     PArg src;

     assert(inst->type == OP_phi);

     block = inst->owner;
     
     i = inst->srcs;
     
     //     fprintf(stderr, "converting Phi func %d.\n", inst->id);

     forEach(block->entrances, l, path) {
	  PInst copy;
	  PArg phiDst;
	  PInst last;

	  src = (PArg)i->data;
	  phiDst = instGetDstByType(inst, OUTPUT_C);
	  prev = path->src;
	  
	  copy = instNew(OP_identity);
	  copy->binInst = 0xdeadbeaf;

	  instAddSrc(copy, src->reg, INPUT_A);
	  instAddDst(copy, phiDst->reg, OUTPUT_C);
	
	  last = basicBlockExitInst(prev);
	  if (last != NULL) {
	       basicBlockInsertInstBeforeInst(prev, last, copy);
	  } else {
	       basicBlockAppendInst(prev, copy);
	  }
	       
	  i = i->next;
     }

     assert(i == NULL);// the list of entrances and inputs should be the same
		       // length.
#endif

}

static PBasicBlock ssaSearchUp(PBasicBlock b, PSet visited, PSet targets)
{
     GList *l;
     PCfgEdge e;

/*      fprintf(stderr, "Traversing %d\n", b->id); */
     
     if (targets != NULL && setElementOf(targets, b->localID)) {
	  return b;
     }

     setAdd(visited, b->localID);

     forEach(b->entrances, l, e) {
	  if (!setElementOf(visited, e->src->localID)) {
	       return ssaSearchUp(e->src, visited, targets);
	  }
     }
     return NULL;
}

/* find the branch responsible for controling the phi-functions in the
   function.  This code isn't quite right, it only handles properly nested
   diamonds, so if-elseif-elsif... won't work right. */
void ssaFindBranchesForPhiFunctions(PFunction func)
{
     GList *i, *j;
     PBasicBlock block;
     PInst inst;

     forEach(func->basicBlocks, i, block) {
	  forEach(block->insts, j, inst) {
	       if (inst->type == OP_phi) {
		    if (g_list_length(inst->srcs) == 2) {
			 PSet firstReached = setNew(func->lengthBasicBlocks + 1);
			 PSet secondReached = setNew(func->lengthBasicBlocks + 1);
/* 			 PArg a1 = (PArg)inst->srcs->data; */
/* 			 PArg a2 = (PArg)inst->srcs->next->data; */
			 PCfgEdge e1 = (PCfgEdge)block->entrances->data;
			 PCfgEdge e2 = (PCfgEdge)block->entrances->next->data;
			 PBasicBlock in1 = e1->src;
			 PBasicBlock in2 = e2->src;
			 PBasicBlock controlSource;
			 PInst controlSourceInst;
			 
			 /* This finds the nearest common ancestor of both
			    incoming basic blocks */ 

/* 			 fprintf(stderr, "Block %d\n", block->id); */
/* 			 fprintf(stderr, "First Search %d\n", in1->id); */
			 ssaSearchUp(in1, firstReached, NULL);
/* 			 fprintf(stderr, "Second Search %d \n", in2->id); */
			 controlSource = ssaSearchUp(in2, secondReached, firstReached);
			 controlSourceInst = basicBlockExitInst(controlSource);

			 if (controlSourceInst != NULL && 
			     instIsType(controlSourceInst, F_COND | F_FPCOND)) {
			      int v;
			      PArg src;
			      inst->phiFunctionControl = controlSourceInst;
			      
			      src = instGetDstByType(controlSourceInst, OUTPUT_OUT);

			      if (src == NULL) {
				   v = getUniqueVariable();
				   instAddDst(controlSourceInst, v, OUTPUT_OUT);
			      } else {
				   v = src->reg;
			      }
			      instAddSrc(inst, v, INPUT_PREDICATE);
			 } else {
			      inst->isUneededPhiFunction = TRUE;
			 }
/* 			 fprintf(stderr, "Control for %d from %d\n", inst->id, controlSource->id); */
			 
			 setFree(secondReached);
			 setFree(firstReached);
		    }
	       }
	  }
     }
}

void ssaConvertFromSSA(PFunction func)
{
     GList *i, *j;
     PBasicBlock block;
     PInst inst;
     PInst delete = NULL;

     forEach(func->basicBlocks, i, block) {
	  forEach(block->insts, j, inst) {
	       if (delete) {
		    basicBlockRemoveInst(block, delete);
		    delete = NULL;
	       }

	       if (inst->type == OP_phi) {
		    ssaConvertPhiFunc(inst);
		    delete = inst;
	       } else {
		    break;
	       } 
	  }
	  if (delete) {
	       basicBlockRemoveInst(block, delete);
	       delete = NULL;
	  }
     }
}


void ssaRenameVariables(PFunction func) 
{
     SetElement V;
     GHashTable *S = g_hash_table_new(g_direct_hash, g_direct_equal);

     forEachElement(func->usedVariables, V) {
	  g_hash_table_insert(S,(void*)V, calloc(sizeof(Stack), 1));	  
     }

     ssaRenameVariables_SEARCH(functionFindBasicBlockByLocalID(func, 1), S);
}

void ssaConvertToSSA(PFunction func)
{

     functionComputeVariableSets(func);
     ssaPlacePhiFunctions(func);
     ssaRenameVariables(func);
     //     ssaFindBranchesForPhiFunctions(func);
}
