#ifndef WRITE_ONCE_INCLUDED
#define WRITE_ONCE_INCLUDED

#define TRUE 1
#define FALSE 0
#include "host.h"

#include <assert.h>

/*
 * These are some macros to implement write-once variables.
 * 
 * After they are declared, they should be initialized exactly once (there is
 * no way to enforce this, since C doesn't gaurentee anything about initial
 * values), and then accessed only using these commands.  If a second attempt
 * to write the value is made, it will throw an assertion.
 *
 */

#ifdef DONT_ENFORCE_WRITE_ONCE

#warn WRITE_ONCE is not enforced

#define WRITE_ONCE_DECLARE(t, v) t v
#define WRITE_ONCE_INIT(v)   
#define WRITE_ONCE_WRITTEN(v)
#define WRITE_ONCE_SET(v, val) (v)=(val)

#define WRITE_ONCE_CHECK(v)   
#define WRITE_ONCE_ACCESS(v)  v

#else

#define WRITE_ONCE_DECLARE(t, v) t v; t v##_COPY; bool_t v##_WRITTEN
#define WRITE_ONCE_INIT(v)      v##_WRITTEN = FALSE
#define WRITE_ONCE_WRITTEN(v)  (v##_WRITTEN = TRUE, v##_COPY = v)
#define WRITE_ONCE_SET(v, val) (v = (v##_WRITTEN ? \
                                   (assert(FALSE), (typeof(v))0) : \
                                   (v##_WRITTEN = TRUE, v##_COPY = (val))))
#define WRITE_ONCE_CHECK(v)    (assert(!v##_WRITTEN || v##_COPY == v))
#define WRITE_ONCE_ACCESS(v)   (WRITE_ONCE_CHECK(v), v)

#endif

#endif
