/**
 * Implementation of barriers for the thread package that uses
 * memory primitives.
 *
 * $Id: mbarrier.c,v 1.4 2003/11/02 21:53:24 schwerin Exp $
 */

#include <memthreads/mthreads.h>
#include <memthreads/mbarrier.h>
#include <assert.h>
#include <stdio.h>
//#define assert(val) (val || (fprintf(stdout, "Assertion failed: " #val " at line %d\n", __LINE__), abort()));

extern void fprintq(FILE *f, volatile const thread_queue_t *q);

void mbarrier_initialize(volatile mbarrier_t *barrier, wsint64 max_count)
{
  TQ_INITIALIZE(&barrier->waiters);
  barrier->max_count = max_count;
  barrier->count = 0;
  barrier->lock = 0;
}

void mbarrier_reset_impl(volatile mbarrier_t *barrier)
{
  mtid_t mtid;
  thread_queue_t *waiters = &barrier->waiters;

  while(!TQ_EMPTY(waiters)) {
    mtid = TQ_DEQUEUE(waiters);
    MT_SPIN_SIGNAL(mtid) = 1;
  }
  barrier->count = 0;
  spinlock_release(&barrier->lock);
}

void mbarrier_wait(volatile mbarrier_t *barrier)
{
  mtid_t mtid;
  spinlock_acquire(&barrier->lock);
  ++barrier->count;
  if(barrier->count == barrier->max_count) {
    mbarrier_reset_impl(barrier);
  }
  else {
    mtid = mthread_get_tid();
    TQ_ENQUEUE(&barrier->waiters, mtid);
    spinlock_release(&barrier->lock);
    spinwait(&MT_SPIN_SIGNAL(mtid));
  }
}
