#include <memthreads/mthreads.h>

#ifndef NULL
#define NULL 0L
#endif

#ifndef false
#define false 0L
#endif

#ifndef true
#define true 1L
#endif

/*
 * This code is adapted from Mellor and Crummey's 1991 article in ACM TOCS, in
 * the 9th page of the article, Algorithm 5.
 */

void mcs_lock_initialize(volatile mcs_lock_t *lock)
{
  *lock = NULL;
}

void mcs_lock_acquire(volatile mcs_lock_t *lock)
{
//  const wsint64 mtid = mthread_get_tid();
  mcs_qnode_t *i = &MT_MCS_QNODE(mthread_get_tid());
  mcs_qnode_t *predecessor;
//  printf("Acquiring lock in thread %d\n", mtid);
  i->next = NULL;
  predecessor = atomic_swap(lock, i);
  if(NULL != predecessor) {
//    printf("Spinning waiting for lock in thread %d\n", mtid);
    i->locked = true;
    predecessor->next = i;
    while(i->locked); // Replace w/ call to spinwait function if performance is poor
  }
//  printf("Lock acquired in thread %d\n", mtid);
}

void mcs_lock_release(volatile mcs_lock_t *lock)
{
//  const wsint64 mtid = mthread_get_tid();
  mcs_qnode_t *i = &MT_MCS_QNODE(mthread_get_tid());
  if(NULL == i->next) { // no known successor
    if(compare_and_swap(lock, i, NULL)) { // c&s returns true iff it swapped
//      printf("No known successor to tid %d\n", mtid);
      return;
    }
//    printf("Waiting on succesor to tid %d\n", mtid);
    while(NULL == i->next); // Replace w/ call to spinwait function if performance is poor
//    printf("Successor to tid %d ready %d\n", mtid);
  }
  i->next->locked = false;
}
