/**
 * Implementation of queue locks for the thread package that uses
 * memory primitives.
 *
 * $Id: qlock.c,v 1.10 2003/11/02 19:24:29 schwerin Exp $
 */

#include <memthreads/mthreads.h>
#include <memthreads/qlock.h>
#include <stdio.h>

#include <assert.h>
//#define assert(val) (val || (fprintf(stdout, "Assertion failed: " #val " at line %d\n", __LINE__), abort()));

void fprintq(FILE *f, volatile const thread_queue_t *q)
{
  wsint64 i = q->head;
  fprintf(f, "TQ(%lld, %lld)[ ", q->head, q->tail);
  while(i != q->tail) {
    i = TQ_INCR(i);
    fprintf(f, "%lld ", q->queue[i]);
  }
  fputs("]\n", f);
}

void qlock_initialize(volatile qlock_t *qlock)
{
  qlock->slock = 0;
  qlock->isBusy = 0;
  TQ_INITIALIZE((&qlock->tqueue));
//  fprintq(stdout, &qlock->tqueue);
}

void qlock_acquire(volatile qlock_t *qlock)
{
  mtid_t tid = mthread_get_tid();
//  fprintf(stdout, "acquiring lock in thread %lld\n", tid);
  spinlock_acquire(&qlock->slock);
//  fprintq(stdout, &qlock->tqueue);
  if(qlock->isBusy) {
    TQ_ENQUEUE((&qlock->tqueue), tid);
//    fprintf(stdout, "%lld: qlock is busy; enqueued\n", tid);
//    fprintq(stdout, &qlock->tqueue);
    spinlock_release(&qlock->slock);
    spinwait(&MT_SPIN_SIGNAL(tid));
//    fputs("signaled!d", stdout);
    spinlock_acquire(&qlock->slock);
    MT_SPIN_SIGNAL(tid) = 0;
  } else {
    qlock->isBusy = 1;
  }
  assert(qlock->isBusy);
  spinlock_release(&qlock->slock);
//  fprintf(stdout, "%lld: lock acquired\n", tid);
}

void qlock_release(volatile qlock_t *qlock)
{
  mtid_t next_tid;
  spinlock_acquire(&qlock->slock);
//  fputs("Releasing!", stdout);
//  fprintq(stdout, &qlock->tqueue);
  if(!TQ_EMPTY((&qlock->tqueue))) {
    next_tid = TQ_DEQUEUE((&qlock->tqueue));
    MT_SPIN_SIGNAL(next_tid) = 1;
  } else {
    assert(TQ_EMPTY((&qlock->tqueue)));
    qlock->isBusy = 0;
  }
  spinlock_release(&qlock->slock);
}
