
#include <threads/defs.h>
#include <asm/procs.h>

#define SPINLOCK_ACQUIRE_IMPL(RW_OUT, LOCKADDR, RW_IN)     \
.scope_explicit                                          | \
.import LOCKADDR as lockaddr_in, RW_IN as rw_in          | \
.import RW_OUT as rw_out                                 | \
.declare_arcs rw_wa, la_wa, result_wa, isBusy_wa         | \
.declare_arcs lockaddr_loop, rw_loop                     | \
    cwa rw_wa, { rw_in, rw_loop }            <?, 1, 2>.0 | \
    wa  la_wa, { lockaddr_in, lockaddr_loop }            | \
                                                         | \
    ldq_l    ldresult_wa, la_wa, 0           <1, 2, 3>.1 | \
    NEI      ldIsBusy_wa, ldresult_wa, 0                 | \
    rho      {lockaddr_loop, lab_wa}, lanb_wa, la_wa, ldIsBusy_wa  | \
    rho      rw_loop, rwnb_wa, rw_wa, ldIsBusy_wa        | \
    mnop     lab_wa                          <2, 3, 1>.1 | \
                                                         | \
    test_and_set result_wa, lanb_wa          <2, 3, 1>.1 | \
    NEI      isBusy_wa, result_wa, 0                 | \
                                                         | \
    rho   lockaddr_loop, { },    lanb_wa, isBusy_wa      | \
    rho   rw_loop,       rw_out, rwnb_wa, isBusy_wa      | \
.endscope

#define SPINLOCK_RELEASE_IMPL(RW_OUT, LOCKADDR, RW_IN)     \
.scope_explicit                                          | \
.import RW_OUT as rw_out                                 | \
.import LOCKADDR as lockaddr_in, RW_IN as rw_in          | \
.declare_arcs rw_wa, la_wa, zero_wa                      | \
    cwa  rw_wa, rw_in                        <?, 1, 2>.0 | \
    wa   la_wa, lockaddr_in                              | \
                                                         | \
    cnst zero_wa, la_wa, 0                               | \
    stq  zero_wa, la_wa, 0                   <1, 2, 3>.1 | \
    mnop_ping rw_out, rw_wa                  <2, 3, 1>.2 | \
.endscope

.section TSECTION

PROC_START(spinlock_acquire)
.declare_arcs rw_out
    SPINLOCK_ACQUIRE_IMPL(rw_out, OUTPUT_A0, RENDEVOUS_WAVE)
    wtd cw_out, rw_out
    dtw cw_rw, rw_out, cw_out
    dtw ra_rw, RENDEVOUS_WAVE, RA
    dtw ra_out, cw_rw, ra_rw
PROC_END(spinlock_acquire, rw_out, rw_out, ra_out)

PROC_START(spinlock_release)
.declare_arcs rw_out
    SPINLOCK_RELEASE_IMPL(rw_out, OUTPUT_A0, RENDEVOUS_WAVE)
    wtd cw_out, rw_out
    dtw cw_rw, rw_out, cw_out
    dtw ra_rw, RENDEVOUS_WAVE, RA
    dtw ra_out, cw_rw, ra_rw
PROC_END(spinlock_release, rw_out, rw_out, ra_out)

;;;
;;; procedure spinlock_acquire_ws
;;;
;;; Inputs:  ra_in       -- the return address
;;           lockaddr_in -- the address of the lock
;;;          rw_in       -- the rendevous wave of the caller
;;;
;;; Outputs: cw_rw       -- the current wave on the rendevous wave.
;;;
;;; Acquires the lock at lockaddr_in, and then returns.
;;;

.declare_labels spinlock_acquire_ws
.scope_explicit_import spinlock_acquire_ws
.import spinlock_acquire_ws
.declare_arcs rw_out
spinlock_acquire_ws:
    func_start ra_in
    func_start lockaddr_in
    func_start rw_in

    SPINLOCK_ACQUIRE_IMPL(rw_out, lockaddr_in, rw_in)

    wtd   cw_out,  rw_out
    dtw   cw_rw,   rw_out, cw_out
    dtw   ra_rw,   rw_in,  ra_in
    isret cw_rw,   ra_rw,  0
.endscope

;;; end procedure spinlock_acquire_ws

;;;
;;; procedure spinlock_release_ws
;;;
;;; Inputs:  ra_in       -- the return address
;;           lockaddr_in -- the address of the lock
;;;          rw_in       -- the rendevous wave of the caller
;;;
;;; Outputs: cw_rw       -- the current wave on the rendevous wave.
;;;
;;; Marks the lock at lockaddr_in as free, then returns.
;;;

.declare_labels spinlock_release_ws
.scope_explicit_import spinlock_release_ws
.import spinlock_release_ws
spinlock_release_ws:

    func_start ra_in
    func_start lockaddr_in
    func_start rw_in

.declare_arcs rw_out
    SPINLOCK_RELEASE_IMPL(rw_out, lockaddr_in, rw_out)

    wtd   cw_out, rw_out
    dtw   cw_rw, rw_out, cw_out
    dtw   ra_rw, rw_in, ra_in
    isret cw_rw, ra_rw, 0
.endscope

;;; end procedure spinlock_release_ws
