;;; FILE:        mutex.S
;;; AUTHOR:      Andrew Petersen (petersen@cs.washington.edu)
;;; 

#ifndef __MUTEX_IMPL
#define __MUTEX_IMPL

#include <asm/procs.h>
#include <threads/defs.h>

.section TSECTION
.version 4

mutex_tc:
	tc { }, { }, { }
mutex_is:
	is ^mutex_tc:0, { }, 0


PROC_START(mutex_create)
;.declare_labels recv_next_mutex
.declare_arcs mutex_id
GET_NEXT_TID(mutex_id, RA)
.export ^mutex_create as mutex_create_local

;;; Call mutex_release here to initialize the lock if 
;;; the lock is to be released on creation

PROC_END_NOMEM(mutex_create, mutex_id, RENDEVOUS_WAVE, RA)


PROC_START(mutex_acquire)
.declare_labels acquired
.import mutex_tc
.import mutex_is
	;;; OUTPUT_A0 is T:w:k, where k is the lock ID (thread)
	nop ^mutex_tc:1, OUTPUT_A0
	cnst ^mutex_is:1, OUTPUT_A0, ^acquired

acquired:
	ir acquired_token

        ;;; Keep the return address from being forwarded until the 
        ;;; token is acquired.  This is to keep the instructions following
        ;;; the lock from proceeding.
	wait RA_OUT, RA, acquired_token
PROC_END_NOMEM(mutex_acquire, acquired_token, RENDEVOUS_WAVE, RA_OUT)


PROC_START(mutex_release)
.import mutex_tc
	;;; OUTPUT_A0 is T:w:k where k is the lock ID (thread)
	;;; The wave number is not important 
	;;; OUTPUT_A1 is T:w:d where d is the lock data

       cwa waRA, RA <?, 1, 2>.0
       wa  waRW, RENDEVOUS_WAVE
       wa  waA0, OUTPUT_A0
       wa  waA1, OUTPUT_A1
       mnop_ping exitRA, waRA <1, 2, 1>.1
	dtt ^mutex_tc:2, waA0, waA1
PROC_END_NOMEM(mutex_release, waA1, waRW, exitRA)

#endif
