#ifndef __BARRIER_IMPL
#define __BARRIER_IMPL
;;; FILE:        barrier.S
;;; AUTHOR:      Andrew Schwerin (schwerin@cs.washington.edu)
;;; DESCRIPTION: An implementation in wave-assembly of a barrier
;;;
;;; $Id: barrier.S,v 1.19 2003/11/02 21:10:00 schwerin Exp $
;;; 
;;; NOTE:
;;; The barrier request message field is either 0, to get behind the barrier,
;;; or it is < 0 to reset the barrier, or it is > 0 to reset the barrier and set the
;;; new threshold to the value of the message.  If the request message is -2, the
;;; barrier is destroyed after it is reset, rather than re-initialized.
;;; 

#include <asm/procs.h>
#include <asm/tid.h>
#include <threads/defs.h>
 
.section TSECTION
        
bwn_tc:
        tc     { }, { }, { }
bwn_rho:
        rho    { }, { }, ^bwn_tc:0, { }

PROC_START(barrier_thread)
.import bwn_tc, bwn_rho

/*
        cwa tdtoken, OUTPUT_RA <?, 1, 2>.0
        thread_destroy tdtoken <1, 2, 1>.1
*/

        nop    barrier_tid_in, OUTPUT_A0
        nop    max_count_in, OUTPUT_A1

;;; receive_barrier_request()
rbreq_tid_pad:
        nop    rbreq_tid_pad_OUT, rbreq_tid_pad_IN
rbreq_wave_pad: 
        nop    rbreq_wave_pad_OUT, rbreq_wave_pad_IN
rbreq_msg_pad: 
        nop    rbreq_msg_pad_OUT, rbreq_msg_pad_IN
;;; end of receive_barrier_request()

        cnst     { tcount_in, nextLink_wave_in, nextLink_tid_in }, max_count_in, 0
        wtd     caller_wave, RA

/*        wa      caller_wave_wa, {caller_wave, caller_wave_lc} */
        cwa      caller_wave_wa, {caller_wave, caller_wave_lc} <?, 1, 1>.0
        wa      barrier_tid_wa, { barrier_tid_in, barrier_tid_lc }
        wa      max_count_wa, { max_count_in, max_count_lc }
        wa      tcount_wa, { tcount_in, tcount_lc }
        wa      nextLink_wave_wa, { nextLink_wave_in, nextLink_wave_lc }
        wa      nextLink_tid_wa, { nextLink_tid_in, nextLink_tid_lc }
	 wtd     ^bwn_tc:2, barrier_tid_wa ; The put that allows this iter to receive

        ;; if(isBarrierMsg(req.msg))
        EQI     isBarrier, rbreq_msg_pad_OUT, 0
        rho     max_count_rp, { }, max_count_wa, isBarrier
        rho     tcount_rp, tcount_nrp, tcount_wa, isBarrier
        rho     nextLink_wave_rp, nextLink_wave_nrp, nextLink_wave_wa, isBarrier
        rho     nextLink_tid_rp, nextLink_tid_nrp, nextLink_tid_wa, isBarrier
        rho     rbreq_tid_pad_rp, { }, rbreq_tid_pad_OUT, isBarrier
        rho     rbreq_wave_pad_rp, { }, rbreq_wave_pad_OUT, isBarrier

        ;; isBarrierMsg(req.msg): TRUE ; data guided along this path ends "_rp"
        dttw    barrier_do_wait::rcl_tid_pad_IN, rbreq_tid_pad_rp, rbreq_wave_pad_rp, nextLink_tid_rp
        dttw    barrier_do_wait::rcl_wave_pad_IN, rbreq_tid_pad_rp, rbreq_wave_pad_rp, nextLink_wave_rp 

        AddI    new_tcount_rp, tcount_rp, 1
        cnst     zero_rp, new_tcount_rp, 0
        phi     tcount_loop_rp, zero_rp, new_tcount_rp, max_count_reached_rp

        ;; if(++tcount == max_tcount)
        EQ      max_count_reached_rp, new_tcount_rp, max_count_rp
        rho     rbreq_tid_mc, { }, rbreq_tid_pad_rp, max_count_reached_rp
        rho     rbreq_wave_mc, { }, rbreq_wave_pad_rp, max_count_reached_rp
        rho     new_tcount_mc, { }, new_tcount_rp, max_count_reached_rp

	 ;; send "chain_release(tcount -1)" to requestor
        SubI    decr_tcount_for_release, { new_tcount_mc, tcount_nrp }, 1
        GTI     are_some_to_release, {new_tcount_mc, tcount_nrp}, 0
        rho     release_count, { }, decr_tcount_for_release, are_some_to_release
        rho     release_tid, { }, {rbreq_tid_mc, nextLink_tid_nrp}, are_some_to_release
        rho     release_wave, { }, {rbreq_wave_mc, nextLink_wave_nrp}, are_some_to_release
        dttw    ^barrier_do_wait::rcr_length_pad:1, release_tid, release_wave, release_count

        ;; Not a barrier (isBarrierMsg(req.msg) == FALSE); some kind of reset
        cnst     tcount_loop_nrp, tcount_nrp, 0

	 ;; if(isSetNewMaxTCountMsg(req.msg))
        GTI     isChangeMax, rbreq_msg_pad_OUT, 0
        phi     new_max_count, rbreq_msg_pad_OUT, max_count_wa, isChangeMax

        ;; Loop exit test
        EQI     isDestroyBarrier, rbreq_msg_pad_OUT, -2
        rho     caller_wave_exit, caller_wave_lc, caller_wave_wa, isDestroyBarrier
        rho     { }, barrier_tid_lc, barrier_tid_wa, isDestroyBarrier
        rho     { }, max_count_lc, new_max_count, isDestroyBarrier
        rho     { }, tcount_lc, { tcount_loop_rp, tcount_loop_nrp }, isDestroyBarrier
        rho     { }, nextLink_wave_lc, rbreq_wave_pad_OUT, isDestroyBarrier
        rho     { }, nextLink_tid_lc, rbreq_tid_pad_OUT, isDestroyBarrier

	  ttd     curr_thread, caller_wave_exit
        wtd     curr_wave, caller_wave_exit
/*        thread_create        memready, curr_thread, curr_wave */
        dtw     curr_wave_in_caller_wave, caller_wave_exit, curr_wave
/*        dtw     RA_in_curr, curr_wave_in_caller_wave, RA */
        dtw     RA_exit, curr_wave_in_caller_wave, RA
/*        wait    RA_exit, RA_in_curr, memready */
        dtw     RENDEVOUS_WAVE_in_curr, curr_wave_in_caller_wave, RENDEVOUS_WAVE
PROC_END(barrier_thread, curr_wave, RENDEVOUS_WAVE_in_curr, RA_exit)

;;;
;;; Barrier resetting procedure
;;; 
;;; To change the barrier's max count as part of the reset process, pass
;;; the new max count as the second parameter.  Any value greater than
;;; zero is considered to be a valid new max count.
;;; 

;.proc   barrier_reset ( bBarrierReset_tid_in, new_max_count )
;.export ^barrier_reset as barrier_reset
PROC_START(barrier_reset)
.import bwn_rho, bwn_tc
        nop {bBarrierReset_tid_in, ^bwn_tc:1}, OUTPUT_A0
        nop new_max_count, OUTPUT_A1
        ;; Get the wave next wave number in the barrier thread
        cnst    ^bwn_rho:3, bBarrierReset_tid_in, 0
        ;; Construct a reset message
        wtd     req_wave, bBarrierReset_tid_in
        ttd     req_tid, bBarrierReset_tid_in
        EQI     isNewMaxCountZero, new_max_count, 0
        cnst    minusone, new_max_count, -1
        phi     req_max_count, minusone, new_max_count, isNewMaxCountZero
        ;; Send the reset message
        dttw    barrier_thread::rbreq_tid_pad_IN, bBarrierReset_tid_in, ^bwn_rho:1, req_tid
        dttw    barrier_thread::rbreq_wave_pad_IN, bBarrierReset_tid_in, ^bwn_rho:1, req_wave
        dttw    barrier_thread::rbreq_msg_pad_IN, bBarrierReset_tid_in, ^bwn_rho:1, req_max_count
PROC_END_NOMEM(barrier_reset, RENDEVOUS_WAVE, RENDEVOUS_WAVE, RA)

;;;
;;; procedure barrier_enq
;;; 
;;; Inputs:  RA       -- the return address,
;;;          bid      -- a barrier id
;;;          RW       -- the return rendevous wave
;;; Outputs: bwn      -- the wave on which to communicate w/ barrier bid
;;;          ew_in_rw -- The exit wave number in the rendevous wave
;;;

.declare_labels barrier_enq
.scope_explicit_import barrier_enq
.import barrier_enq
.import bwn_rho, bwn_tc
barrier_enq:
    func_start RA
    func_start barrier_id
    func_start RW

    cwa  { bid_wa, ^bwn_tc:1 }, barrier_id <?, 1, 2>.0
    wa   ena_wa, enq_notification_addr
    wa   RA_wa, RA
    wa   RW_wa, RW
    mnop_ping exitRA, RA_wa <1, 2, 1>.1

    cnst ^bwn_rho:3, bid_wa, -1
    wtd  cw_wa, RW_wa
    dtw  cw_rw, RW_wa, cw_wa
    dtw  ra_rw, RW_wa, exitRA

    isret ^bwn_rho:0, exitRA, 0
    isret cw_rw, ra_rw, 1
.endscope
;;; end procedure barrier_enq

;;;
;;; procedure barrier_do_wait
;;;
;;; Inputs:  RA
;;;          bid
;;;          bwn
;;;          RW
;;; Outputs: ew_in_rw
;;;
.declare_labels barrier_do_wait
.scope_explicit_import barrier_do_wait
.import barrier_do_wait
barrier_do_wait:
    func_start RA
    func_start bid
    func_start bwn
    func_start RW

;;; receive_chain_link() landing pads
rcl_tid_pad:
        nop    rcl_tid_pad_OUT, rcl_tid_pad_IN
rcl_wave_pad:
        nop    rcl_wave_pad_OUT, rcl_wave_pad_IN
;;; end of receive_chain_link() landing pads

;;; receive_chain_release() landing pad
rcr_length_pad:
        nop    rcr_length_pad_OUT, rcr_length_pad_IN
;;; end of receive_chain_release() landing pad

    ttd  req_tid,  bid
    wtd  req_wave, bid
    cnst req_msg,  bid, 0

    dttw barrier_thread::rbreq_tid_pad_IN,  bid, bwn, req_tid
    dttw barrier_thread::rbreq_wave_pad_IN, bid, bwn, req_wave
    dttw barrier_thread::rbreq_msg_pad_IN,  bid, bwn, req_msg

    ;; if(length) send_chainRelease(nextLink, length - 1)
    NEI     hasNext, rcr_length_pad_OUT, 0
    rho     length_hn, { }, rcr_length_pad_OUT, hasNext
    rho     nextLink_tid_hn, { }, rcl_tid_pad_OUT, hasNext
    rho     nextLink_wave_hn, { }, rcl_wave_pad_OUT, hasNext
    SubI    decr_length_hn, length_hn, 1
    dttw    rcr_length_pad_IN, nextLink_tid_hn, nextLink_wave_hn, decr_length_hn

    ;; exit code.  Caller must make subsequent instructions dependent on the return value,
    ;; ew_in_rw
    wait RA_OUT, RA, rcr_length_pad_OUT
    wtd  exit_wave, RA
    dtw ew_in_rw, RW, exit_wave
    dtw ra_in_rw, RW, RA_OUT

    isret ew_in_rw, ra_in_rw, 0
.endscope
;;; end procedure barrier_do_wait


PROC_START(barrier_wait)
.declare_labels barrier_enq_rtn, barrier_do_wait_rtn
.import barrier_enq, barrier_do_wait

    cnst barrier_enq_call_arc, RA, ^barrier_enq
    cnst barrier_enq_ra_arc, RA, ^barrier_enq_rtn
    wtd  curr_wave0, RA

    ;; Call to barrier_enq
    is barrier_enq_ra_arc, barrier_enq_call_arc, 0
    is OUTPUT_A0,          barrier_enq_call_arc, 1
    is curr_wave0,         barrier_enq_call_arc, 2
barrier_enq_rtn:
    ir bwn_enqd
    ir enqdwave_in_curr_wave0
    ;; end of return from call to barrier_enq

    cnst barrier_do_wait_call_arc, bwn_enqd, ^barrier_do_wait
    cnst barrier_do_wait_ra_arc, bwn_enqd, ^barrier_do_wait_rtn
    wtd  enqd_wave, bwn_enqd
    dtw  bid_enqd, enqdwave_in_curr_wave0, OUTPUT_A0
    dtw  rw_enqd,  enqdwave_in_curr_wave0, RENDEVOUS_WAVE
    dtw  ra_enqd,  enqdwave_in_curr_wave0, RA

    ;; Call to barrier_do_wait
    is barrier_do_wait_ra_arc, barrier_do_wait_call_arc, 0
    is bid_enqd,               barrier_do_wait_call_arc, 1
    is bwn_enqd,               barrier_do_wait_call_arc, 2
    is enqd_wave,              barrier_do_wait_call_arc, 3
barrier_do_wait_rtn:
    ir waited_wave_in_enqd_wave
    ;; end of return from call to barrier_do_wait

    dtw rw_waited, waited_wave_in_enqd_wave, rw_enqd
    dtw ra_waited, waited_wave_in_enqd_wave, ra_enqd
PROC_END_NOMEM(barrier_wait, rw_waited, rw_waited, ra_waited)
    

/*
;;; 
;;; Get behind the barrier; come on, everyone's doing it.
;;; 
;.proc barrier_wait(bBarrierWait_tid_in)
;.export ^barrier_wait as barrier_wait
PROC_START(barrier_wait)
.import bwn_rho, bwn_tc
        cwa    { bBarrierWait_tid_in, ^bwn_tc:1 }, OUTPUT_A0 <?, 1, 2>.0
        wa     advRENDEVOUS_WAVE, RENDEVOUS_WAVE
        wa     advRA, RA
        mnop_ping exitRA, advRA <1, 2, 1>.1
;;; receive_chain_link() landing pads
rcl_tid_pad:
        nop    rcl_tid_pad_OUT, rcl_tid_pad_IN
rcl_wave_pad:
        nop    rcl_wave_pad_OUT, rcl_wave_pad_IN
;;; end of receive_chain_link() landing pads

;;; receive_chain_release() landing pad
rcr_length_pad:
        nop    rcr_length_pad_OUT, rcr_length_pad_IN
;;; end of receive_chain_release() landing pad

        ;; Get the next wave number in the barrier thread
        cnst     ^bwn_rho:3, bBarrierWait_tid_in, -1
        ;; Construct a barrier request message
        wtd     req_wave, bBarrierWait_tid_in
        ttd     req_tid, bBarrierWait_tid_in
        cnst    req_msg, bBarrierWait_tid_in, 0
        ;; Send the barrier request message
        dttw    barrier_thread::rbreq_tid_pad_IN, bBarrierWait_tid_in, ^bwn_rho:0, req_tid
        dttw    barrier_thread::rbreq_wave_pad_IN, bBarrierWait_tid_in, ^bwn_rho:0, req_wave
        dttw    barrier_thread::rbreq_msg_pad_IN, bBarrierWait_tid_in, ^bwn_rho:0, req_msg

	 ;; if(length) send_chainRelease(nextLink, length - 1)
        NEI     hasNext, rcr_length_pad_OUT, 0
        rho     length_hn, { }, rcr_length_pad_OUT, hasNext
        rho     nextLink_tid_hn, { }, rcl_tid_pad_OUT, hasNext
        rho     nextLink_wave_hn, { }, rcl_wave_pad_OUT, hasNext
        SubI    decr_length_hn, length_hn, 1
        dttw    rcr_length_pad_IN, nextLink_tid_hn, nextLink_wave_hn, decr_length_hn

	 ;; exit code.  Caller must make subsequent instructions dependent on the return value,
        ;; rcr_length_pad_out
;       wtd     RA_WAVE, RA
;       isret      RA_WAVE, RA, 0
;       isret      rcr_length_pad_OUT, RA, 1
        wait RA_OUT, exitRA, rcr_length_pad_OUT
PROC_END_NOMEM(barrier_wait, rcr_length_pad_OUT, advRENDEVOUS_WAVE, RA_OUT)
*/
#endif
