#ifndef __CONDVAR_IMPL
#define __CONDVAR_IMPL
;;; FILE: condvar.S
;;; AUTHOR: Andrew Schwerin (schwerin@cs.washington.edu)
;;; DESCRIPTION: Condition variable functions that cannot
;;; be written in C
;;; 
;;; $Id: condvar.S,v 1.3 2003/10/11 18:39:15 schwerin Exp $
;;;

#include <asm/procs.h>
#include <threads/defs.h>
#include <mutex/mutex.S>
#include <threads/barrier.S>

.section TSECTION

;;;
;;; OUTPUT_A0 is the condition variable (barrier)
;;; OUTPUT_A1 is the guard mutex
;;;
;;; Waits on the barrier and unlocks the mutex atomically
;;;
PROC_START(condvar_wait)
.import barrier_enq, barrier_do_wait, mutex_release, mutex_acquire
.declare_labels barrier_enq_rtn, barrier_do_wait_rtn, mutex_release_rtn, mutex_acquire_rtn

    cnst barrier_enq_call_arc, OUTPUT_A0, ^barrier_enq
    cnst barrier_enq_ra_arc, OUTPUT_A0, ^barrier_enq_rtn
    wtd  start_wave, RA

    ;; Call to barrier_enq
    is barrier_enq_ra_arc, barrier_enq_call_arc, 0
    is OUTPUT_A0,          barrier_enq_call_arc, 1
    is start_wave,         barrier_enq_call_arc, 2
barrier_enq_rtn:
    ir bwn_enqd
    ir enqd_wave_in_start_wave
    ;; end of return from call to barrier_enq

    dtw RA_enqd, enqd_wave_in_start_wave, RA
    dtw RW_enqd, enqd_wave_in_start_wave, RENDEVOUS_WAVE
    dtw bid_enqd, enqd_wave_in_start_wave, OUTPUT_A0
    dtw mutex_enqd, enqd_wave_in_start_wave, OUTPUT_A1
    wtd enqd_wave_enqd, bwn_enqd

    cnst mutex_release_call_arc, bwn_enqd, ^mutex_release
    cnst mutex_release_ra_arc, bwn_enqd, ^mutex_release_rtn
    cnst zero, bwn_enqd, 0
    
    CALL_NB(mutex_release_call_arc, mutex_release_ra_arc, zero, zero,
            mutex_enqd, zero, zero, zero, zero, zero,
            zero, zero, zero, zero, zero, zero,
            zero, zero, zero, zero, zero, zero,
            enqd_wave_enqd)
    RECV_RETURN(mutex_release_rtn, { }, { }, { }, { }, { }, { },
                mr_wave_in_enqd_wave)
    
    dtw RA_mr, mr_wave_in_enqd_wave, RA_enqd
    dtw RW_mr, mr_wave_in_enqd_wave, RW_enqd
    dtw bid_mr, mr_wave_in_enqd_wave, bid_enqd
    dtw mutex_mr, mr_wave_in_enqd_wave, mutex_enqd
    dtw bwn_mr, mr_wave_in_enqd_wave, bwn_enqd

    cnst barrier_do_wait_call_arc, RA_mr, ^barrier_do_wait
    cnst barrier_do_wait_ra_arc, RA_mr, ^barrier_do_wait_rtn
    wtd  mr_wave_mr, RA_mr

    ;; Call barrier_do_wait
    is barrier_do_wait_ra_arc, barrier_do_wait_call_arc, 0
    is bid_mr,                 barrier_do_wait_call_arc, 1
    is bwn_mr,                 barrier_do_wait_call_arc, 2
    is mr_wave_mr,                  barrier_do_wait_call_arc, 3
barrier_do_wait_rtn:
    ir bdone_wave_in_mr_wave
    ;; end call to barrier_do_wait

    dtw RA_bdone, bdone_wave_in_mr_wave, RA_mr
    dtw RW_bdone, bdone_wave_in_mr_wave, RW_mr
    dtw mutex_bdone, bdone_wave_in_mr_wave, mutex_mr

    cnst mutex_acquire_call_arc, mutex_bdone, ^mutex_acquire
    cnst mutex_acquire_ra_arc, mutex_bdone, ^mutex_acquire_rtn
    cnst zero_bdone, mutex_bdone, 0
    wtd  bdone_wave_bdone, mutex_bdone

    CALL_NB(mutex_acquire_call_arc, mutex_acquire_ra_arc, zero_bdone, zero_bdone,
            mutex_bdone, zero_bdone, zero_bdone, zero_bdone, zero_bdone, zero_bdone,
            zero_bdone, zero_bdone, zero_bdone, zero_bdone, zero_bdone, zero_bdone,
            zero_bdone, zero_bdone, zero_bdone, zero_bdone, zero_bdone, zero_bdone, 
            bdone_wave_bdone)
    RECV_RETURN(mutex_acquire_rtn, { }, { }, { }, { }, { }, { }, ma_wave_in_bdone_wave)
    dtw RA_ma, ma_wave_in_bdone_wave, RA_bdone
    dtw RW_ma, ma_wave_in_bdone_wave, RW_bdone
PROC_END(condvar_wait, RW_ma, RW_ma, RA_ma)
#endif
