;;; FILE:        threads.S
;;; AUTHOR:      Andrew Schwerin (schwerin@cs.washington.edu)
;;; DESCRIPTION: Thread library functions
;;; 
;;; $Id: threads.S,v 1.8 2003/10/19 02:04:53 schwerin Exp $
;;;

#include <asm/procs.h>
#include <threads/defs.h>
#include <mutex/mutex.S>

.section TSECTION

PROC_START(thread_get_tid)
        ttd tid, RA
PROC_END_NOMEM(thread_get_tid, tid, RENDEVOUS_WAVE, RA)

;;;
;;; OUTPUT_A0 is the PC of the function to run inside the thread
;;;
PROC_START(thread_create_pc)
.declare_labels thread_fn_rtn
.declare_arcs tid_in
        GET_NEXT_TID(tid_in, RA)
        cnst wave, RA, 0
        thread_create tid, tid_in, wave

        ;; Move all the damned values into (tid, wave)
        dttw a0_in_tid, tid, wave, OUTPUT_A0
        dttw a1_in_tid, tid, wave, OUTPUT_A1
        dttw a2_in_tid, tid, wave, OUTPUT_A2
        dttw a3_in_tid, tid, wave, OUTPUT_A3
        dttw a4_in_tid, tid, wave, OUTPUT_A4
        dttw a5_in_tid, tid, wave, OUTPUT_A5
        dttw oret_in_tid, tid, wave, OUTPUT_RET
        dttw ofp_in_tid, tid, wave, OUTPUT_FP
        dttw ora_in_tid, tid, wave, OUTPUT_RA
        dttw pv_in_tid, tid, wave, OUTPUT_PV
        dttw gp_in_tid, tid, wave, OUTPUT_GP
        dttw fp0_in_tid, tid, wave, OUTPUT_FP0
        dttw fp1_in_tid, tid, wave, OUTPUT_FP1
        dttw fpa0_in_tid, tid, wave, OUTPUT_FPA0
        dttw fpa1_in_tid, tid, wave, OUTPUT_FPA1
        dttw fpa2_in_tid, tid, wave, OUTPUT_FPA2
        dttw fpa3_in_tid, tid, wave, OUTPUT_FPA3
        dttw fpa4_in_tid, tid, wave, OUTPUT_FPA4
        dttw fpa5_in_tid, tid, wave, OUTPUT_FPA5
        ttd  tid_in_tid, a0_in_tid
        wtd  wave_in_tid, a0_in_tid
        cnst thread_fn_ra_arc, a0_in_tid, ^thread_fn_rtn
        sll_I tid_shifted_in_tid, tid_in_tid, 32
        bis_I sp_in_tid, tid_shifted_in_tid, 0x8000000000000000
        ;; CALL_PC_NB with the damned values
;;; SS The pv should hold the address of the function we are calling.
        CALL_PC_NB(a0_in_tid, thread_fn_ra_arc, oret_in_tid, ofp_in_tid,
                   tid_in_tid, a1_in_tid, a2_in_tid, a3_in_tid, a4_in_tid, a5_in_tid,
                   ora_in_tid, a0_in_tid, gp_in_tid, sp_in_tid, fp0_in_tid, fp1_in_tid,
                   fpa0_in_tid, fpa1_in_tid, fpa2_in_tid, fpa3_in_tid, fpa4_in_tid, fpa5_in_tid,
                   wave_in_tid)
        ;; RECV_RETURN the damned values
        RECV_RETURN(thread_fn_rtn, retval_in_new_wave, { }, { }, { }, { }, { }, { })
        ttd tid_in_new_wave,   retval_in_new_wave
        cwa tid_in_final_wave, tid_in_new_wave <?, 1, 1>.0
        wa  retval_in_final_wave, retval_in_new_wave
        wtd final_wave_in_final_wave, retval_in_final_wave
        cnst mutex_release_fn, tid_in_final_wave, FN_LABEL(mutex_release)
        cnst mutex_release_ra_arc, tid_in_final_wave, ^mutex_release_rtn
        cnst zero, tid_in_final_wave, 0
        CALL_NB(mutex_release_fn, mutex_release_ra_arc, zero, zero, tid_in_final_wave,
                retval_in_final_wave, zero, zero, zero, zero, zero, zero, zero, zero, zero, zero,
                zero, zero, zero, zero, zero, zero, final_wave_in_final_wave)
        RECV_RETURN(mutex_release_rtn, final_ret, { }, { }, { }, { }, { }, { })
        thread_destroy final_ret <?, 1, 1>.1
PROC_END_NOMEM(thread_create_pc, tid, RENDEVOUS_WAVE, RA)

;;; OUTPUT_A0 is the tid of the thread to detach
PROC_START(thread_detach)
        cnst mutex_acquire_fn, RA, FN_LABEL(mutex_acquire)
        cnst mutex_acquire_ra_arc, RA, ^mutex_acquire_rtn
        cnst zero, RA, 0
        wtd  curr_wave, RA
        CALL_NB(mutex_acquire_fn, mutex_acquire_ra_arc, zero, zero,
                OUTPUT_A0, zero, zero, zero, zero, zero, 
                zero, zero, zero, zero, zero, zero,
                zero, zero, zero, zero, zero, zero, 
                curr_wave)
        RECV_RETURN(mutex_acquire_rtn, { }, { }, { }, { }, { }, { }, { })
PROC_END_NOMEM(thread_detach, zero, RENDEVOUS_WAVE, RA)
