#!/usr/bin/perl -w

package Drip::Elf::Section;
use strict;
use Pebble::HashObject;
use Pebble::Stringy;
use Carp;
use Drip::Type qw( :ALL );
use Drip::Elf::ElfConsts qw( SHT SHF );

BEGIN {
    our @ISA = qw( Pebble::HashObject Pebble::Stringy );
    our %FIELDS = (
	'name' => undef,
	'type' => undef,
	'flags' => undef,
	'varname' => undef,
	'representation' => 'Drip::Type',
	'pad' => undef,
	'vartype' => undef,
    );
    __PACKAGE__->DECLARE_ACCESSORS;
}

sub initialize {
    my $self = shift;
    $self->name(shift);
    $self->type(shift);
    $self->flags(shift);
    $self->varname(shift);
    $self->representation(shift);
    $self->pad(0);
    $self->vartype('Pebble::ArrayList') if UNIVERSAL::isa($self->representation, 'Drip::Type::ListType');
    return $self;
}

sub size {
    my ($self, $generator) = @_;
    my $varname= $self->varname;
    my $size = $self->type == SHT->values->{NOBITS} ?
	$generator->$varname :
	$self->representation->sizeInstance($generator->$varname);
#    print STDERR "section ". $self->name ." size = " . $size . "\n";
    return $size;
}

sub entsize {
    my $self = shift;
    my $type = $self->representation;
    return $type->_itemType->entsize if UNIVERSAL::isa($type,'Drip::Type::ListType');
    return 0;
}

sub toString {
    return "section " . $_[0]->name . " (" . $_[0]->representation->toString() . ")";
}

1;
