#!/usr/bin/perl -w

package Drip::Elf::SectionDefinitions;
use Exporter;
use Drip::Elf::ElfDefs qw( :ALL );
use Drip::Elf::ElfConsts qw( :ALL );
use Drip::Elf::Sections qw( :ALL );
use Drip::Type qw( :ALL );
use Drip::InstructionFormat qw( :ALL );
use strict;
use Carp;

our @ISA = qw( Exporter );
our @EXPORT_OK = qw( SECTION_INDEX_TEXT SECTION_INDEX_DATA SECTION_INDEX_BSS SECTION_INDEX_STRTAB );
our @EXPORT_TAGS = ( ALL => \@EXPORT_OK );

BEGIN {
    foreach (qw( SHT::WS_verdef SHT::WS_functions SHT::WS_sharedarcs SHT::PROGBITS 
                 SHT::NOBITS SHT::REL SHT::SYMTAB SHT::STRTAB SHF::NONE SHF::ALLOC 
                 SHF::EXECINSTR SHF::WRITE )) {
	my ($enum, $name) = split("::", $_);
	eval qq%use constant % . $enum."_".$name . qq% => $enum->values->{$name};%;
	die $@ . " when importing $enum::$name\n" if $@;
    }
}

section('.version',  SHT_WS_verdef,  SHF_NONE, 
	'version', &UINT(LONG));
section('.text',     SHT_PROGBITS,   SHF_ALLOC | SHF_EXECINSTR,
	'encodedinstrs', &List(&EncodedInstruction));
section('.data',     SHT_PROGBITS,   SHF_ALLOC | SHF_WRITE,
	'encodeddata', &ByteString);
section('.bss',	     SHT_NOBITS,     SHF_ALLOC | SHF_WRITE,
        'bsssize', &UINT(LONG));
section('.rel.text', SHT_REL,        SHF_NONE,
	'textreloc', &Array(&Elf64_Rel));
section('.rel.data', SHT_REL,        SHF_NONE,
	'datareloc', &Array(&Elf64_Rel));
section('.sym',	     SHT_SYMTAB,     SHF_NONE,
	'symboltable', &Array(&Elf64_Sym));
section('.strtab',   SHT_STRTAB,     SHF_NONE,
	'stringtable', &ByteString);
section('.func',    SHT_WS_functions,  SHF_NONE,
	'functions', &Array(&ElfFunction));
section('.sarc',     SHT_WS_sharedarcs, SHF_NONE,
	'sharedarcs', &List(&EncodedSharedArc));

1;
