#!/bin/bash

PROTECT=__protect__

DEFS=$1\Defs.pm
CONSTS=$1\Consts.pm
DEFS_HEAD=__defs_head
CONSTS_HEAD=__consts_head
HEAD_TOP=__head_top
TEMP=__temp

CPP="/usr/bin/cpp -I$PEBBLE_DIR/Drip"

sed "/^# *define[ \t][ \t]*[_a-zA-Z][_a-zA-Z0-9]*[ \t]/{
  p;
  s/# *define[ \t][ \t]*\([_a-zA-Z][_a-zA-Z0-9]*\)[ \t].*$/const int $PROTECT\1 \1;/;}" \
| perl -we "# this code turns #define macros into perl subroutines
  open EXPORTS, qq{>>$DEFS_HEAD};
  while (\$line = <STDIN>) {
    if (\$line =~ m{^#\\s*define\\s+(\\w+)\\(((\\s*\\w+\\s*,?)*)\\)\\s+(\\S.*)\$}) {
      my \$fnname = \$1;
      my @vars = split /\\s*,\\s*/, \$2;
      my \$expansion = \$4; 
      foreach my \$var (@vars) { print STDOUT qq{#define \$var \\\$\$var\\n}; }
      print STDOUT qq(sub \$fnname {\\n);
      foreach my \$var (@vars) { print STDOUT qq{  my \$var = shift;\\n}; }
      print STDOUT qq(  return \$expansion;\\n\\}\\n);
      foreach my \$var (@vars) { print STDOUT qq{#undef \$var\\n}; }
      print EXPORTS qq{ \$1};
#     } elsif (\$line =~ m{^#\\s*define\\s+(\\w+)\\s+(\\S.*)\$}) {
#       print STDOUT \$line;
#       print STDOUT qq{const &int, singlequote__\$1__, \$1;};
    }
    print STDOUT \$line;
  }
  close EXPORTS
" \
| $CPP -D__WORDSIZE=64 \
| sed "{s/$PROTECT//;
        /^sub/,/^}/{    # make subroutines PERL-friendly
            # fix up hex literals
            s/0\(x\|X\)\([0-9a-fA-F]*\)/hex(qq{\2})/g;
            # remove typecasts
            s/(\( *\(unsigned\|signed\|short\|long\|char\|int\)\)* *\**)//g;
            s/\(static\|reinterpret\|const\|dynamic\)_cast<[^>]*>//g;
            s/( *Elf[^)]*)//g;  #HACK: remove type casts to Elf types
        };
   #HACK workaround: elf.h has apparent problem; resolves dependency
        /DT_PROCNUM/s/DT_MIPS_NUM/0x32/;
   #remove blank lines, comments
        /^#/d;/^$/d;
}" \
| ./eatstruct -h $DEFS_HEAD > $DEFS

echo "#!/usr/bin/perl -w

package Drip::Elf::ElfDefs;
use Exporter;
use Drip::Type qw( :ALL );
use Data::Dumper;
use strict;
use Carp;
use Pebble::Constants;

our @ISA = qw( Exporter );
our @EXPORT_OK = qw(" > $HEAD_TOP

echo ");
our %EXPORT_TAGS = ( ALL => \\@EXPORT_OK );

Drip::Type::include 'Drip::Elf::ElfConsts';
" \
| cat $HEAD_TOP $DEFS_HEAD - > $TEMP; mv $TEMP $DEFS_HEAD

# extract constants
sed -n '/^const/p' $DEFS > $CONSTS
sed -i -n '/^const/!p' $DEFS
echo "
1;" \
| cat $DEFS_HEAD $DEFS - > $TEMP; mv $TEMP $DEFS

# remove include guard
# put prefixes at front
# stable sort by prefix then numeric value
# remove prefixes
cat $CONSTS \
| sed "/'_/d" \
| sed "{#these are hacks that are specific to elf.h
s/ELF\(CLASS\|DATA\)\([^_]\)/ELF\1_\2/;
s/\(R\|SYMINFO\)_\([^_]*\)_/\1\2_/;
s/\([^_]*\)_1_/\11_/;
s/EI_/EI/;}" \
| sed "{/_/!s/^.*$/0 &/; # lines without underscores will not be turned into enums
        s/^.*, '\([^_]*\)_\(.*\)$/\1, '\2/;}" \
| sort -t , -k 1,1 -k 3n,3 -s \
| uniq \
| perl -we "# this code turns constants into enums
  @export_ok = ();
  while (\$line = <STDIN>) {
    if (\$line =~ m{^0 (.* '(.*)'.*)$}) { 
      # strip the leader off of constants
      print STDOUT qq{\$1\n};
      push @export_ok, \$2;
    } elsif (\$line =~ m{^(.*), '(.*)', (.*);}) {
      # make consecutive constants with the same prefix into enums
      \$prefix = \$1;
      push @export_ok, \$1;
      print STDOUT qq{typedef &ENUM('\$2',\\t\$3,\\n};
      while (\$line = <STDIN>) {
        if (\$line =~ m{^\$prefix, '(.*)', (.*);}) {
          (\$name, \$value) = (\$1, \$2);
          print STDOUT qq{\\t '};
          print STDOUT qq{_} if \$name =~ /^\d/;
          print STDOUT qq{\$name',\\t\$value,\\n};
        } else {
          last;
        }
      }
      print STDOUT qq{\\t), '\$prefix';\\n};
      redo if \$line;
    }
  }
  print STDOUT qq{\n1;\n};

  open EXPORTS, qq{>>$CONSTS_HEAD};
  print EXPORTS qq{ @export_ok };
  close EXPORTS
" \
> $TEMP; mv $TEMP $CONSTS

echo ");
our %EXPORT_TAGS = ( ALL => \\@EXPORT_OK );
" | cat $HEAD_TOP $CONSTS_HEAD - $CONSTS \
| sed 's/ElfDefs/ElfConsts/' > $TEMP; mv $TEMP $CONSTS

rm -f $DEFS_HEAD $CONSTS_HEAD $HEAD_TOP $TEMP

