#!/usr/bin/perl -w

package Drip::TabPrint;
use strict;
use Carp;
use Pebble::HashObject;

use constant DEFAULT_TABWIDTH = 4;
use constant DEFAULT_TABDEPTH = 1;

BEGIN {
    our @ISA = qw( Pebble::HashObject );
    our %FIELDS = (
        'ostream' => undef,
        'tabdepth' => undef,
        'tabwidth' => undef
    );
    __PACKAGE__->DECLARE_ACCESSORS;
}

sub initialize {
    my $self = shift;
    $self->ostream(shift);
    my $tw = shift;
    if(defined $tw) { $self->tabwidth($tw); }
    else { $self->tabwidth(DEFAULT_TABWIDTH); }
    my $td = shift;
    if(defined $td) { $self->tabdepth($td); }
    else { $self->tabdepth(DEFAULT_TABDEPTH); }
    return $self;
}

sub increaseTab {
    my $self = shift;
    my $incr = shift;
    $self->tabdepth($self->tabdepth() + $incr);
}

sub decreaseTab {
    my $self = shift;
    $self->increaseTab(- shift());
}

sub print {
    my $self = shift;
    my $spaces = " " x ($self->tabdepth() * $self->tabwidth);
    my $td = $self->tabdepth;
    s/\n/\n$spaces/g foreach @_;
    s/\t/$td/g foreach @_;
    $self->ostream()->print(@_);
}

sub printf {
    my $self = shift;
    $self->ostream->print(sprintf(@_));
}

1;
