#!/usr/bin/perl -w

package Drip::Type::ENUMType;
use Carp;
use Pebble::ArrayList;

BEGIN {
    our @ISA = qw( Drip::Type::INTType );
    our %FIELDS = (
        'values' => 'HASH',
	 'order' => 'Pebble::ArrayList'
		   );
    __PACKAGE__->DECLARE_ACCESSORS;
}

sub initialize
{
    my $self = shift;
    my $pkg = shift;
    $self = $self->SUPER::initialize( $pkg, Drip::Type::LONG );
    $self->values({ });
    $self->order(new Pebble::ArrayList);
    $self->setValues(@_);
    return $self;
}

sub setValues
{
    my $self = shift;
    my $nextval = 0;
    while(my $name = shift(@_)) {
        croak("ENUM already contains '$name'") if(defined $self->values->{$name});
        $self->order->push($name);
        $nextval = shift if (scalar @_) && isNum($_[0]);
        $self->values->{$name} = $nextval++;
    }
}

sub instanceDeclaration
{
    my $self = shift;
    my $name = shift;
    my $retval = "enum {";
    foreach my $i (@{$self->order}) {
	$retval .= "$i = " . $self->values->{$i} . ", ";
    }
    $retval .= "} $name";
    return $retval;
}

sub isNum($) { return ($_[0] =~ m/\d+/); }

sub getCFnNameStem { return undef; die 'ENUM'; return 'ENUM'; }


sub validate
{
    my $self = shift;
    my $value = shift;
    croak("undefined value") unless defined $value;
    $value = $self->values->{$value} if defined $self->values->{$value};
    return $self->SUPER::validate($value);
}

1;
