#!/usr/bin/perl -w

package Drip::Type::TypedefType;
use strict;
use Carp;

BEGIN {
    our @ISA = qw( Drip::Type );
    our %FIELDS = (
        'name' => undef,
        'basetype' => 'Drip::Type'
    );
    __PACKAGE__->DECLARE_ACCESSORS;
}

our $AUTOLOAD;

sub initialize
{
    my $self = shift;
    my $pkg = shift;
    $self = $self->SUPER::initialize($pkg);
    scalar @_ == 2 or croak("Wrong number of parameters");
    my $name = shift;
    my $basetype = shift;
    croak("Undefined name") unless defined $name;
    croak("Undefined base type for type '$name'") unless defined $basetype;
    $self->name($name);
    $self->basetype($basetype);
    return $self;
}

sub defaultValue { my $self = shift; return $self->basetype->defaultValue; }

sub readInstance
{
    my $self = shift;
    my $fh = shift;
    my $type = shift;
    $type or $type = $self;
    $self->basetype->readInstance($fh, $type);
}

sub generateCReadFn
{
    my $self = shift;
    my $fh = shift;
    my $type = shift;
    $type or $type = $self;
    $self->basetype->generateCReadFn($fh, $type);
}

sub getCFnNameStem
{
    my $self = shift;
    return $self->name;
}

sub instanceDeclaration {
    my $self = shift;
    my $name = shift;
    return $self->name . ($name ? " $name" : "");
}

sub typeDeclaration {
    my $self = shift;
    return "typedef " . $self->basetype->instanceDeclaration($self->name);
}

sub requiredTypes
{
    my $self = shift;
    return ( $self->basetype );
}

sub toString
{
    my $self = shift;
    return $self->name();
}

sub DESTROY { }

sub AUTOLOAD
{
    no strict 'refs';
    my $name = $AUTOLOAD;
    my $pkg = $AUTOLOAD;
    $name =~ s/.*://;
    $pkg =~ s/(::[^:]*?)$//x;

    my $toEval = qq(package $pkg;\
sub $name {\
    my \$self = shift;\
    my \$retval;\
    eval { \$retval = \$self->basetype->$name(\@_); };\
    if(\$@) { croak(\$@); }\
    return \$retval;\
};);
    eval $toEval;
    goto &$AUTOLOAD;
}

1;
