#!/usr/bin/env perl

use warnings;
use strict;

package LCTools::Description;
use IO::Handle;
use IO::File;
use Data::Dumper;
use Carp;
use Exporter;

our @ISA = qw( Exporter );
our @EXPORT_OK = qw(loadFromFileName loadFromHandle storeToFileName storeToHandle combineDescriptions);

sub loadFromFileName
{
    my $fname = shift;
    my $fh = IO::File->new;
    $fh->open("<$fname") or Carp::croak "Couldn't open $fname for reading";
    return loadFromHandle($fh);
}

sub loadFromHandle
{
    my $fh = shift;
    my $string = join(' ', <$fh>);
    my @retval = eval($string);
    $@ and Carp::croak $@;
    return @retval;
}

sub storeToFileName
{
    my ($filename, $ref, @rest) = @_;
    Carp::croak("Too many arguments") if scalar @rest;
    my $fh = IO::File->new;
    $fh->open(">$filename") or croak("Couldn't open $filename for writing");
    storeToHandle($fh, $ref);
}

sub storeToHandle
{
    my ($fh, $ref, @rest) = @_;
    Carp::croak("Too many arguments") if scalar @rest;
    local $Data::Dumper::Terse = 1;
    $fh->print("(", join(", ", Dumper($ref)), ")");
}

sub combineDescriptions
{
    my %primary = %{shift()};
    while(scalar @_) {
        my %subordinate = %{shift()};
        my $subordparams = ($subordinate{parameters} or [ ]);
        delete $subordinate{parameters};
        foreach my $key (keys %subordinate) {
            defined $primary{$key} or $primary{$key} = $subordinate{$key};
        }
        unshift @{$primary{parameters}}, @{$subordparams};
    }
    return \%primary;
}

sub storeToFileWithIncludes
{
    my ($fh, $desc, @includes) = @_;
    local $Data::Dumper::Terse = 1;

    print $fh "use LCTools::Description qw(loadFromFileName combineDescriptions);\n";
    print $fh ("combineDescriptions(",
               join(", ", (Dumper($desc),
                           map { "loadFromFileName('$_')" } @includes)), ");\n");
}

1;
