#!/usr/bin/env perl

#
# $Id: DoSimulate.pm,v 1.6 2004/06/14 04:34:26 swanson Exp $
#

use warnings;
use strict;

package LCTools::DoSimulate;

use Pebble::HashObject;
use LCTools::LockedFileQueue;

BEGIN {
    our @ISA=qw( Pebble::HashObject );
    our %FIELDS = ( 'resultsPrefix' => undef,
                    'resultsSuffix' => undef,
                    'jobq'          => 'LCTools::LockedFileQueue'
                  );
    __PACKAGE__->DECLARE_ACCESSORS;
}

sub initialize
{
    my $self = shift;
    %$self = @_;
    return $self;
}

sub generateResultsFileName
{
    my ($self, $app, $sim) = @_;
    return $self->resultsPrefix() . $app->{filename_part} . '-' . $sim->{filename_part} . $self->resultsSuffix();
}

sub generateCmdLine
{
    my ($self, $app, $sim) = @_;
    my $resultsFileName = $self->generateResultsFileName($app, $sim);
    defined $resultsFileName or $resultsFileName = "/dev/null";

    my $simulator = ($sim->{simulator} or die "Undefined simulator name");
    my @simparams = @{$sim->{parameters}};
    my $alphaexec = ($app->{alphaexec} or die "Undefined alpha executable");
    my $wsexec    = ($app->{wsexec} or die "Undefined wave scalar executable");
    my @appparams = @{$app->{parameters}};

    return ("$simulator " . join(" ", @simparams) .
            " -- $wsexec $alphaexec " . join(" ", @appparams));
}

sub generateBashCmd
{
  my ($self, $app, $sim) = @_;
  return ($self->generateCmdLine($app, $sim) . " >" .
          $self->generateResultsFileName($app, $sim) . " 2>&1" . ($app->{stdin} ? " < " . $app->{stdin} : ""));
}

sub generateCondorCmd
{
  my ($self, $app, $sim) = @_;
  return ("nohup condor_run \"".$self->generateBashCmd($app, $sim) . "\"");
}

sub simulate {
    my ($self, $app, $sim) = @_;
    my $todo = $self->generateBashCmd($app, $sim);
    print "$todo\n";
    return system $todo;
}

1;
