#!/usr/bin/perl -w

package Pebble::ASMLexer;

use Parse::Lex;
use Carp;
use Pebble::ArrayObject;

our @ISA = qw( Pebble::ArrayObject );

my $PLEXER        = 0;
#my $WAITINGTOKENS = 1;

my @tokens = (
    'PERL', [";perl", "(?:[^;].*?|;(?!endperl))+", ";endperl"], sub {
        $_[1] =~ s/;perl//m;
        $_[1] =~ s/;endperl//m;
        return $_[1];
    },
    qw~
       AS                         \bas\b
       ATMACRO                    \@macro\b
       ATENDMACRO                 \@endmacro\b
       DOTATTRS                   \.attrs\b
       DOTPROC                    \.proc\b
       DOTENDPROC                 \.endproc\b
       DOTSECTION                 \.section\b
       DOTENDSECTION              \.endsection\b
       DOTVERSION                 \.version\b
       DOTSHARE                   \.share\b
       DOTEXTERN                  \.extern\b
       IMPLICITSCOPETYPE          \.scope\b
       EXPLICITDECLARESCOPETYPE   \.scope_explicit_declare\b
       EXPLICITIMPORTSCOPETYPE    \.scope_explicit_import\b
       FULLYEXPLICITSCOPETYPE     \.scope_explicit\b
       ENDSCOPE                   \.endscope\b
       DOTDECLLABELS              \.declare_labels\b
       DOTDECLARCS                \.declare_arcs\b
       DOTIMPORT                  \.import\b
       DOTEXPORT                  \.export\b
       EOL                        \n
       NAME                       \b[_[:alpha:]]\w*\b
       HEXDIGITS                  \b0x[[:xdigit:]]+\b
       OCTDIGITS                  \b0[0-7]+\b
       SCOPETOK                   ::
       COLON                      :
       DOT                        \.
       QS                         \?
       LBRACE                     {
       RBRACE                     }
       CARET                      \^
       DOLLAR                     \$
       LT                         <
       EQ                         =
       GT                         >
       PLUS                       \+
       MINUS                      -
       DECDIGITS                  \b(?:[1-9]\d*|0)\b
       LPAREN                     \(
       RPAREN                     \)
       PIPE                       \|
      ~
    , 'COMMA'                     , ','
    , 'COMMENT'                   , '[;#].*?\n'
    ,qw(ERROR .*), sub { Carp::croak("lexing error: can't analyze \"$_[1]\" at line " .
                                     $_[0]->lexer->line . ":\n" . $_[0]->lexer->buffer); }
);

sub initialize
{
    my $self = shift;
    $self->_lexer(new Parse::Lex(@tokens));
    $self->_lexer->skip( '([ \t\r]|[\\\\]\n)+');
    return $self;
}

sub nextToken
{
    my $self = shift;
#    $self->_waitingtokens and return $self->nextWaitingToken;

    my $token = $self->_lexer->next;
    $self->_lexer->eoi and return ('', undef);

#    if($token->name eq 'PERL') {
#        $self->_waitingtokens( \$token->text );
#    }

    return ($token->name, $token->text);
}

sub from { shift->_lexer->from(@_); }

#sub nextWaitingToken
#{
#    my $self = shift;
#    my @token = ( $self->_waitingtokens->shift, $self->waitingtokens->shift );
#    scalar @$self->_waitingtokens or $self->_waitingtokens = undef;
#    return @token;
#}

sub _lexer
{
    my $self = shift;
    scalar @_ and $self->[$PLEXER] = shift;
    return $self->[$PLEXER];
}

#sub _waitingtokens
#{
#    my $self = shift;
#    scalar @_ and $self->[$WAITINGTOKENS] = shift;
#    return $self->[$WAITINGTOKENS];
#}

sub debug
{
    my $lexer = new Pebble::ASMLexer();
    $lexer->from(\*STDIN);
    while(my ($nam, $val) = $lexer->nextToken) {
        return if $nam eq '';
        print "($nam, $val)\n";
    }
}

#$lexer = Parse::Lex->new(@tokens);
#$lexer->skip( '([ \t]|[\\\\]\n)+');
#$lexer->from(\*STDIN);
#$lexer->every( sub { print $_[0]->name . "\t-->" . $_[0]->text . "<--\n"; } );

1;
