#!/usr/bin/perl -w

package Pebble::Arc;
use strict;
use Carp;
use Pebble::HashObject;
use Pebble::Operand;

BEGIN {
    our @ISA = qw( Pebble::HashObject Pebble::Operand );
    our %FIELDS = (
        'name'          => undef,
        '_producers'     => 'Pebble::ArrayList',
        '_consumers'    => 'Pebble::ArrayList'
    );
    __PACKAGE__->DECLARE_ACCESSORS;
}

sub initialize {
    my $self = shift;
    my $name = shift;
    $self->name($name);
    $self->_producers(new Pebble::ArrayList);
    $self->_consumers(new Pebble::ArrayList);
    return $self;
}

sub addProducer {
    my $self = $_[0];
    my $producer = $_[1];
    unless($self->_producers->contains($producer)) {
        $self->_producers->push($producer);
    } else {
        carp($producer->toString() . " already a producer of arc " . $self->toString);
    }
}

sub producers { return @{$_[0]->_producers}; }

sub addConsumer {
    my $self = shift;
    my $consumer = shift;
    unless($self->_consumers->contains($consumer)) {
        $self->_consumers->push($consumer);
    } else {
        carp($consumer->toString() . " already a consumer of arc " . $self->toString);
    }
}

# sub removeConsumer {
#     my $self = shift;
#     my $consumer = shift;
#     my $cidx = $self->_consumers->indexOf($consumer);
#     if($cidx >= 0) {
#         splice @{$self->_consumers}, $cidx, 1;
#     }
# }

sub consumers {
    my $self = shift;
    return @{$self->_consumers};
}

sub referingInstrInstances {
    return ($_[0]->consumers, $_[0]->producers);
}


sub toString { return $_[0]->name; }

1;
