#!/usr/bin/perl -w

package Pebble::ArcSet;
use strict;
use Carp;
use Pebble::HashObject;
use Pebble::Operand;
use Pebble::ArrayList;

BEGIN {
    our @ISA = qw( Pebble::HashObject Pebble::Operand );
    our %FIELDS = (
        '_arcs' => 'Pebble::ArrayList'
    );
    __PACKAGE__->DECLARE_ACCESSORS;
}

sub initialize
{
    my $set = shift;
    $set->_arcs(new Pebble::ArrayList);
    $set->addArc($_) foreach(@_);
    return $set;
}

sub arcs { return @{$_[0]->_arcs}; }

sub producers { return map { $_->producers; } $_[0]->arcs; }
sub consumers { return map { $_->consumers; } $_[0]->arcs; }

sub addArc
{
    my $set = $_[0];
    my $arc = $_[1];
    croak("$arc not an arc") unless(UNIVERSAL::isa($arc, 'Pebble::Arc'));
    $set->_arcs->push($arc) unless $set->contains($arc);
}

sub addArcSet
{
    my $thisSet  = $_[0];
    my $otherSet = $_[1];
    $thisSet->addArc($_) foreach $otherSet->arcs();
}

sub contains
{
    my $set = $_[0];
    my $arc = $_[1];
    return $set->_arcs->contains($arc);
}

sub toString {
    return "{ ". join(", ", map { $_->toString; } $_[0]->arcs) . " }";
}

1;
