#!/usr/bin/perl -w

package Pebble::AssemblyUnit;

use strict;
use Carp;
use Pebble::HashObject;
use Pebble::Section;
use Pebble::InstructionFactory;
use Pebble::ArrayList;
use Pebble::ArcSet;

BEGIN {
    our @ISA = qw( Pebble::HashObject );
    our %FIELDS = (
        '_sections'           => 'HASH',
        '_definedSections'    => 'Pebble::ArrayList',
        '_instructionFactory' => 'Pebble::InstructionFactory',
        '_exportedSymbols'    => 'HASH',
        '_externSymbols'      => 'HASH',
        '_sharedArcs'         => 'Pebble::ArcSet',
        '_labels'             => 'Pebble::ArrayList',
        'version'             =>  undef
    );
    __PACKAGE__->DECLARE_ACCESSORS;
}

sub initialize
{
    my $self = shift;

    $self->_sections({ });
    $self->_exportedSymbols({ });
    $self->_externSymbols({ });
    $self->_labels(new Pebble::ArrayList);
    $self->_instructionFactory (new Pebble::InstructionFactory);
    $self->_sharedArcs(new Pebble::ArcSet);
    return $self;
}

sub addExportSymbol {
    my ($self, $label, $name, @rest) = @_;
    croak("Two instances of export symbol $name") if defined $self->_exportedSymbols->{$name};
    $self->_exportedSymbols->{$name} = $label;
}

sub addExternSymbol {
    my ($self, $label, $name, @rest) = @_;
    croak("Two instances of extern symbol $name") if defined $self->_externSymbols->{$name};
    $self->_externSymbols->{$name} = $label;
}

sub addSharedArcs {
    my $self = shift;
    $self->_sharedArcs->addArc($_) foreach @_;
}

sub getSharedArcs {
    return $_[0]->_sharedArcs->arcs;
}

sub addLabel
{
    my $self = shift;
    my $label = shift;
    $self->_labels->push($label);
}


sub labels
{
    my $self = shift;
    return @{$self->_labels};
}


sub section
{
    my $self = shift;
    my $sectionName = shift;
    $self->_sections->{$sectionName} or
        $self->_sections->{$sectionName} = new Pebble::Section($sectionName, $self);
}

sub sections {
    my $self = shift;
    return @{$self->_definedSections};
}

sub setSections
{
    my $self = shift;
    my $definedSections = shift;
    my %sections = map { ( $_->name, $_ ) } @{$definedSections};
    $definedSections = new Pebble::ArrayList(values %sections);
    $self->_definedSections($definedSections);
    undef;
}

sub getMacroDef
{
    my $self = shift;
    return $self->_instructionFactory->getMacroDef(@_)
}

sub setMacroDef
{
    my $self = shift;
    $self->_instructionFactory->setMacroDef(@_);
}

sub instructionFactory { my $self = shift; return $self->_instructionFactory; }


sub visit { $_[1]->visitAssemblyUnit($_[0]); }

1;
