#!/usr/bin/perl -w

package Pebble::HashObject;

use strict;
use English;
use Pebble::Object;
use Carp;

BEGIN {
    our @ISA    = qw( Pebble::Object );
    our %FIELDS = ( );
}
our $AUTOLOAD;

sub new
{
    my $kind = shift;
    return $kind->_newImpl({ }, @_);
}

sub initialize
{
    my $self = shift;
    $self->SUPER::initialize(@_);
    return $self;
}

sub AUTOLOAD
{
    no strict "refs";
    my $fullname = $AUTOLOAD;
    my $self = $_[0];
    my $name = $fullname;
    my $pkg = $fullname;
    $name =~ s/.*://; # strip fully qualified portion
    $pkg =~ s/(::[^:]*?)$//x;

    if($name eq "DESTROY") {
        return undef;
    }
    if(my $accessor = FIND_ACCESSOR($pkg, $name)) {
        goto $accessor;
    }
    croak("No such method: $fullname");
}

sub FIND_ACCESSOR
{
    my $pkg = shift;
    my $field = shift;
    my $fullname = "${pkg}::${field}";
#    carp("Declaring $fullname");
    ref($pkg) and $pkg = ref($pkg);
    no strict "refs";
    my $accessor = *$fullname{CODE};
    return $accessor if $accessor;
    my $fields = \%{"${pkg}::FIELDS"};
    use strict 'refs';
    if( exists( $fields->{$field} ) )
    {
        my $type = $fields->{$field};
        my $toEval = qq( \
package $pkg;\
sub $field {\
     my \$self = \$_[0];\
#     scalar \@_ > 2 and croak("Too many parameters to $field!");\
     if(2 == scalar \@_) { \
         my \$newval = \$_[1];\
     );
     if(defined $type) {
        $toEval .= qq(    if( \$newval && !UNIVERSAL::isa(\$newval, '$type') ) {\
              croak("Field '$field' must have type $type, not " . ref(\$newval));\
         } \
); 
     }
        $toEval .= qq(         return \$self->{$field} = \$newval;\
     }\
     return \$self->{$field};
           } );

        eval $toEval;
        if($@) { die $@; }
        no strict 'refs';
        $accessor = *{$fullname};
        use strict 'refs';
    } else {
        no strict 'refs';
        ($accessor = FIND_ACCESSOR($_, $field) and last) foreach(@{"${pkg}::ISA"});
    }
    return $accessor;
}

sub DECLARE_ACCESSORS
{
    no strict "refs";
    my $pkg = shift;
    ref($pkg) and $pkg = ref($pkg);
    foreach my $field (keys(%{"${pkg}::FIELDS"})) {
        $pkg->can($field) and croak("$pkg already has a field named '$field'");
        FIND_ACCESSOR($pkg, $field);
    }
}

BEGIN { __PACKAGE__->DECLARE_ACCESSORS; }

1;
