#!/usr/bin/perl -w

package Pebble::InstrDef;
use strict;
use Carp;
use Pebble::HashObject;


BEGIN {
    our @ISA = qw( Pebble::HashObject );
    our %FIELDS = (
        'opname' => undef
    );
    __PACKAGE__->DECLARE_ACCESSORS;
}

sub initialize {
    my $self = shift;
    my $opname = shift;
    $opname or croak("Opname must not be undefined!");
    $self->opname($opname);
    return $self;
}

sub instantiate
{
    return new Pebble::InstrInstance(@_);
}

sub validateParameters
{
    my $self = shift;
    my $operandsRef = shift;
    my @operands = @$operandsRef;
    my $waveordering = shift;
    $self->validateWaveOrdering($waveordering);

    for(my $i = 0; $i < $self->numArcOperands; $i++) {
        my $arcset = shift @operands;
        croak("Insufficient quantity of arc operands") unless $arcset;
        unless(UNIVERSAL::isa($arcset, 'Pebble::ArcSet')) {
            croak("Output and non-immediate input operands must be arcsets");
        }
    }
    if($self->HasImmediate()) {
        croak("Missing immediate operand") unless scalar @operands;
        my $immediate = shift @operands;

        croak("Expected immediate operand; found " . ref($immediate))
            unless UNIVERSAL::isa($immediate, 'Pebble::Immediate');
    }
    scalar @operands and croak("Too many operands");
}



sub toString { return $_[0]->opname; }

use Pebble::InstrDef::Alias;
use Pebble::InstrDef::Instr;
use Pebble::InstrDef::MemInstr;

1;
