#!/usr/bin/perl -w

package Pebble::InstrDef::Instr;
use strict;
use Carp;
use Pebble::HashObject;

use constant OUTPUT_PORT_NAME => ('OUTPUT_A', 'OUTPUT_B', 'OUTPUT_C');
use constant INPUT_PORT_NAME  => ('INPUT_A',  'INPUT_B',  'INPUT_C');

BEGIN {
    our @ISA = qw( Pebble::InstrDef );
    our %FIELDS = (
        'opcode'     => undef,
        'InputArcs'  => 'ARRAY',
        'OutputArcs' => 'ARRAY',
        'HasImmediate' => undef
    );
    __PACKAGE__->Pebble::HashObject::DECLARE_ACCESSORS;
}

sub initialize {
    my $self = shift;
    my $opname = shift;
    $self = $self->SUPER::initialize( $opname );
    return $self;
}

sub numInputArcs {
    my $self = shift;
    return scalar @{$self->InputArcs()} ;
}

sub numOutputArcs {
    my $self = shift;
    return scalar @{$self->OutputArcs} ;
}

sub numArcOperands {
    my $self = shift;
    return $self->numInputArcs() + $self->numOutputArcs();
}

sub numOperands {
    my $self = shift;
    return $self->numArcOperands() + ($self->HasImmediate ? 1 : 0);
}

sub getPortType {
    my $self = shift;
    my $portnum = shift;
    if($portnum < $self->numOutputArcs) {
        return (OUTPUT_PORT_NAME)[$portnum];
    } elsif($portnum < $self->numArcOperands) {
        return (INPUT_PORT_NAME)[$portnum - $self->numOutputArcs];
    }
    croak("Invalid portnum $portnum for instruction ", $self->opname);
}

sub validateWaveOrdering
{
    my $self = shift;
    my $waveordering = shift;
    if($waveordering) {
        croak("Wave ordering on non-memory instruction '" . $self->opname() . "'");
    }
}

sub isMemoryInstr { return undef; }

1;
