#!/usr/bin/perl -w

package Pebble::InstrInstance;
use strict;
use Carp;
use Pebble::CodeInstance;
use Pebble::OperandPort;
use Pebble::InstrDef;
use Pebble::ArrayList;

BEGIN {
    our @ISA = qw ( Pebble::CodeInstance );
    our %FIELDS = ( 'instrdef' => 'Pebble::InstrDef' );
    __PACKAGE__->DECLARE_ACCESSORS ;
}

sub initialize
{
    croak("Wrong number of parameters") unless 4 == scalar @_;
    my $self = shift;
    my $instrdef = shift;
    $self = $self->SUPER::initialize( @_ );
    $self->instrdef( $instrdef );
    $self->_setProducersAndConsumers;
    return $self
}

sub toString {
    my $self = shift;
    return join(", ", $self->instrdef()->opname(), map { $_->toString; } @{$self->operands});
}

sub _setProducersAndConsumers
{
    my $self = shift;
    my @operands = @{$self->operands};
    my $port = 0;
    for(my $i = 0; $i < $self->instrdef->numOutputArcs; $i++, $port++) {
        my $outArcSet = shift @operands;
        foreach my $arc ($outArcSet->arcs) {
            $arc->addProducer(new Pebble::OperandPort($self, $port));
        }
    }
    for(my $i = 0; $i < $self->instrdef->numInputArcs; $i++, $port++) {
        my $inArcSet = shift @operands;
        foreach my $arc ($inArcSet->arcs) {
            $arc->addConsumer(new Pebble::OperandPort($self, $port));
        }
    }
}

sub addArc
{
    my $self = shift;
    my $arc = shift;
    my $port = shift;
    croak("Port number out of range") if $port < 0 || $port >= $self->instrdef->numArcOperands;
    return if $self->operands->[$port]->contains($arc);
    $self->operands->[$port]->addArc($arc);
    if($port < $self->instrdef->numOutputArcs) {
        $arc->addProducer(new Pebble::OperandPort($self, $port));
    }
    else {
        $arc->addConsumer(new Pebble::OperandPort($self, $port));
    }
}


sub visit { return $_[1]->visitInstrInstance($_[0]); }

1;
