#!/usr/bin/perl -w

package Pebble::InstrLabel;
use strict;
use Pebble::HashObject;
use Pebble::InstrInstance;
use Pebble::ArrayList;
use Pebble::Arc;
use Carp;

BEGIN {
    our @ISA = qw( Pebble::HashObject Pebble::Operand );
    our %FIELDS = (
        'name'                => undef,
        '_isExtern'           => undef,
        '_instr'              => 'Pebble::InstrInstance',
        '_arcs'               => 'Pebble::ArrayList',
        '_refinstrs'          => 'Pebble::ArrayList', # of Pebble::InstrInstance objects
        'attributes'          => 'HASH'
    );
    __PACKAGE__->DECLARE_ACCESSORS;
}

sub initialize {
    my $self = shift;
    my $name = shift;
    my $asmunit = shift;
    croak("Undefined asmunit") unless defined $asmunit;
    $self->name($name);
    $self->_arcs(new Pebble::ArrayList);
    $self->_refinstrs(new Pebble::ArrayList);
    $asmunit->addLabel($self);
    return $self;
}


sub addReference
{
    my $refinstrs = $_[0]->_refinstrs;
    my $instr = $_[1];
    $refinstrs->push($instr) unless $refinstrs->contains($instr);
}


sub arcs { return @{$_[0]->_arcs}; }
sub nonArcReferences { return @{$_[0]->_refinstrs}; }

sub arcByLabel {
    my $self = shift;
    my $operandport = shift;
    my $arc = $self->_arcs->[$operandport];
    if(!defined $arc) {
        $arc = $self->_arcs->[$operandport] =
            new Pebble::Arc("^" . $self->name() . ":" . $operandport);
        if(defined $self->_instr) {
            $self->_instr->addArc($arc, $operandport);
        }
    }
    return $arc;
}

sub setExtern
{
    my $self = shift;
    croak("Cannot mark label " . $self->name() .
          " as extern because it already has a definition")
        if $self->_instr;
    $self->_isExtern(1);
}

sub setInstruction
{
    my $self = shift;
    croak("Label " . $self->name() . "already has a definition") if $self->_instr;
    croak("Attempt to define extern'd label " . self->name) if $self->_isExtern;
    $self->_instr ( shift );
    for(my $i = 0; $i < scalar @{$self->_arcs}; $i++) {
        $self->_instr->addArc( $self->_arcs->[$i], $i )
            if defined $self->_arcs->[$i];
    }
}

# True if _isExtern is true or _instr has a value; both should never be true
sub isSpecified { return $_[0]->_isExtern() || $_[0]->_instr(); }

sub getInstruction { return shift->_instr; }

sub toString {
    my $self = shift;
    return "^" . $self->name() . ": " . (defined $self->_instr() ? $self->_instr->toString()
                                         : "undef");
}

1;
