#!/usr/bin/perl -w

package Pebble::InstructionFactory;
use strict;
use Pebble::HashObject;
use Data::Dumper;
use Carp;
use Pebble::InstrDef;
use Pebble::InstrDef::InstructionDefs;
use Pebble::InstrInstance;
our @ISA = qw( Pebble::HashObject );

sub initialize
{
    my $self = shift;
    $self->{instrDefs} = { %Pebble::InstrDef::InstructionDefs::IDEFS };
    $self->{macroDefs} = { };
    return $self;
}

sub createInstruction
{
    my $self = shift;
    my $opname = shift;
    my $def = $self->getDef($opname);
    Carp::croak("Undefined operation $opname") unless $def;
    $def->validateParameters(@_);

    my $instr = $def->instantiate(@_);
    Carp::croak("Instantiation failed") unless defined $instr;
    return $instr;
}

sub getDef
{
    my $self = shift;
    my $name = shift;
    my $retval;
    $retval = $self->getInstrDef($name) or
        $retval = $self->getMacroDef($name);
    return $retval;
}

sub setMacroDef
{
    my $self = shift;
    my $macroname = shift or Carp::croak("Undefined macro name!");
    my $macro = shift or Carp::croak("Undefined macro definition!");

    $self->getInstrDef($macroname)
        and Carp::croak("Attempt to redefine instruction $macroname as a macro");

    $self->getMacroDef($macroname)
        and Carp::croak("Attempt to redefine macro $macroname");

    $self->_macrodefs->{$macroname} = $macro;
    return undef;
}

sub getMacroDef
{
    my $self = shift;
    my $macroname = shift;
    return $self->_macrodefs->{$macroname};
}

sub getInstrDef
{
    my $self = shift;
    my $instrname = shift;
    return $self->_instrdefs->{$instrname};
}


sub setInstrDef
{
    my $self = shift;
    my $instrname = shift or Carp::croak("Undefined instruction name!");
    my $instrdef = shift or Carp::croak("Undefined instruction definition!");

    $self->getInstrDef($instrname)
        and Carp::croak("Attempt to redefine instruction $instrname");

    $self->_instrdefs->{$instrname} = $instrdef;
    return undef;
}

sub _macrodefs { my $self = shift; return $self->{macroDefs}; }
sub _instrdefs { my $self = shift; return $self->{instrDefs}; }

1;
