#!/usr/bin/perl -w

use strict;
use Carp;
use Pebble::ArrayObject;
use Pebble::Namespace;

package Pebble::Name;
our @ISA = qw( Pebble::ArrayObject );

my $NAMESPACE = 0;
my $NAMESTR   = 1;

sub initialize
{
    my ($self, $namespace, $namestr, @rest) = @_;
    carp("Too many parameters to initialize") if scalar @rest;

    $self->namespace($namespace);
    $self->name($namestr);
    return $self;
}

sub namespace
{
    my $self = shift;
    scalar @_ and $self->[$NAMESPACE] = shift;
    return $self->[$NAMESPACE];
}

sub name
{
    my $self = shift;
    scalar @_ and $self->[$NAMESTR] = shift;
    return $self->[$NAMESTR];
}

sub declareLocally
{
    my $self = shift;
    $self->namespace->declareLocally($self->name);
    return undef;
}

sub isDeclaredLocally
{
    my $self = shift;
    return $self->namespace->isDeclaredLocally($self->name);
}

sub getByNameTyped
{
    my $self = shift;
    my $type = shift;
    return $self->namespace->getByNameTyped($self->name, $type);
}

sub arcByName { shift->getByNameTyped("Pebble::Arc"); }
sub labelByName { shift->getByNameTyped("Pebble::InstrLabel"); }

sub toString
{
    my $self = shift;
    return ($self->namespace ? $self->namespace->toString : 'UNDEF') . "::" . $self->name;
}

sub print
{
    my $self = shift;
    my $fh = \*STDOUT;
    scalar @_ and $fh = shift;

    print $fh $self->toString;
}

sub debug
{
    my $ns = new Pebble::Namespace undef;
    my $name = new Pebble::Name $ns, "foo";

    $name->arcByName;
    $name->arcByName;
    $name->labelByName;
}


package Pebble::FullyQualifiedName;
our @ISA = qw( Pebble::Name );

sub getByNameTyped
{
    my $self = shift;
    my $type = shift;
    return $self->namespace->getByNameTyped($self->name, $type,
                                            \&Pebble::Namespace::getByNameLocally);
}

1;
