#define THREAD_SEND_MSG(TARGET_INSTR, TARGET_THREAD, TARGET_WAVE, VAL) \
  {                                                                    \
    asm ( "dttw " #TARGET_INSTR "^1, " #TARGET_THREAD ", "             \
          #TARGET_WAVE " , " #VAL "\n" )

#define THREAD_SET_LANDING_PAD(LP) asm ( #LP ": send { }, { }\n" )

#define THREAD_LANDING_PAD_RESULT(LP) asm_arcbystring ( #LP "^0" )

typedef struct _Link {
  ThreadID tid;
  WaveID wave;
} Link;

typedef struct _BarrierRequest {
  Link link;
  BarrierMsg msg;
} BarrierRequest;

void barrier_thread(Barrier b, int max_tcount)
{
  BarrierRequest req;
  Link nextLink = { tid:0, wave:0 };
  int tcount = 0;
  
  for( ; ; ) {
    req = receive_barrier_request();
    if(isBarrierWaitMsg(req.msg)) {
      send_chain_link(req.link, nextLink);
      nextLink = req.link;
      if(++tcount == max_tcount) {
        send_chain_release(req.link, tcount-1);
        tcount = 0;
      }
    }
    else { // isn't a barrier message, so must be a release message
      send_chain_release(nextLink, tcount-1);
      tcount = 0;
      if(isDestroyBarrierMsg(req.msg)) return;
      if(isSetNewMaxTCountMsg(req.msg))
        max_tcount = getNewMaxTCount(req.msg);
    }
  }
}

int barrier_wait(Barrier b)
{
  BarrierRequest req;
  Link nextLink;
  int length;

  req.link.tid = CURR_THREAD;
  req.link.wave = CURR_WAVE;
  req.msg = BARRIER_MSG;
  send_barrier_request(b, req);
  nextLink = receive_chain_link();
  int length = receive_chain_release( );
  if(length) send_chain_release( nextLink, length - 1);
  return length;
}

void barrier_reset(Barrier b, int new_max)
{
  BarrierRequest req;
  req.link.tid = CURR_THREAD;
  req.link.wave = CURR_WAVE;
  req.msg = new_max ? BARRIER_RESET_NEWMAX_MSG(new_max) : BARRIER_RESET_MSG;
  send_barrier_request(b, req);
}

void barrier_destroy( Barrier b )
{
  BarrierRequest req;
  req.link.tid = CURR_THREAD;
  req.link.wave = CURR_THREAD;
  req.msg = BARRIER_DESTROY;
  send_barrier_request(b, req);
}

void send_barrier_request(Barrier b, BarrierRequest req)
{
  WaveID barrier_wave = take(b.tcaddr, b.tid);
  THREAD_SEND_MSG("rbreq_tid_pad", b.tid, barrier_wave, req.link.tid);
  THREAD_SEND_MSG("rbreq_wave_pad", b.tid, barrier_wave, req.link.wave);
  THREAD_SEND_MSG("rbreq_msg_pad", b.tid, barrier_wave, req.msg);
}

BarrierRequest receive_barrier_request(void)
{
  BarrierRequest retval;
  put(b.tcaddr, b.tid, CURR_WAVE);
  THREAD_SET_LANDING_PAD("rbreq_tid_pad");
  THREAD_SET_LANDING_PAD("rbreq_wave_pad");
  THREAD_SET_LANDING_PAD("rbreq_msg_pad");
  retval.link.tid = THREAD_LANDING_PAD_RESULT("rbreq_tid_pad");
  retval.link.wave = THREAD_LANDING_PAD_RESULT("rbreq_wave_pad");
  retval.msg = THREAD_LANDING_PAD_RESULT("rbreq_msg_pad");
  return retval;
}


void send_chain_link(Link curr, Link next)
{
  THREAD_SEND_MSG("rcl_tid_pad", curr.tid, curr.wave, next.tid);
  THREAD_SEND_MSG("rcl_wave_pad", curr.tid, curr.wave, next.wave);
}

Link receive_chain_link()
{
  Link retval;
  THREAD_SET_LANDING_PAD("rcl_tid_pad");
  THREAD_SET_LANDING_PAD("rcl_wave_pad");
  retval.tid = THREAD_LANDING_PAD_RESULT("rcl_tid_pad");
  retval.wave = THREAD_LANDING_PAD_RESULT("rcl_wave_pad");
  return retval;
}  

void send_chain_release(Link target, int length)
{
  THREAD_SEND_MSG("rcr_len_pad", target.tid, target.wave, length);
}

int receive_chain_release( )
{
  THREAD_SET_LANDING_PAD("rcr_len_pad");
  return THREAD_LANDING_PAD_RESULT("rcr_len_pad");
}
