;;; FILE:        barrier_a.s
;;; AUTHOR:      Andrew Schwerin (schwerin@cs.washington.edu)
;;; DESCRIPTION: An implementation in wave-assembly of the barrier described in
;;;              very C-like pseudocode in barrier_a.c
;;;
;;; $Id: barrier_a.s,v 1.6 2004/02/12 22:24:03 schwerin Exp $
;;; 
;;; NOTE:
;;; The barrier request message field is either 0, to get behind the barrier,
;;; or it is < 0 to reset the barrier, or it is > 0 to reset the barrier and set the
;;; new threshold to the value of the message.  If the request message is -2, the
;;; barrier is destroyed after it is reset, rather than re-initialized.
;;; 
 
.section .code

.share rbreq_tid_pad_OUT
.share nextLink_wave_in
.share nextLink_wave_iN
.share ^bwn_tc:2
        
bwn_tc:
        tc     { }, { }, { }
bwn_rho:
        rho    { }, { }, ^bwn_tc:0, { }

.proc barrier_thread ( barrier_tid_in, max_count_in )
.export ^barrier_thread as barrier_thread .attrs( localInvocationPC = 18032 )
;;; receive_barrier_request()
rbreq_tid_pad:
        identity    rbreq_tid_pad_OUT, rbreq_tid_pad_IN
rbreq_wave_pad: 
        identity    rbreq_wave_pad_OUT, rbreq_wave_pad_IN
rbreq_msg_pad: 
        identity    rbreq_msg_pad_OUT, rbreq_msg_pad_IN
;;; end of receive_barrier_request()

        cnst     { tcount_in, nextLink_wave_in, nextLink_tid_in }, max_count_in, 0
        wtd     RA_WAVE, RA

	 wa      RA_WAVE_wa, {RA_WAVE, RA_WAVE_lc}
        wa      barrier_tid_wa, { barrier_tid_in, barrier_tid_lc }
        wa      max_count_wa, { max_count_in, max_count_lc }
        wa      tcount_wa, { tcount_in, tcount_lc }
        wa      nextLink_wave_wa, { nextLink_wave_in, nextLink_wave_lc }
        wa      nextLink_tid_wa, { nextLink_tid_in, nextLink_tid_lc }
	 wtd     ^bwn_tc:2, barrier_tid_wa ; The put that allows this iter to receive

        ;; if(isBarrierMsg(req.msg))
        EQI     isBarrier, rbreq_msg_pad_OUT, 0
        rho     max_count_rp, { }, max_count_wa, isBarrier
        rho     tcount_rp, tcount_nrp, tcount_wa, isBarrier
        rho     nextLink_wave_rp, nextLink_wave_nrp, nextLink_wave_wa, isBarrier
        rho     nextLink_tid_rp, nextLink_tid_nrp, nextLink_tid_wa, isBarrier
        rho     rbreq_tid_pad_rp, { }, rbreq_tid_pad_OUT, isBarrier
        rho     rbreq_wave_pad_rp, { }, rbreq_wave_pad_OUT, isBarrier

        ;; isBarrierMsg(req.msg): TRUE ; data guided along this path ends "_rp"
        dttw    barrier_wait::rcl_tid_pad_IN, rbreq_tid_pad_rp, rbreq_wave_pad_rp, nextLink_tid_rp
        dttw    barrier_wait::rcl_wave_pad_IN, rbreq_tid_pad_rp, rbreq_wave_pad_rp, nextLink_wave_rp 

        AddI    new_tcount_rp, tcount_rp, 1
        cnst     zero_rp, new_tcount_rp, 0
        phi     tcount_loop_rp, zero_rp, new_tcount_rp, max_count_reached_rp

        ;; if(++tcount == max_tcount)
        EQ      max_count_reached_rp, new_tcount_rp, max_count_rp
        rho     rbreq_tid_mc, { }, rbreq_tid_pad_rp, max_count_reached_rp
        rho     rbreq_wave_mc, { }, rbreq_wave_pad_rp, max_count_reached_rp
        rho     new_tcount_mc, { }, new_tcount_rp, max_count_reached_rp

	 ;; send "chain_release(tcount -1)" to requestor
        SubI    decr_tcount_for_release, { new_tcount_mc, tcount_nrp }, 1
        dttw    ^barrier_wait::rcr_length_pad:1, { rbreq_tid_mc, nextLink_tid_nrp }, \
                { rbreq_wave_mc, nextLink_wave_nrp }, decr_tcount_for_release

        ;; Not a barrier (isBarrierMsg(req.msg) == FALSE); some kind of reset
        cnst     tcount_loop_nrp, tcount_nrp, 0

	 ;; if(isSetNewMaxTCountMsg(req.msg))
        GTI     isChangeMax, rbreq_msg_pad_OUT, 0
        phi     new_max_count, rbreq_msg_pad_OUT, max_count_wa, isChangeMax

        ;; Loop exit test
        EQI     isDestroyBarrier, rbreq_msg_pad_OUT, -2
        rho     RA_WAVE_exit, RA_WAVE_lc, RA_WAVE_wa, isDestroyBarrier
        rho     { }, barrier_tid_lc, barrier_tid_wa, isDestroyBarrier
        rho     { }, max_count_lc, new_max_count, isDestroyBarrier
        rho     { }, tcount_lc, { tcount_loop_rp, tcount_loop_nrp }, isDestroyBarrier
        rho     { }, nextLink_wave_lc, rbreq_wave_pad_OUT, isDestroyBarrier
        rho     { }, nextLink_tid_lc, rbreq_tid_pad_OUT, isDestroyBarrier

        wtd     curr_wave, RA_WAVE_exit
        dtw     curr_wave_in_old, RA_WAVE_exit, curr_wave
        is      curr_wave_in_old, RA, 0
.endproc

;;;
;;; Barrier resetting procedure
;;; 
;;; To change the barrier's max count as part of the reset process, pass
;;; the new max count as the second parameter.  Any value greater than
;;; zero is considered to be a valid new max count.
;;; 

.proc   barrier_reset ( bBarrierReset_tid_in, new_max_count )
.export ^barrier_reset as barrier_reset
        ;; Get the wave next wave number in the barrier thread
        cnst    { ^bwn_rho:3, ^bwn_tc:1}, bBarrierReset_tid_in, 0
        ;; Construct a reset message
        wtd     req_wave, bBarrierReset_tid_in
        ttd     req_tid, bBarrierReset_tid_in
        EQI     isNewMaxCountZero, new_max_count, 0
        cnst    minusone, new_max_count, -1
        phi     req_max_count, minusone, new_max_count, isNewMaxCountZero
        ;; Send the reset message
        dttw    barrier_thread::rbreq_tid_pad_IN, bBarrierReset_tid_in, ^bwn_rho:1, req_tid
        dttw    barrier_thread::rbreq_wave_pad_IN, bBarrierReset_tid_in, ^bwn_rho:1, req_wave
        dttw    barrier_thread::rbreq_msg_pad_IN, bBarrierReset_tid_in, ^bwn_rho:1, req_max_count
        wtd     RA_WAVE, RA
        is      RA_WAVE, RA, 0
.endproc

;;; 
;;; Get behind the barrier; come on, everyone's doing it.
;;; 
.proc barrier_wait(bBarrierWait_tid_in)
.export ^barrier_wait as barrier_wait

;;; receive_chain_link() landing pads
rcl_tid_pad:
        identity    rcl_tid_pad_OUT, rcl_tid_pad_IN
rcl_wave_pad:
        identity    rcl_wave_pad_OUT, rcl_wave_pad_IN
;;; end of receive_chain_link() landing pads

;;; receive_chain_release() landing pad
rcr_length_pad:
        identity    rcr_length_pad_OUT, rcr_length_pad_IN
;;; end of receive_chain_release() landing pad

        ;; Get the next wave number in the barrier thread
        cnst     {^bwn_rho:3, ^bwn_tc:1}, bBarrierWait_tid_in, -1
        ;; Construct a barrier request message
        wtd     req_wave, bBarrierWait_tid_in
        ttd     req_tid, bBarrierWait_tid_in
        cnst    req_msg, bBarrierWait_tid_in, 0
        ;; Send the barrier request message
        dttw    barrier_thread::rbreq_tid_pad_IN, bBarrierWait_tid_in, ^bwn_rho:0, req_tid
        dttw    barrier_thread::rbreq_wave_pad_IN, bBarrierWait_tid_in, ^bwn_rho:0, req_wave
        dttw    barrier_thread::rbreq_msg_pad_IN, bBarrierWait_tid_in, ^bwn_rho:0, req_msg

	 ;; if(length) send_chainRelease(nextLink, length - 1)
        NEI     hasNext, rcr_length_pad_OUT, 0
        rho     length_hn, { }, rcr_length_pad_OUT, hasNext
        rho     nextLink_tid_hn, { }, rcl_tid_pad_OUT, hasNext
        rho     nextLink_wave_hn, { }, rcl_wave_pad_OUT, hasNext
        SubI    decr_length_hn, length_hn, 1
        dttw    rcr_length_pad_IN, nextLink_tid_hn, nextLink_wave_hn, decr_length_hn

	 ;; exit code.  Caller must make subsequent instructions dependent on the result returned by the
        ;; second indirect send in order to cause the instruction to wait on the barrier
        wtd     RA_WAVE, RA
        is      RA_WAVE, RA, 0
        is      rcr_length_pad_OUT, RA, 1
.endproc
