/**
 * Assume the following assembler macros:
 *
 * @macro mstruct take, take_ra, put
 *     tc takenVal, take, put
 *     is take_ra, takenVal
 * @endmacro
 *
 * @macro take mstructLabel, mtid, ra
 *     send mstructLabel^0, mtid
 *     send mstructLabel^1, ra
 * @endmacro
 *
 * @macro put mstructLabel, mtid, val
 *     dtt mstructLabel^2, val, mtid
 * @endmacro
 */

int take(TID tid, addr mstruct)
{
  asm ( "take mstruct, tid, $RA" );
  // The above assembly actually causes the return to occur
}

void put(TID tid, addr mstruct, int val)
{
  asm ("put mstruct, tid, val");
}

void takeandput(TID tid, addr mstruct)
{
  int val = take(tid, mstruct);
  put(tid, mstruct, val);
  return val;
}

typedef struct _Barrier {
  TID tid;
  addr tcount, max_tcount, waiting_threads;
} Barrier;

Barrier barrierCreate( int barrier_max_tcount )
{
  Barrier b;
  b.tid = allocateBarrierTid(/*?*/);
  b.tcount = allocateMstruct(/*?*/);
  b.max_tcount = allocateMstruct(/*?*/);
  b.waitingThreads = allocateMstruct(/*?*/);
  
  put(b.tid, b.tcount, 0);
  put(b.tid, b.max_tcount, barrier_max_tcount);
  return b;
}

int barrier(Barrier b)
{
  const TID tid = b.tid;
  const addr waitingThreads = b.waitingThreads;
  const addr tcount = b.tcount;
  const addr max_tcount = b.max_tcount;
  const int max_tcount = takeandput(tid, max_tcount);

  int new_tcount = take(tid, tcount) + 1;
  if(new_tcount == max_tcount)
    while(new_tcount--) put(tid, waitingThreads, new_tcount);

  take_nb(tid, waitingThreads);
  put(tid, tcount, new_tcount);

/*   asm{ */
/*     take tid, waitingThreads, $+1^1\n; */
/*     send finished, { } ; */
/*     pass ftid, tid, finished ; */
/*     is $RA, finished ; */
/*     put tcount, ftid, new_tcount ; */
/*   }; */
}
