;;; FILE:        barrier_a.s
;;; AUTHOR:      Andrew Schwerin (schwerin@cs.washington.edu)
;;; DESCRIPTION: An implementation in wave-assembly of the barrier described in
;;;              very C-like pseudocode in barrier_a.c
;;;
;;; $Id: barrier_test.s,v 1.1 2003/09/05 23:30:13 schwerin Exp $
;;; 
;;; NOTE:
;;; The barrier request message field is either 0, to get behind the barrier,
;;; or it is < 0 to reset the barrier, or it is > 0 to reset the barrier and set the
;;; new threshold to the value of the message.  If the request message is -2, the
;;; barrier is destroyed after it is reset, rather than re-initialized.
;;; 

@macro eqz result, val
        ceqi result, val, 0
@endmacro 

@macro nez result, val
        neqi result, val, 0
@endmacro
        
.section .code
        
        tc     bwn_takeout, {bBarrierReset_tid_in, bBarrierWait_tid_in} , bwn_put
        rho    bwn_to_barrier, bwn_to_reset, bwn_takeout, barrier_req_isnt_reset

.proc barrier_thread ( barrier_tid_in, max_count_in )
;;; receive_barrier_request()
rbreq_tid_pad:
        send    rbreq_tid_pad_OUT, rbreq_tid_pad_IN
rbreq_wave_pad: 
        send    rbreq_wave_pad_OUT, rbreq_wave_pad_IN
rbreq_msg_pad: 
        send    rbreq_msg_pad_OUT, rbreq_msg_pad_IN
;;; end of receive_barrier_request()

        .scope
        lda     { tcount_in, nextLink_wave_in, nextLink_tid_in }, 0, max_count_in

        wa      barrier_tid_wa, { barrier_tid_in, barrier_tid_lc }
        wa      max_count_wa, { max_count_in, max_count_lc }
        wa      tcount_wa, { tcount_in, tcount_lc }
        wa      nextLink_wave_wa, { nextLink_wave_in, nextLink_wave_lc }
        wa      nextLink_tid_wa, { nextLink_tid_in, nextLink_tid_lc }
	 wtd     bwn_put, barrier_tid_wa ; The put that allows this iter to receive

        ;; if(isBarrierMsg(req.msg))
        eqz     isBarrier, req_msg
        rho     max_count_rp, { }, mac_count_wa, isBarrier
        rho     tcount_rp, tcount_nrp, tcount_wa, isBarrier
        rho     nextLink_wave_rp, nextLink_wave_nrp, nextLink_wave_wa, isBarrier
        rho     nextLink_tid_rp, nextLink_tid_nrp, nextLink_tid_wa, isBarrier
        rho     rbreq_tid_pad_rp, { }, rbreq_tid_pad_OUT, isBarrier
        rho     rbreq_wave_pad_rp, { }, rbreq_wave_pad_OUT, isBarrier

        ;; isBarrierMsg(req.msg): TRUE ; data guided along this path ends "_rp"
        dttw    rcl_tid_pad_IN, rbreq_tid_pad_rp, rbreq_wave_pad_rp, nextLink_tid_rp
        dttw    rcl_wave_pad_IN, rbreq_tid_pad_rp, rbreq_wave_pad_rp, nextLink_wave_rp 

        addi    new_tcount_rp, tcount_rp, 1
        lda     zero_rp, 0, new_tcount_rp
        phi     tcount_loop_rp, zero_rp, new_tcount_rp, max_count_reached_rp

        ;; if(++tcount == max_tcount)
        ceq     max_count_reached_rp, new_tcount_rp, max_count_rp
        rho     rbreq_tid_mc, { }, rbreq_tid_pad_rp, max_count_reached_rp
        rho     rbreq_wave_mc, { }, rbreq_wave_pad_rp, max_count_reached_rp
        rho     new_tcount_mc, { }, new_tcount_rp, max_count_reached_rp

	 ;; send "chain_release(tcount -1)" to requestor
        subi    decr_tcount_for_release, { new_tcount_mc, tcount_nrp }, 1
        dttw    rcr_len_pad, { rbreq_tid_mc, nextLink_tid_nrp }, \
                { rbreq_wave_mc, nextLink_wave_nrp }, decr_tcount_for_release

        ;; Not a barrier (isBarrierMsg(req.msg) == FALSE); some kind of reset
        lda     tcount_loop_nrp, 0, tcount_nrp

	 ;; if(isSetNewMaxTCountMsg(req.msg))
        gti     isChangeMax, rbreq_msg_pad_OUT, 0
        phi     new_max_count, rbreq_msg_pad_OUT, max_count_wa, isChangeMax

        ;; Loop exit test
        ceq     isDestroyBarrier, rbreq_msg_pad_OUT, -2
        rho     { }, barrier_tid_lc, barrier_tid_wa, isDestroyBarrier
        rho     { }, max_count_lc, new_max_count, isDestroyBarrier
        rho     { }, tcount_lc, { tcount_loop_rp, tcount_loop_nrp }, isDestroyBarrier
        rho     { }, nextLink_wave_lc, rbreq_wave_pad_OUT, isDestroyBarrier
        rho     { }, nextLink_tid_lc, rbreq_tid_pad_OUT, isDestroyBarrier
        .endscope
.endproc

;;;
;;; Barrier resetting procedure
;;; 
;;; To change the barrier's max count as part of the reset process, pass
;;; the new max count as the second parameter.  Any value greater than
;;; zero is considered to be a valid new max count.
;;; 

.proc   barrier_reset ( bBarrierReset_tid_in, new_max_count )
        .scope
        ;; Get the wave next wave number in the barrier thread
        lda     barrier_req_isnt_reset, FALSE, bBarrierReset_tid_in
        ;; Construct a reset message
        wtd     req_wave, bBarrierReset_tid_in
        ttd     req_tid, bBarrierReset_tid_in
        eqz     isNewMaxCountZero, new_max_count
        lda     minusone, -1, new_max_count
        phi     req_max_count, minusone, new_max_count, isNewMaxCountZero
        ;; Send the reset message
        dttw    rbreq_tid_pad_IN, bBarrierReset_tid_in, bwn_to_reset, req_tid
        dttw    rbreq_wave_pad_IN, bBarrierReset_tid_in, bwn_to_reset, req_wave
        dttw    rbreq_msg_pad_IN, bBarrierReset_tid_in, bwn_to_reset, req_max_count
        .endscope
.endproc

;;; 
;;; Get behind the barrier; come on, everyone's doing it.
;;; 
.proc barrier_wait(bBarrierWait_tid_in)

;;; receive_chain_link() landing pads
rcl_tid_pad:
        send    rcl_tid_pad_OUT, rcl_tid_pad_IN
rcl_wave_pad:
        send    rcl_wave_pad_OUT, rcl_wave_pad_IN
;;; end of receive_chain_link() landing pads

;;; receive_chain_release() landing pad
rcr_length_pad:
        send    rcr_length_pad_OUT, rcr_length_pad_IN
;;; end of receive_chain_release() landing pad

        .scope
        ;; Get the next wave number in the barrier thread
        lda     barrier_req_isnt_reset, TRUE, bBarrierWait_tid_in
        ;; Construct a barrier request message
        wtd     req_wave, bBarrierWait_tid_in
        ttd     req_tid, bBarrierWait_tid_in
        lda     req_msg, 0, bBarrierWait_tid_in
        ;; Send the barrier request message
        dttw    rbreq_tid_pad_IN, bBarrierWait_tid_in, bwn_to_barrier, req_tid
        dttw    rbreq_wave_pad_IN, bBarrierWait_tid_in, bwn_to_barrier, req_wave
        dttw    rbreq_msg_pad_IN, bBarrierWait_tid_in, bwn_to_barrier, req_msg

	 ;; if(length) send_chainRelease(nextLink, length - 1)
        nez     hasNext, rcr_length_pad_OUT
        rho     length_hn, { }, rcr_length_pad_OUT, hasNext
        rho     nextLink_tid_hn, { }, rcl_tid_pad_OUT, hasNext
        rho     nextLink_wave_hn, { }, rcl_wave_pad_OUT, hasNext
        subi    decr_length_hn, length_hn, 1
        dttw    rcr_length_pad_OUT, nextLink_tid_hn, nextLink_wave_hn, decr_length_hn
        is      RA, rcr_length_pad_OUT
        .endscope
.endproc
